-- DIALOG.E v1.4
-- Smaller implementation of Text-based GUI

without warning
without trace

include get.e
include graphics.e
include image.e
include mouse.e

-- dialog window styles
global constant
    YES_NO      = 1,            -- dialog with "Yes" and "No" buttons
    OK_CANCEL   = 2,            -- dialog with "OK" and "Cancel" buttons
    OK          = 3,            -- dialog with "OK" button

-- colors
    BackColor   = WHITE,        -- dialog background
    TextColor   = BLACK,        -- dialog text colors
    DataColor   = BLACK,        -- text data colors
    LitColor    = BRIGHT_WHITE, -- dialog lit color

-- key definitions      
    Backspace   = 8,    
    Tab         = 9,
    Enter       = 13,
    Escape      = 27,
    Space       = 32,
    ShiftTab    = 271,
    Home        = 327,
    Up          = 328,
    PageUp      = 329,
    Left        = 331,
    Right       = 333,
    End         = 335,
    Down        = 336,
    PageDown    = 337
    

constant
    Type    = 1,    -- type of control
    Title   = 2,    -- title of control
    Row     = 3,    -- row of title
    Col     = 4,    -- column of title
    Wide    = 5,    -- width of control
    Deep    = 6,    -- depth of control
    Data    = 7,    -- data value of control
    Anchor  = 8,    -- offset into data values
    Cursor  = 9,    -- current data value indexed
    Hotkey  = 10    -- character to highlight
    
global constant
    Button  = 1,
    Check   = 2,
    Label   = 3,
    List    = 4,
    Radio   = 5,
    Sle     = 6

integer     
    wRow,       -- window starting row
    wCol        -- window starting column
    
global sequence
    controls,
    helpText
    
    controls = {}
    helpText = {}

sequence
    wTitle     -- window title
    
constant 
    minWide     = 30    -- minimum size for a dialog

-----------------------------------------------------------------------------
function center( sequence text, integer wide )
    
    -- return offset to center text
    return floor((wide/2) - (length( text )/2))

end function



-----------------------------------------------------------------------------
procedure w_position( integer row, integer col )
    
    -- position relative to window
    position( wRow+row, wCol+col )
    
end procedure


-----------------------------------------------------------------------------    
procedure shadow( integer row, integer col, sequence image, integer wide )

    -- "shadows" the bytes in the screen image

    integer at
    
    -- position in image    
    at = ((col) * 2)-1

    -- color
    text_color( GRAY )
    bk_color( BLACK )
    
    -- bytes
    for i = 1 to wide do
        position( row, col-1+i )
        puts( 1, image[row][at] )
        at = at + 2        
    end for
    
    -- restore colors
    text_color( TextColor )
    bk_color( BackColor )
    
end procedure
    
    

-----------------------------------------------------------------------------    
procedure draw_window( integer wide, integer deep, sequence title, sequence image )

    integer tmp

    -- calculate center
    wTitle = title
    wRow = floor( (25 - deep)/2 )
    wCol = floor( (80 - wide)/2 )

    -- set colors
    bk_color( BackColor )
    text_color( TextColor )

    -- draw top
    w_position( 0, 0 )
    puts( 1, 218 & repeat( 196, wide ) & 191 )    

    -- title
    if length( title ) then
        w_position( 0, center( title & " ", wide ) )
        puts( 1, " " & title & " " )
    end if

    -- body of window    
    for i = 1 to deep-3 do   
        
        -- position
        w_position( i, 0 )
        puts( 1, 179 & repeat( ' ', wide ) & 179 )
        
        -- shadow
        shadow( wRow+i, wCol+wide+2, image, 2 )
        
    end for    

    -- buttons go at the bottom        
    tmp = deep - 2
    
    -- bar
    w_position( tmp, 0 )
    puts( 1, 195 & repeat( 196, wide ) & 180 )
    
    -- shadow
    shadow( wRow+tmp, wCol+wide+2, image, 2 )
        
    -- space for buttons
    w_position( tmp+1, 0 )
    puts( 1, 179 & repeat( ' ', wide ) & 179 )
    
    -- shadow
    shadow( wRow+tmp+1, wCol+wide+2, image, 2 )
        
    -- bottom
    w_position( tmp+2, 0 )
    puts( 1, 192 & repeat( 196, wide ) & 217 )
    
    -- shadow
    shadow( wRow+tmp+2, wCol+wide+2, image, 2 )
        
    -- shadow
    shadow( wRow+tmp+3, wCol+2, image, wide+2 )

end procedure


-----------------------------------------------------------------------------
global procedure add_control(  integer type_of,
                        sequence title,
                        integer row,
                        integer col,
                        integer wide,
                        integer deep,
                        object data )

    integer at
    sequence control
    
    control = repeat( 0, 10 )

    -- look for hotkey
    at = find( "&", title )
    if at then
        -- remove ampersand
        title = title[1..at-1] & title[at..length(title)]
    end if

    
    control[Type]   = type_of
    control[Title]  = title
    control[Row]    = row
    control[Col]    = col
    control[Wide]   = wide
    control[Deep]   = deep
    control[Data]   = data
    control[Anchor] = 0
    control[Cursor] = 1
    control[Hotkey] = at
    
    controls = append( controls, control )
                        
end procedure

-----------------------------------------------------------------------------
global procedure add_button(   sequence title,
                        integer wide )

    controls = append( controls,  
        {Button, title, 0, 0, wide, 0, 0} )
                        
end procedure


-----------------------------------------------------------------------------
procedure draw_button( sequence self, integer active )

    -- draw control button                       
    
    integer at
    
    w_position( self[Row], self[Col] )

    -- light, if active        
    if active then
        text_color( LitColor )
    end if             
    
    -- draw button
    puts( 1, "<" & repeat( ' ', self[Wide]+2) & ">" )   
        
    -- clear light
    text_color( TextColor )

    -- text position
    at = center( self[Title], self[Wide] )+2
        
    w_position( self[Row], self[Col] + at )
    puts( 1, self[Title] )
        
    -- position cursor
    w_position( self[Row], self[Col] + at )

end procedure


-----------------------------------------------------------------------------
procedure draw_check( sequence self, integer active )

    -- draw check box

    -- position control
    w_position( self[Row], self[Col] )

    -- data color
    if active then
        text_color( DataColor )
    else
        text_color( TextColor )
    end if

    -- check box    
    if self[Data] then
        puts( 1, "[X] " )        
    else
        puts( 1, "[ ] " )
    end if      

    -- normal color
    text_color( TextColor )
    
    -- title
    puts( 1, self[Title] )
    
    -- position cursor
    w_position( self[Row], self[Col]+1 )
        
end procedure


-----------------------------------------------------------------------------
procedure draw_radio( sequence self, integer active )

    -- draw radio button
    
    -- position control
    w_position( self[Row], self[Col] )

    -- data color
    if active then
        text_color( DataColor )
    else
        text_color( TextColor )
    end if
    
    -- button
    if self[Data] then
        puts( 1, "(" & 7 & ") " )
    else
        puts( 1, "( ) " )            
    end if               

    -- normal color
    text_color( TextColor )
    
    -- title
    puts( 1, self[Title] )

    -- position cursor
    w_position( self[Row], self[Col]+1 )

end procedure


-----------------------------------------------------------------------------
procedure draw_sle( sequence self, integer active )

    -- draw a single line edit        
                              
    integer at
                              
    -- position of label
    w_position( self[Row], self[Col] )
    puts( 1, self[Title] & " " )

    -- position of frame
    at = self[Col] + length( self[Title] ) + 1

    -- top of frame
    w_position( self[Row]-1, at )
    puts( 1, 218 & repeat( 196, self[Wide] ) & 191 )

    -- middle
    w_position( self[Row], at )
    puts( 1, 179  )

    -- active?    
    if active then
        text_color( DataColor )
    end if

    -- sle data            
    puts( 1, self[Data] & repeat( ' ', self[Wide] - length( self[Data] ) ) )
            
    -- normal color
    text_color( TextColor )
    
    -- right side of SLE
    puts( 1, 179 ) 

    -- bottom of frame
    w_position( self[Row]+1, at )
    puts( 1, 192 & repeat( 196, self[Wide] ) & 217 )
        
    -- position cursor
    w_position( self[Row], at+length(self[Data])+1 )

end procedure


-----------------------------------------------------------------------------
function list_thumb_at( sequence self )

    -- returns the position of the list thumb
    -- or 0 if no thumb

    integer at
    
    -- position thumb
    if length( self[Data] ) > self[Deep] then
        at = self[Anchor] + self[Cursor]
        -- thumb position
        return floor( (at / length( self[Data] ) ) * (self[Deep]-3) )+1
    else
        -- non-existant
        return 0
    end if

end function


-----------------------------------------------------------------------------
procedure draw_list( sequence self, integer active )

    -- draw a list box
    
    integer at, thumb
    sequence text

    -- label
    w_position( self[Row], self[Col]+1 )
    puts( 1, self[Title] )
    
    -- top of box
    at = self[Row]
    w_position( at+1, self[Col] )
    puts( 1, 218 & repeat( 196, self[Wide] ) & 191 )
    
    -- middle
    for i = 1 to self[Deep] do
        if self[Anchor]+i > length( self[Data] ) then
            -- blank
            text = repeat( ' ', self[Wide] )
            
        else
            -- assign text
            text = self[Data][self[Anchor]+i]
            
            -- clip text
            if length( text ) > self[Wide] then
                -- trim
                text = text[1..self[Wide]] 
            else
                -- pad
                text = text & repeat( ' ', self[Wide] - length(text) )
            end if      
        end if

        -- position
        w_position( at+i+1, self[Col] )

        if  i = self[Cursor] then
            
            -- left border
            puts( 1, 179 )
            
            -- reverse colors
            text_color( BackColor )
            bk_color( TextColor )
            
            -- selection
            puts( 1, text )
            
            -- normal colors
            text_color( TextColor )
            bk_color( BackColor )

        else
            -- plain text
            puts( 1, 179 & text )
        end if

        -- right border
        if length( self[Data] ) > self[Deep] then
        
            -- reverse colors
            text_color( BackColor )
            bk_color( TextColor )
                    
            if i = 1 then
                -- up arrow
                puts( 1, 24 )
                
            elsif i = self[Deep] then                    
                -- down arrow
                puts( 1, 25 )
                
            else                
                -- border
                puts( 1, 178 )
            end if
        
            -- normal colors
            text_color( TextColor )
            bk_color( BackColor )
            
        else
            -- plain border
            puts( 1, 179 ) 
        end if
            
    end for                                        
 
    -- position thumb
    thumb = list_thumb_at( self )
    if thumb > 0 then
        w_position( thumb+3, self[Col]+self[Wide]+1 )
        puts( 1, 219 )
    end if
    

    -- bottom of box
    w_position( self[Row]+self[Deep]+2, self[Col] )
    puts( 1, 192 & repeat( 196, self[Wide] ) & 217 )

    -- title
    text = " " & self[Title]
    
    -- position cursor
    w_position( self[Row]+self[Cursor]+1, self[Col]+1 )

end procedure


-----------------------------------------------------------------------------
procedure draw_label( sequence self, integer active )

    -- draw a label

    w_position( self[Row], self[Col] )
    puts( 1, self[Title] )
        
end procedure

-----------------------------------------------------------------------------
procedure draw_control( sequence self, integer active )

    -- draw the given control

    if    self[Type] = Button   then draw_button( self, active )
    elsif self[Type] = Check    then draw_check( self, active )
    elsif self[Type] = Radio    then draw_radio( self, active )
    elsif self[Type] = Sle      then draw_sle( self, active )
    elsif self[Type] = List     then draw_list( self, active )
    elsif self[Type] = Label    then draw_label( self, active )
    end if

end procedure
    

-----------------------------------------------------------------------------
function edit_sle( sequence self, integer key )

    if key = Backspace then

        if length( self[Data] ) then
            self[Data] = self[Data][1..length(self[Data])-1]
        end if

    elsif key >= ' ' 
    and   key <= 126 then            
        -- alpha numeric
        if length( self[Data] ) < self[Wide] then
            self[Data] = self[Data] & key
        end if

    end if

    return self
    
end function



-----------------------------------------------------------------------------
function edit_check( sequence self, integer key )

    if key = Space then
        -- toggle
        self[Data] = not self[Data]        
    end if
    
    return self
        
end function



-----------------------------------------------------------------------------
function edit_radio( sequence self, integer key )    

    if key = Space then
        
        -- clear all radios
        for i = 1 to length( controls ) do
            if controls[i][Type] = Radio then
                controls[i][Data] = 0
                draw_control( controls[i], 0 )
            end if
        end for
        
        -- set this one on
        self[Data] = 1
        
    end if
    
    return self
        
end function


-----------------------------------------------------------------------------
function edit_list( sequence self, integer key )

    if key = Up then
    
        if self[Cursor] = 1 then
            -- move anchor up 1 item
            if self[Anchor] > 0 then
                self[Anchor] = self[Anchor] - 1
            end if
        else  
            -- move up 1 item
            self[Cursor] = self[Cursor] - 1
        end if
        
    elsif key = Down then

        -- room to move?
        if self[Anchor]+self[Cursor] < length( self[Data] ) then
            -- move cursor?            
            if self[Cursor] < self[Deep] then
                -- move cursor down 1
                self[Cursor] = self[Cursor] + 1
            else
                -- move anchor down 1
                self[Anchor] = self[Anchor] + 1
            end if
        end if

    elsif key = PageUp then

        -- move to top if list less than full
        if length( self[Data] ) <= self[Deep] then
            -- move to bottom of list
            self[Cursor] = 1
            
        else                   
        
            -- page up
            self[Anchor] = self[Anchor] - self[Deep]

            -- gone too far?
            if self[Anchor] < 0 then
                -- position at top
                self[Anchor] = 0
                self[Cursor] = 1
            end if
        end if

    elsif key = PageDown then

        -- move to bottom if list less than full
        if length( self[Data] ) <= self[Deep] then
            -- move to bottom of list
            self[Cursor] = length( self[Deep] )
        else
            -- page down
            self[Anchor] = self[Anchor] + self[Deep]
            
            -- gone too far?
            if self[Anchor] + self[Deep] > length( self[Data] ) then
                -- position at bottom
                self[Anchor] = length( self[Data] ) - self[Deep]
                self[Cursor] = self[Deep]
            end if
        end if
    

    elsif key = Home then
        -- move to top of list
        self[Anchor] = 0
        self[Cursor] = 1
        
    elsif key = End then
        -- move to bottom of list
        
        -- data shorter than list?
        if length( self[Data] ) <= self[Deep] then
            self[Cursor] = length( self[Deep] )
        else
            self[Anchor] = length( self[Data] ) - self[Deep]
            self[Cursor] = self[Deep]
        end if

    end if
            
    return self
        
end function


-----------------------------------------------------------------------------
function edit_control( sequence self, integer key )    

    -- select correct handler for control


    if    self[Type] = Radio        then return edit_radio( self, key )            
    elsif self[Type] = Check        then return edit_check( self, key )            
    elsif self[Type] = Sle          then return edit_sle( self, key )
    elsif self[Type] = List         then return edit_list( self, key )
    
    end if

    return self

end function   


----------------------------------------------------------------------------- 
procedure position_buttons( integer wide, integer deep )

    integer totalWide, count, space, tmp

    -- position the buttons in the button portion of the window
    
    -- calculate the space taken                            
    count = 0
    totalWide = 0
    for i = 1 to length( controls ) do
        if controls[i][Type] = Button then
            count = count + 1
            totalWide = totalWide + controls[i][Wide] + 4
        end if        
    end for
    
    -- any buttons?

    if count > 0 then
        -- divide evenly        
        space = floor( (( wide - totalWide ) / (count+1)) )
        tmp = 2 + space
        for i = 1 to length( controls ) do
            if controls[i][Type] = Button then
                controls[i][Col] = tmp
                controls[i][Row] = deep-1
                tmp = tmp + controls[i][Wide] + 4 + space
            end if        
        end for
    end if

end procedure


-----------------------------------------------------------------------------
function mouse_click_list( sequence mouse, sequence self )

    -- handle a mouse click on a list

    integer row, col, wide
    integer thumb, at
    

    -- calculate positions
    row = self[Row] + wRow + 2
    col = self[Col] + wCol
    wide = self[Wide] + 2
    
    -- on a control?
    if mouse[2] = col + wide - 1 then

        -- up arrow?
        if mouse[3] = row then

            -- room to move?
            if self[Anchor] > 0 then
                self[Anchor] = self[Anchor]-1
            end if

            
        -- down arrow?
        elsif mouse[3] = row + self[Deep] - 1 then
            
            -- room to move?
            if self[Anchor]+self[Deep] < length( self[Data] ) then
                -- move
                self[Anchor] = self[Anchor]+1
            end if
            
        -- scroll bar
        else

            -- find thumb position
            thumb = list_thumb_at( self )

            -- is there a thumb?
            if thumb > 0 then

                -- mouse position on scroll bar
                at = mouse[3] - row
            
                -- move towards the thumb
                if at < thumb then
                    -- page down
                    self = edit_list( self, PageUp )
                
                elsif at > thumb then                
                    -- page up
                    self = edit_list( self, PageDown )
                end if

            end if

        end if
        
    -- clicked inside list
    else
    

        -- position in list
        self[Cursor] = (mouse[3] - row) + 1

        -- past end of list?
        if self[Anchor] + self[Cursor] > length( self[Data] ) then
            -- clip
            self[Cursor] = length( self[Data] ) - self[Anchor]
        end if

    end if

    -- redraw list
    draw_control( self, 1 )

    return self
            
end function


-----------------------------------------------------------------------------
function mouse_control()

    -- this points to current control
    -- return { control, key }
    -- or -1 if no event
    
    integer row, col, wide, deep, clicked, key
    object mouse


    -- look for a mouse event
    mouse = get_mouse()
    if not sequence( mouse ) then
        -- no mouse event
        return -1
    end if
    
    -- look for event
    if not and_bits(mouse[1], LEFT_DOWN) then
        -- no click event
        return -1        
    end if
    
    -- scale the mouse controls
    mouse[2..3]=mouse[2..3]/8+1

    -- clicked control
    clicked = 0

    -- examine each control    
    for i = 1 to length( controls ) do

        -- skip labels    
        if controls[i][Type] != Label then
        
            -- default position
            row = controls[i][Row] + wRow
            col = controls[i][Col] + wCol
            wide = controls[i][Wide]
            deep = 1 
                
            -- adjust for specific controls
            if controls[i][Type] = Button then
                -- control button                       
                wide = wide + 4
                
            elsif controls[i][Type] = Check then
                -- check box   
                wide = wide + length( controls[i][Title] ) + 2
            
            elsif controls[i][Type] = Radio then
                -- radio button
                wide = wide + length( controls[i][Title] ) + 4
            
            elsif controls[i][Type] = Sle then
                wide = wide + length( controls[i][Title] ) + 2
    
            elsif controls[i][Type] = List then
                wide = wide + 2
                deep = controls[i][Deep]
                row = row + 2
    
            end if         

            -- same row?
            if  mouse[3] >= row 
            and mouse[3] <= row+deep-1 then
            
                -- same col?       
                if  mouse[2] >= col
                and mouse[2] <= col+wide-1 then
                
                    -- return match
                    clicked = i
                    exit
                    
                end if
            end if
        end if
    end for

    -- clicked an item?
    if clicked = 0 then
        -- no control clicked
        return -1        
    end if
    
    -- if list, needs special handling
    if controls[clicked][Type] = List then                 
        -- routine to handle click
        controls[clicked] = mouse_click_list( mouse, controls[clicked] )
    end if

    -- choose key, based on type of controls
    if controls[clicked][Type] = Sle then
        key = 0
        
    elsif controls[clicked][Type] = Button then
        key = Space
        
    else
        key = Space
        
    end if

    return { clicked, key }    
    
end function



-----------------------------------------------------------------------------    
function build_msgbox( integer style, sequence text )

    integer wide, deep
    sequence screen
    
    -- creates a dialog in a given style
    
    -- was one line passed, or many?
    if not sequence( text[1] ) then
        -- place the text in a sequence
        if length( text ) < minWide then
            -- add padding to center it       
            text = repeat( ' ', floor( (minWide-length(text)) /2 ) )
                    & text
        end if
        
        -- put it in a sequence
        text = { text }
        
    end if

    -- clear the controls
    controls = {}

    -- build labels, and find widest line
    wide = minWide
    for i = 1 to length( text ) do                
        -- add label
        add_control( Label, text[i], i+1, 4, 0, 0, 0 )

        -- widest encountered line    
        if length( text[i] )  > wide then
            wide = length( text[i] )
        end if
        
    end for
    wide = wide + 6
    
    -- depth
    deep = length( text ) + 5
    
    -- save screen
    screen = save_text_image( {1, 1}, {25, 80} )
    
    -- build the buttons
    if style = YES_NO then
        add_button( "Yes", 6 )
        add_button( "No", 6 )

    elsif style = OK_CANCEL then
        add_button( "OK", 6 )
        add_button( "Cancel", 6 )

    else
        add_button( "OK", 6 )

    end if

    -- return window size
    return { wide, deep }
end function


----------------------------------------------------------------------------- 
global function run_dialog( integer wide, integer deep, sequence title )

    integer this, key, oldRow, oldCol, defaultButton
    sequence text, oldControls, oldTitle, result, screen
    object tmp


    -- save screen
    screen = save_text_image( {1, 1}, {25, 80} )

    -- calculate button positions
    position_buttons( wide, deep )
                     
    -- put up window
    draw_window( wide, deep, title, screen )
                     
    -- draw all controls; set defaults
    defaultButton = 0
    this = 0
    for i = 1 to length( controls ) do
        -- default button is first button
        if  controls[i][Type] = Button
        and defaultButton = 0 then
            defaultButton = i            
        end if
        
        -- focus on first non-label
        if  controls[i][Type] != Label
        and this = 0 then
            this = i            
        end if
        
        -- draw control
        if defaultButton = i then
            -- highlight default button
            draw_control( controls[i], 1 )
        else                              
            -- no highlight
            draw_control( controls[i], 0 )
        end if
        
        
    end for
    
    -- main loop
    while 1 do

        -- draw current control
        draw_control( controls[this], 1 )    
        
        mouse_events( LEFT_DOWN )

        while 1 do         
        
            -- clear key
            key = -1

            -- activate mouse            
            mouse_pointer(1)
    
            -- look for a mouse
            tmp = mouse_control()
            if sequence(tmp) then      

                -- dim current control
                draw_control( controls[this], 0 )
    
                -- new current control
                this = tmp[1]                      
                
                -- key to give control
                key = tmp[2]       
                
            end if

            -- check for key, if no mouse event
            if key = -1 then
                key = get_key()
            end if

            -- leave loop on key            
            if key != -1 then
                exit
            end if
        end while

        -- hide mouse
        mouse_pointer(0)
        
        -- give to control
        controls[this] = edit_control( controls[this], key )
        draw_control( controls[this], 1 )
        
        
        -- check for navigation
        if key = Enter
        or (key = Space and controls[this][Type] = Button )
        then

            -- button?
            if controls[this][Type] = Button then
                -- help button?
                if compare( controls[this][Title], "Help" ) = 0 then

                    -- save old values
                    oldControls = controls
                    oldRow = wRow
                    oldCol = wCol
                    oldTitle = wTitle
        
                    -- text
                    if length( helpText ) = 0 then
                        text = {"No help is available on " & wTitle & "." }
                    else
                        text = helpText
                    end if

                    -- build the dialog
                    tmp = build_msgbox( OK, text )
                    
                    -- run the dialog
                    tmp =  run_dialog( tmp[1], tmp[2], "Help On " & wTitle )

                    -- restore controls
                    controls = oldControls
                    wRow = oldRow
                    wCol = oldCol
                    wTitle = oldTitle
                    
                    -- colors
                    text_color( TextColor )
                    bk_color( BackColor )
                    
                    
                else
                    -- leave
                    result = { controls[this][Title], {} }
                    exit         
                    
                end if   
                
            else
                -- exit, using default button
                    result = { controls[defaultButton][Title], {} }
                    exit
            end if
            
        elsif key = Escape then

            result = { "Cancel", {} }
            exit
            
        elsif key = Tab
        or    key = Right
        or    key = Down
        then            
        
            if key = Down
            and controls[this][Type] = List then
                -- ignore; handled in edit
            else
                -- move ahead
                draw_control( controls[this], 0 )    

                while 1 do                
                    
                    -- move
                    this = this + 1
                    if this > length( controls ) then
                        this = 1
                    end if
                    
                    -- ok to stop?
                    if controls[this][Type] != Label then
                        exit
                    end if 
                    
                end while 
                
                -- highlight/dim default button
                if controls[this][Type] != Button then      
                    -- light it
                    draw_control( controls[defaultButton], 1 )    
                else
                    -- dim it
                    draw_control( controls[defaultButton], 0 )    
                end if
                                              
            end if
                
        elsif key = ShiftTab
        or    key = Left
        or    key = Up
        then
            -- move back
            
            if Up
            and controls[this][Type] = List then
                -- ignore; handled in edit
            else  
                -- move backwards
                draw_control( controls[this], 0 )    

                -- loop until safe spot found
                while 1 do                
                    this = this - 1
                    if this < 1 then
                        this = length( controls )
                    end if                               
                    
                    -- ok to stop?
                    if controls[this][Type] != Label then
                        exit
                    end if
                end while
            end if
        
            -- highlight/dim default button
            if controls[this][Type] != Button then      
                -- light it
                draw_control( controls[defaultButton], 1 )    
            else
                -- dim it
                draw_control( controls[defaultButton], 0 )    
            end if
            
        else
            -- position( 1, 1 )
            -- ? key        
        end if
    
    end while                     

    -- restore screen             
    display_text_image( {1, 1}, screen )
    
    -- append values to result
    for i = 1 to length( controls ) do
        -- get control data
        tmp = controls[i][Data]
        
        -- if list, get selected item
        if controls[i][Type] = List then
            if length( tmp ) > 0 then
                -- index to selected item
                tmp = tmp[ controls[i][Anchor] + controls[i][Cursor] ]
            end if
        end if

        -- append to list        
        result[2] = append( result[2], tmp )
    end for

    -- restore default colors
    text_color( WHITE )
    bk_color( BLACK )

    return result

end function

-----------------------------------------------------------------------------
global function msgbox( integer style, sequence title, sequence text )

    -- run a messagebox in a given style
    -- return the button that caused an exit
    
    sequence size, result
    
    -- calculate and build
    size = build_msgbox( style, text )
    
    -- run dialog
    result = run_dialog( size[1], size[2], title )
    
    -- return button
    return result[1]
    
end function

