-- QBasic Compatible Routines for Euphoria

without warning
without trace
without type_check

include msgbox.e
include limits.e
include graphics.e
include wildcard.e
include get.e
include ports.e


-- returned value from main function
    global object tmp

-- marks an argument missing
global constant MISSING = -999999

-- variable storage
global constant
    NUM         = 1,    -- numeric
    STR         = 2,    -- string
    NUMS        = 3,    -- numeric arrays
    STRS        = 4,    -- string array
    USER_TYPE   = 5     -- user defined types

global constant
    OFS         = 1,    -- holds array offsets
    VALS        = 2     -- holds array values

global sequence 
    stack,                              -- stores prior locals
    gNum, gNums, gStr, gStrs, gUser,    -- global variables
    vNum, vNums, vStr, vStrs, vUser,    -- local variables
    c                                   -- constants
    
    stack = {}                          -- init stack
    c = repeat( 0, MAX_CONSTS )         -- init constants

-- arrays
sequence
    INIT_NUM_ARRAY,     -- holds default empty numeric array
    INIT_STR_ARRAY,     -- holds default empty string array
    INIT_USER_TYPE      -- holds default user type element
    
    if FAST_ARRAYS then
        -- just data for arrays and types
        INIT_NUM_ARRAY = repeat( 0, 11 )
        INIT_STR_ARRAY = repeat( "", 11 ) 
        INIT_USER_TYPE = {}
    else
        -- data and offsets
        INIT_NUM_ARRAY = { { 1 }, repeat( 0, 11 ) }
        INIT_STR_ARRAY = { { 1 }, repeat( "", 11 ) }
        INIT_USER_TYPE = { {}, {} }
    end if
        
-- setup global variables
    gNum = repeat( 0, MAX_GLOBALS )
    gNums = repeat( INIT_NUM_ARRAY, MAX_GLOBALS )
    
    gStr = repeat( {}, MAX_GLOBALS )
    gStrs = repeat( INIT_STR_ARRAY, MAX_GLOBALS )
    
    gUser = repeat( INIT_USER_TYPE, MAX_TYPES )

-- locals
    vNum = {}
    vNums = {}    
    vStr = {}
    vStrs = {}    
    vUser = {}


-- used for jumps
    global integer goto
    global integer gosub
    global integer returnFlag
    goto = 0
    gosub = 0
    returnFlag = 0

-- data support
    global sequence data
    global integer dataPos
    data = {}
    dataPos = 0

-- for peek and poke
    atom segment
    segment = 0
    
-- for extended keys in Inkey$
-- convert to QB mapping of keys

    sequence extendedKey
    extendedKey = repeat( 0, 400 )
    
    extendedKey[271] = 15       -- shift+tab

    extendedKey[272] = 16       -- alt+q
    extendedKey[273] = 17       -- alt+w
    extendedKey[274] = 18       -- alt+e
    extendedKey[275] = 19       -- alt+r
    extendedKey[276] = 20       -- alt+t
    extendedKey[277] = 21       -- alt+y
    extendedKey[278] = 22       -- alt+u
    extendedKey[279] = 23       -- alt+i
    extendedKey[280] = 24       -- alt+o
    extendedKey[281] = 25       -- alt+p

    extendedKey[282] = 30       -- alt+a
    extendedKey[283] = 31       -- alt+s
    extendedKey[284] = 32       -- alt+d
    extendedKey[285] = 33       -- alt+f
    extendedKey[286] = 34       -- alt+g
    extendedKey[287] = 35       -- alt+h
    extendedKey[288] = 36       -- alt+j
    extendedKey[289] = 37       -- alt+k
    extendedKey[290] = 38       -- alt+l
                                
    extendedKey[300] = 44       -- alt+z
    extendedKey[301] = 45       -- alt+x
    extendedKey[302] = 46       -- alt+c
    extendedKey[303] = 47       -- alt+v
    extendedKey[304] = 48       -- alt+b
    extendedKey[305] = 49       -- alt+n
    extendedKey[306] = 50       -- alt+m
                               
    extendedKey[315] = 59       -- f1
    extendedKey[316] = 60       -- f2
    extendedKey[317] = 61       -- f3
    extendedKey[318] = 62       -- f4
    extendedKey[319] = 63       -- f5
    extendedKey[320] = 64       -- f6
    extendedKey[321] = 65       -- f7
    extendedKey[322] = 66       -- f8
    extendedKey[323] = 67       -- f9
    extendedKey[324] = 68       -- f10
    extendedKey[325] = 69       -- f11
    extendedKey[326] = 70       -- f12

    extendedKey[327] = 71       -- home
    extendedKey[328] = 72       -- up 
    extendedKey[329] = 73       -- page up
    extendedKey[331] = 75       -- left
    extendedKey[333] = 77       -- right
    extendedKey[335] = 79       -- end
    extendedKey[336] = 80       -- down
    extendedKey[337] = 81       -- page down
    extendedKey[338] = 82       -- insert
    extendedKey[339] = 83       -- delete

    extendedKey[376] = 120      -- alt+1
    extendedKey[377] = 121      -- alt+2
    extendedKey[378] = 122      -- alt+3
    extendedKey[379] = 123      -- alt+4
    extendedKey[380] = 124      -- alt+5
    extendedKey[381] = 125      -- alt+6
    extendedKey[382] = 126      -- alt+7
    extendedKey[383] = 127      -- alt+8
    extendedKey[384] = 128      -- alt+9
    extendedKey[385] = 129      -- alt+0
    extendedKey[386] = 130      -- alt+-
    extendedKey[387] = 131      -- alt+=

-----------------------------------------------------------------------------
global function jump_to( integer label )
    -- true if goto equals label
    -- or goto is blank

    integer result


    if returnFlag then
        -- don't jump; a return is pending
        result = 0
    elsif goto = 0 then
        -- set goto so locals will not nest
        goto = label
        result = -1
    else
        -- return true if label matches
        if goto = label then
            result = -1
        else    
            result = 0
        end if
    end if

    return result

end function


-----------------------------------------------------------------------------
global function old_jump_to( integer label )
    -- true if goto equals label
    -- or goto is zero


    if returnFlag then
        -- don't jump; a return is pending
        return 0
    elsif goto = 0 then
        -- set goto so locals will not nest
        goto = label
        return -1
    else
        -- return true if label matches
        return (goto=label)
    end if

end function

-----------------------------------------------------------------------------
global procedure setup_frame()

    -- save old, and initialize new local variables

    -- check stack space
    if length( stack ) >= MAX_STACK then
        abort_error( "Run Time Error", "Out of stack space." )
    end if

    -- only push if you need to
    if length( vNum ) then
        -- push prior locals: MUST BE IN THIS ORDER!
        stack = append( stack, { vNum, vStr, vNums, vStrs, vUser } )
    end if
    
    -- initialize current locals
    vNum = repeat( 0, MAX_LOCALS )
    vNums = repeat( INIT_NUM_ARRAY, MAX_LOCALS )
    
    vStr = repeat( {}, MAX_LOCALS )
    vStrs = repeat( INIT_STR_ARRAY, MAX_LOCALS )    
    
    vUser = repeat( INIT_USER_TYPE, MAX_TYPES )
    
end procedure

-----------------------------------------------------------------------------
procedure assign_to_passback( object v, sequence a, integer tos )

    -- assign value v to the passback address a
    integer n

    -- get length
    n = length( a )
    
    -- address to assign?
    if n < 3 then
        return
    end if
    
    -- resolve references
    if a[1] = 'v' then   
        -- local variable
        if n = 3 then       stack[tos-1][a[2]][a[3]] = v
        elsif n = 4 then    stack[tos-1][a[2]][a[3]][a[4]] = v
        elsif n = 5 then    stack[tos-1][a[2]][a[3]][a[4]][a[5]] = v
        elsif n = 6 then    stack[tos-1][a[2]][a[3]][a[4]][a[5]][a[6]] = v
        elsif n = 7 then    stack[tos-1][a[2]][a[3]][a[4]][a[5]][a[6]][a[7]] = v
        
        else
            abort_error( "Run Time Error",
                         "Passing by reference only supports 4 indexes." )
        end if

    elsif a[1] = 'g' then
        -- global variable
        if n = 3 then       stack[tos][a[2]][a[3]] = v
        elsif n = 4 then    stack[tos][a[2]][a[3]][a[4]] = v
        elsif n = 5 then    stack[tos][a[2]][a[3]][a[4]][a[5]] = v
        elsif n = 6 then    stack[tos][a[2]][a[3]][a[4]][a[5]][a[6]] = v
        elsif n = 7 then    stack[tos][a[2]][a[3]][a[4]][a[5]][a[6]][a[7]] = v
        
        else
            abort_error( "Run Time Error",
                         "Passing by reference only supports 4 indexes." )
        end if

    else
        abort_error( "Run Time Error",
                     "Error in passback list." )
    end if
end procedure

-----------------------------------------------------------------------------
global procedure cleanup( sequence argVals, sequence passBack )

    -- clean up on exit of function or sub
    
    integer tos, passCount

    -- assign passbacks?
    passCount = length( passBack )

    -- are there args?
    if length( passBack ) then

        -- are any not null?
        if compare( repeat( {}, passCount ), passBack ) != 0 then

            -- push globals on the stack
            stack = append( stack, { gNum, gStr, gNums, gStrs, gUser } )    
            
            -- calculate tos
            tos = length( stack )
    
            -- check which variables get values passed back
            for i = 1 to length( passBack ) do
                -- pass back if not null
                if length( passBack[i] ) then
                    -- assign to caller
                    assign_to_passback( argVals[i], passBack[i], tos )
                end if
            end for
    
            -- get the globals off the stack
            gNum = stack[tos][1]
            gStr = stack[tos][2]
            gNums = stack[tos][3]
            gStrs = stack[tos][4]
            gUser = stack[tos][5]
    
            -- pop the stack
            stack = stack[1..tos-1]

        end if
    end if


    -- get the top of stack
    tos = length( stack )

    -- only pop if there's stuff to pop
    if tos then
    
        -- pop the locals off the stack
        vNum = stack[tos][1]
        vStr = stack[tos][2]
        vNums = stack[tos][3]
        vStrs = stack[tos][4]
        vUser = stack[tos][5]
        
        -- pop stack
        stack = stack[1..length(stack)-1]
        
    end if
    
end procedure



-----------------------------------------------------------------------------
-- QBasic compatible functions

-----------------------------------------------------------------------------
-- ABS()
global function qb_abs( sequence s )
    -- return absolute value of number

    if s < 0 then
        return -s
    else
        return s
    end if

end function

-----------------------------------------------------------------------------
-- ASC
global function qb_asc( sequence s )
    -- return ascii value of first char in string
    if length( s ) = 0 then
        return 0
    else
        return s[1]
    end if
end function
-- ATN      arctan( n )

-----------------------------------------------------------------------------
-- CDBL     n/a
-----------------------------------------------------------------------------
-- CHR$     {n}
-----------------------------------------------------------------------------
-- CINT     n/a
-----------------------------------------------------------------------------
-- CLNG     n/a
-----------------------------------------------------------------------------
-- COS      cos( n )
-----------------------------------------------------------------------------
-- CSNG     n/a
-----------------------------------------------------------------------------
-- CSRLIN   
global function qb_csrlin( sequence s )
    -- return line cursor is on    

    s = get_position()
    return s[1]

end function

-----------------------------------------------------------------------------
--  CVD     n/a
-----------------------------------------------------------------------------
-- CVDMBF   unsupported
-----------------------------------------------------------------------------
-- CVI      n/a
-----------------------------------------------------------------------------
-- CVL      n/a
-----------------------------------------------------------------------------
-- CVS      n/a
-----------------------------------------------------------------------------
-- CVSMBF   unsupported
-----------------------------------------------------------------------------
-- DATE$    
global function qb_date()
    -- return current date "mm-dd-yyyy"
    sequence s
    
    s = date()
    return sprintf( "%02d-%02d-%04d",{s[2],s[3],s[1]+1900} )

end function

-- ENVIRON$
global function qb_environ( sequence s )

    -- return value of environment string

    object result

    result = getenv( s[1] )
    if integer(result) then
        return ""
    else
        return result
    end if

end function

-----------------------------------------------------------------------------
-- EOF
-----------------------------------------------------------------------------
-- ERDEV 
-----------------------------------------------------------------------------
-- ERDEV$ 
-----------------------------------------------------------------------------
-- ERL 
-----------------------------------------------------------------------------
-- ERR 
-----------------------------------------------------------------------------

global constant CONST_E = 2.718281828459045

-- EXP
global function exp(object a) -- this should be part of the interpreter...
    return power(CONST_E, a)
end function


-----------------------------------------------------------------------------
-- FILEATTR          
-----------------------------------------------------------------------------
-- FIX            
global function qb_fix( atom n )
    -- discard non-integer portion of number
    if n < 1 then
        return floor(n)+1
    else
        return floor(n)
    end if
end function

-----------------------------------------------------------------------------
-- FRE 
-----------------------------------------------------------------------------
-- FREEFILE 
-----------------------------------------------------------------------------
-- HEX$
-----------------------------------------------------------------------------
-- INKEY$                    
global function qb_inkey()
    -- return current key press, if any
    -- convert extended keys to QBasic equivalents
    integer key
    key = get_key()
    if key = -1 then
        return ""
    elsif   key < 257 then  
        return {key}        -- normal key
    elsif key > 400 then
        return {0}
    else
        -- convert to extended key
        return {0, extendedKey[key]}
    end if
end function

-----------------------------------------------------------------------------
-- INP
global function qb_inp( atom port )
    -- use Jacques Deschenes code

    return Input( port )

end function
-----------------------------------------------------------------------------
-- INPUT$                    
-----------------------------------------------------------------------------
-- INSTR2
global function qb_instr2( sequence s1, sequence s2 )
    -- return position of s2 in s1
    if length(s2) then
        return match( s2, s1 )
    else
        return 0
    end if
end function

-----------------------------------------------------------------------------
global function qb_instr3( atom n, sequence s1, sequence s2 )
    -- return position of s2 in s1
    integer at
    
    if n < 0 then                   
        -- too small
        return 0
    elsif n > length(s1) then       
        -- too large
        return 0
    else            
        -- look
        at = match( s2, s1[n..length(s1)])
        -- match?
        if at > 0 then
            -- add skipped chars
            at = at + n - 1
        end if
    end if
    
    return at
    
end function

-----------------------------------------------------------------------------
-- INT 
-----------------------------------------------------------------------------
-- IOCTL$ 
-----------------------------------------------------------------------------
-- LBOUND  
-----------------------------------------------------------------------------
-- LCASE$                    
-----------------------------------------------------------------------------
-- LEFT$
global function qb_left( sequence s, atom n )
    -- return charaters in string, starting from left pos.
    if length(s) = 0 then               return ""
    elsif n < 1 then                    return ""
    elsif n > length( s ) then          return s
    else                                return s[1..n]
    end if
end function
-----------------------------------------------------------------------------
-- LEN             
-----------------------------------------------------------------------------
-- LOC                
-----------------------------------------------------------------------------
-- LOCATE
global procedure qb_locate( atom row, atom col, atom visible )
    -- position cursor
    -- also set file position

    integer r, c
    sequence pos

    -- visible?
    if visible = 0 then
        cursor( NO_CURSOR )
    else
        cursor( UNDERLINE_CURSOR )        
    end if

    -- get current position
    pos = get_position()

    -- set row
    if row != -1 then
        r = row
    else
        r = pos[1]
    end if

    -- set col
    if col != -1 then
        c = col
    else
        c = pos[2]
    end if

    -- move to position
    position( r, c )

    -- save
    file_set_pos( 0, col )
    
end procedure

-----------------------------------------------------------------------------
-- LOF 
-----------------------------------------------------------------------------
-- LOG 
-----------------------------------------------------------------------------
-- LPOS 
-----------------------------------------------------------------------------
-- LTRIM$
global function qb_ltrim( sequence s )
    -- remove trailing white space (spaces and tabs)
    for i = length( s ) to 1 by -1 do
        if  s[i] != 9             -- tab
        and s[i] != ' ' then      -- space
            return s[1..i]
        end if
    end for
    -- nothing but spaces
    return ""
end function

-----------------------------------------------------------------------------
-- MID$
global function qb_mid( sequence s, atom start, atom size )
    -- return substring from string

    integer n

    -- third arg missing?
    if size = MISSING then
        -- calculate substring size
            size = length( s ) - start + 1
            if size < 0 then
                size = 0
            end if
    end if    

    -- too small?
    if start < 0 then
        abort_error( "Run Time Error", "Illegal function call in MID$." )
    end if

    -- legal, but null
    if size = 0
    or start > length( s ) then
        return ""
    end if

    -- check against substring size
    n = start + size - 1
    if n < 1 then
        return ""
    elsif n > length( s ) then
        n = length( s )
    end if

    -- return substring
    return s[start..n]

end function

-----------------------------------------------------------------------------
global function qb_assign_mid2( sequence s1, integer start, sequence s2 )

    -- mid$( s1, start ) = s2
    integer n
    
    -- legal values?
    if start > length( s1 )
    or start < 1 then
        abort_error( "Run Time Error", "Illegal function call in MID$." )
    end if

    -- if smaller, paste prior
    s2 = s1[1..start-1] & s2
    if length( s2 ) < length( s1 ) then
        s2 = s2 & s1[length(s2)+1..length(s1)]
    end if
    
    -- trim
    return s2[1..length(s1)]
    
    
    
    
end function


-----------------------------------------------------------------------------
global function qb_assign_mid3( sequence s1, integer start, integer long,
                                    sequence s2 )
    -- alter s2 based on long
    if length( s2 ) > long then
        s2 = s2[1..long]
    end if

    -- call mid2    
    return qb_assign_mid2( s1, start, s2 )

end function

-----------------------------------------------------------------------------
-- MKD$                      
-----------------------------------------------------------------------------
-- MKDMBF$                   
-----------------------------------------------------------------------------
-- MKI$ 
-----------------------------------------------------------------------------
-- MKL$ 
-----------------------------------------------------------------------------
-- MKS$ 
-----------------------------------------------------------------------------
-- MKSMBF$ 
-----------------------------------------------------------------------------
-- OCT$
-----------------------------------------------------------------------------
-- PEEK                      
global function qb_peek( atom addr )
    -- this will need to use the segment as well
    return peek( segment + addr )
end function    
    
-----------------------------------------------------------------------------
-- PEN                       
-----------------------------------------------------------------------------
-- PLAY
    -- shared buffer
        sequence music      -- music holding music string
        integer musicPos    -- current position in string

    constant          
        C = 1, D = 3, E = 5, F = 6, G = 8, A = 10, B = 12,
        PAUSE = 0,
                
        FREQ = {{   0,   0,   0,  39,  41,  44,  46,  49,  52,  55,  58,  62},
                {  65,  69,  73,  78,  82,  87,  93,  98, 104, 110, 117, 123},
                { 131, 139, 147, 156, 165, 175, 185, 196, 208, 220, 233, 247},
                { 262, 277, 294, 311, 330, 349, 370, 392, 415, 440, 466, 494},
                { 523, 554, 587, 622, 659, 698, 740, 784, 831, 880, 932, 988},
                {1047,1109,1175,1245,1329,1397,1480,1568,1661,1760,1865,1975},
                {2093,2217,2349,2489,2637,2794,2960,3136,3322,3520,3729,3951},
                {4186,   0,0,0,0,0,0,0,0,0,0,0},
                {0,0,0,0,0,0,0,0,0,0,0,0}},

        NORMAL      = 1,
        LEGATO      = 2,
        STACCATO    = 3

------------------------------------------------------------------------------
function read_int_from_music()

    -- attempt to read the next part of the music buffer as a number
    -- increment music position
    -- returns -1 if conversion failed

    integer result

    -- room for a number?
    if musicPos >= length( music ) then
        return -1
    end if  
    
    -- digit in position?
    if not find( music[musicPos+1], "0123456789" ) then
        return -1
    end if

    -- read digits
    result = 0
    while 1 do           
        -- at end of string?
        if musicPos = length( music ) then
            return result
        end if

        -- current char a digit?        
        if not find( music[musicPos+1], "0123456789" ) then        
            return result
        end if
        
        -- add digit to result
        result = ( result * 10 ) + ( music[musicPos+1] - '0' )
        
        -- move ahead
        musicPos = musicPos + 1

    end while
            
end function         

------------------------------------------------------------------------------
global procedure qb_play(sequence s)

    -- plays QBASIC music command strings; supports all QBASIC music notation
    -- except MF(foreground play) and MB(background play) - this function 
    -- will always play in the foreground
    -- Background play on the pc speaker involves machine code and hooking 
    -- interrupts, which is out of my range for the moment.
    -- The only other difference is that this function will stop on an
    -- ESCAPE key press.

    atom noteLen, tmpLen
    integer note, oct, char, noteSize, tempo, tmp, noteIndex, style, tmpOct
    sequence list

    -- move s into a shared music
    music = s

    style = NORMAL -- normal
    tempo = 32
    noteSize = 4
    noteLen = .1
    oct = 2
    musicPos = 1
    list = {}

    while musicPos <= length( music ) do

        -- parse instruction
        char = lower(music[musicPos])        

        -- ' ' blank
        if char = ' ' then
            -- ignore

        -- 'o' select octave
        elsif char = 'o' then            
            tmp = read_int_from_music()
            if tmp >= 0 and tmp <= 6 then
                oct = tmp + 2  
            end if

        -- '>' up 1 octave
        elsif char = '>' then
            if oct < 8 then
                oct = oct + 1
            end if

        -- '<' down 1 octave                    
        elsif char = '<' then
         if oct > 2 then
            oct = oct - 1
         end if

        -- 't' tempo         
        elsif char = 't' then
            tmp = read_int_from_music()
            if tmp >= 32 and tmp <= 255 then
                tempo = tmp
            end if
        
        -- 'l' length    
        elsif char = 'l' then
            tmp = read_int_from_music()
            if tmp > 0 then
                noteSize = tmp
            end if
            
        -- 'm' mode
        elsif char = 'm' then
        
            if musicPos < length( music ) then
                musicPos = musicPos + 1
                
                -- modes
                if      lower(music[musicPos]) = 'l' then style = LEGATO
                elsif   lower(music[musicPos]) = 'n' then style = NORMAL
                elsif   lower(music[musicPos]) = 's' then style = STACCATO
                else  -- this function doesn't 'F' or 'B'
                end if

                -- set note style
                if    style = LEGATO then
                    noteLen = (240/tempo)/noteSize

                elsif style = NORMAL then
                    noteLen = ((240/tempo)/noteSize)*(.9)

                elsif style = STACCATO then
                    noteLen = ((240/tempo)/noteSize)*(.5)

                end if

            end if

        else            
            -- 'p' pitch
            if char = 'p' then
                tmp = read_int_from_music()
                if tmp > 0 then
                    noteSize = tmp
                    list = append( list, { PAUSE,(240/tempo)/noteSize } )
                end if

            -- 'a'
            elsif char = 'a' then
                noteIndex = A
    
            -- 'b'            
            elsif char = 'b' then
                noteIndex = B
                
            -- 'c'             
            elsif char = 'c' then
                noteIndex = C
    
            -- 'd'         
            elsif char = 'd' then
                noteIndex = D
            
            -- 'e'    
            elsif char = 'e' then
                noteIndex = E
    
            -- 'f'            
            elsif char = 'f' then
                noteIndex = F
                
            -- 'g'
            elsif char = 'g' then
                noteIndex = G
                
            -- 'n'
            elsif char = 'n' then
                tmp = read_int_from_music()
                if tmp > 0 then
                    if remainder(tmp,12) != 0 then
                        noteIndex = remainder(tmp,12)
                        tmpOct = floor(tmp/12) + 2
                    else
                        noteIndex = (tmp/12)
                        tmpOct = floor(tmp/12) + 1
                    end if
                      
                    noteSize = tmp

                else
                
                    noteIndex = 0
                                        
                end if
            
            else                
                -- unknown
                abort_error( "Play Command", 
                                "Illegal Function Call: '" & char & "'" )
                
            end if
            
            -- default pitch            
            note = FREQ[oct][noteIndex]
                        
            if not noteIndex then
                note = PAUSE
                
            elsif char = 'n' then
                if noteSize > 0 then 
                    note = FREQ[tmpOct][noteIndex]
                end if

            -- accidental?                
            elsif musicPos < length( music ) then
             
                -- '-' flat
                if music[musicPos+1] = '-' then
                    if char = 'c' then
                        note = FREQ[oct-1][B]            
                    else     
                        note = FREQ[oct][noteIndex-1]
                   end if
                   musicPos = musicPos + 1

                -- '#' or '+' sharp                   
                elsif find(music[musicPos+1],"#+") then
                    if char = 'b' then
                        note = FREQ[oct+1][C]            
                    else
                        note = FREQ[oct][noteIndex+1]
                    end if
                    musicPos = musicPos + 1

                end if
                
             end if

            -- dotted?
            tmpLen = noteLen
            while musicPos < length( music ) do
                if music[musicPos + 1] != '.' then
                    exit
                end if
                
                -- increase duration
                tmpLen = tmpLen * 1.5
                musicPos = musicPos + 1
                
            end while

            -- play the sound
            list = append( list, {note,tmpLen} )

            if style = LEGATO then -- legato
                -- no pause
             
            else -- normal OR staccato
                list = append( list,  {PAUSE, noteLen/9} )
             
            end if
            
        end if                
      
        -- check for an escape key
        if get_key() = 27 then -- escape key pressed
            exit
        end if

        -- move position
        musicPos = musicPos + 1
        
    end while

    -- play the music
    tick_rate(500)
    for i = 1 to length( list ) do
        -- play the note
        sound( list[i][1] )
        -- pause
        tmpLen = time()
        while (time()-tmpLen) < list[i][2] do 
        end while    
    end for
    tick_rate(0)

    -- cancel any sound   
    sound(0)
    
end procedure

-----------------------------------------------------------------------------
-- PMAP
-----------------------------------------------------------------------------
-- POKE
global procedure qb_poke( atom address, atom byte )
    -- should use segment
    poke( segment + address, byte )
end procedure

-----------------------------------------------------------------------------
-- POINT 
-----------------------------------------------------------------------------
-- POS
global function qb_pos( object n )
    -- return column cursor is on
    
    sequence s
    
    s = get_position()
    return s[2]

end function

-----------------------------------------------------------------------------
-- RIGHT$
global function qb_right( sequence s, atom n )
    -- return chars in string, starting from left pos.
    integer l

    l = length( s )
    
    if l = 0    then        return ""
    elsif n < 1 then        return ""
    elsif n > l then        return s
    else                    return s[l-(n-1)..l]
    end if
end function

-----------------------------------------------------------------------------
-- RND
global function qb_rnd()
    -- return random number between 0 and 1
    return rand(999999)/1000000
end function

-----------------------------------------------------------------------------
-- RTRIM$
global function qb_rtrim( sequence s )
    -- remove trailing white space (spaces and tabs)
    for i = 1 to length( s ) do
        if  s[i] != 9             -- tab
        and s[i] != ' ' then      -- space
            return s[i..length(s)]
        end if
    end for
    -- nothing but spaces
    return ""
end function

-----------------------------------------------------------------------------
-- SCREEN
global function qb_fn_screen( atom row, atom col, atom colorFlag )
    -- returns char/attribute at row/col
    -- colorFlag indicates which

    sequence s

    -- get char and at tribute
    s = save_text_image( {row, col}, {row, col} )

    if colorFlag = MISSING
    or colorFlag = 0 then
        return s[1][1]
    else
        return s[1][2]        
    end if
    
end function

-----------------------------------------------------------------------------
global procedure qb_sub_screen( atom mode, atom colorSwitch,
                                    atom activePage, atom visualPage )

    if colorSwitch = 0 then
        if mode = 0 then
            -- disable color
        else
            -- enable color
        end if
    elsif colorSwitch = 1 then
        if mode = 1 then
            -- enable color
        else
            -- disable color
        end if
    end if

    if activePage != MISSING then
        -- set screen page
    end if

    if activePage != MISSING then
        -- set active page
    end if

end procedure

-----------------------------------------------------------------------------
-- SEEK                      
-----------------------------------------------------------------------------
-- SGN                       
-----------------------------------------------------------------------------
-- SIN
-----------------------------------------------------------------------------
-- SLEEP
global procedure qb_sleep(atom n)
    -- an n second pause while a message is on the screen
    -- out of order because it's used by 'beep'
    atom t

    t = time()
    while time() < t + n do
    end while
end procedure

-----------------------------------------------------------------------------
-- SPACE$
-----------------------------------------------------------------------------
-- SPC                       
-----------------------------------------------------------------------------
-- SQR 
-----------------------------------------------------------------------------
-- STICK 
-----------------------------------------------------------------------------
-- STR$ 
-----------------------------------------------------------------------------
-- STRIG

-----------------------------------------------------------------------------
-- STRING$
global function qb_string( integer wide, object pattern )
    -- return string of given length from pattern
    -- string$( length, ascii code | string )

    sequence result

    if integer( pattern ) then
        return repeat( pattern, wide )
    else
        result = {}
        for i = 1 to wide do
            result = result & pattern
        end for
        return result
    end if

end function

-----------------------------------------------------------------------------
-- TAB      file_tab()

-----------------------------------------------------------------------------
-- TIME$
global function qb_time()
    -- return time in "hh:mm:ss" format
    sequence s
    
    -- get date
    s = date()

    -- 12 hour clock
    if s[4] > 12 then
        s[4] = s[4] - 12
    end if

    -- time
    return sprintf( "%02d:%02d:%02d", s[4..6] )

end function

-----------------------------------------------------------------------------
-- TIMER
-----------------------------------------------------------------------------
-- UBOUND                    
-----------------------------------------------------------------------------
-- UCASE$

-----------------------------------------------------------------------------
-- VAL
global function qb_val( sequence s )
    -- return value of string

    s = value( s )
    if s[1] = GET_SUCCESS then
        return s[2]
    else
        return 0
    end if
end function

-----------------------------------------------------------------------------
-- VARPTR                     
-----------------------------------------------------------------------------
-- VARPTR$                    
-----------------------------------------------------------------------------
-- VARSEG


-- SUPPORT WORDS
global function qb_pad( sequence s, integer n )
    -- limits/pads a string to given size
    -- used for string types
    
    if length( s ) > n then
        return s[1..n]
    else
        return s & repeat( ' ', n - length(s) )
    end if
end function

-----------------------------------------------------------------------------
-- DATA
global function qb_read()
    -- return next element in data
    if dataPos < length( data )-1 then
        dataPos = dataPos + 1
        return data[ dataPos ]
    else
        abort_error( "Run Time Error", "Out of DATA" )  
    end if
end function





-----------------------------------------------------------------------------
-- QBasic compatible statements

-- BEEP
global procedure qb_beep()
    -- beep briefly

    sound( 1000 )
    qb_sleep( 0.07 )
    sound( 0 )

end procedure

-----------------------------------------------------------------------------
-- DEF SEG
global procedure qb_def_seg( atom n )
    -- set up segment

    segment = n * #10

end procedure


-----------------------------------------------------------------------------
-- OUT
global procedure qb_out( atom port, atom b )
    -- use Jacques Deschenes code

    Output( b, port)

end procedure
-----------------------------------------------------------------------------
-- SOUND
global procedure qb_sound( atom pitch, atom duration )
    -- make sound for duration

    sound( pitch )
    qb_sleep( duration )
    sound( 0 )

end procedure


-----------------------------------------------------------------------------
-- WIDTH
global procedure qb_width( atom cols, atom rows )
    -- set screen size
    -- only rows is supported
    
    object result

    -- assume graphics mode 3
    result = graphics_mode(3)    
    
    -- cols is ignored for now
    
    
    if rows != MISSING then
        result = text_rows( rows )
    end if

end procedure

-----------------------------------------------------------------------------
-- QuickBasic compatible statements and functions
-- COMMAND$
global function qb_command()
    -- return command line
    sequence s, cmd
    s = command_line()
    if length( s ) > 2 then
        cmd = ""
        for i = 3 to length( s ) do
            cmd = cmd & s[i] & " "
        end for
        return cmd
    else
        return ""
    end if
end function


