-- QBasic to Euphoria Translator
-- (c) 1999 David Cuny
-- low level routines (c) Jacques Deschenes
-- pc speaker tools (c) J. Craig Gilbert


-- BUGS:
-- Euphoria benchmark: An array is passed as a parameter to a Sub, and
-- then Dimensioned. It's being Dim'ed wrong
-- play() dummied in
-- rnd(1) dummied in
-- AS ANY not supported

include msgbox.e
include wildcard.e
include graphics.e
include get.e
include dialog.e

-- text mode
    set_text_mode()

-- version
constant VERSION = "v.0.60 Bleeding Edge Alpha"

-- for tracing
global integer traceFile
    traceFile = open( "trace", "w" )

without warning

-- modes of optimization
constant
    FAST        = 0,
    COMPATIBLE  = 1,
    TRUE        = -1,
    FALSE       = 0

-- options for optimization    
global integer
    ForLoopMode,
    PrintMode,
    FileReadMode,
    ArrayMode,
    PassByReference
    
global sequence errText

-- compile-time variable information
-- placed here, because it's shared with GRAMMAR.E

global sequence
    locals,             -- local variables is current process
    globals,            -- all global variables
    constants,          -- all constants
    types,              -- user defined data types
    elements            -- elements of user defined types

-- clear the global and local variables
-- clear the constants
    globals = repeat( {}, 5 )
    locals = repeat( {}, 5 )
    constants = {}

global integer
    useLocals,          -- true if not in main process
    sharedFlag,         -- toggled on by SHARED
    staticFlag          -- toggled on by STATIC

    useLocals   = 0     -- false, until in sub/function
    sharedFlag  = 0     -- if true, variable built by var_proto is common
    staticFlag  = 0     -- false, until set by static

global constant
                    -- in the sequences locals and globals
    NUM         = 1,        --  numeric variables
    STR         = 2,        -- string variables
    NUMS        = 3,        -- numeric array variables
    STRS        = 4,        -- string array variables
    USER_TYPE   = 5,        -- user defined; can be array        
    NUM_TYPES   = 5,        -- number of different types

                    -- in the sequences locals[] & globals[]:
    V_NAME      = 1,        -- variable name
    V_TYPE      = 2,        -- variable type
    V_SCOPE     = 3,        -- is global is visible in current scope
    V_INDEXES   = 4,        -- indexes

                        -- within each V_INDEX in an array variable
    V_LOW       = 1,        -- start value
    V_HIGH      = 2,        -- end value

                        -- values for V_VISIBLE
    V_HIDDEN    = 0,        -- only visible from main module
    V_LOCAL     = 1,        -- local within current module
    V_SHARED    = 2,        -- visible to all modules
    V_STATIC    = 3,        -- local, but created in globals

-- types
    T_NAME      = 1,        -- type name
    T_DEFAULT   = 2,        -- default data structure

-- elements
    E_NAME      = 1,        -- element name
    E_TYPE      = 2,        -- element type
    E_INDEX     = 3,        -- position in sequence
    E_OWNER     = 4,        -- element's owner's index
    E_SIZE      = 5         -- element's size; used in strings


-- data types
-- positive values are indexes to user-defined types

global constant
    INTEGER     = -1,
    LONG        = -2,
    SINGLE      = -3,
    DOUBLE      = -4,
    STRING      = -5

global integer
    dataType,               -- current data type
    dataSize                -- used to limit string sizes
    
    dataType    = SINGLE    -- default data type
    dataSize    = 0         -- no size limit

global constant
    SUFFIXES = "!#$%&"

global integer
    hideElements            -- if true, returns hides element names
    hideElements = -1       -- initially true

include limits.e
include grammar.e

-- for line breaks
constant LINE_MAX = 120
constant BREAK_CHAR = " +-*/[]"

integer     deferFlag,          -- if true, code is deferred to uses code
            labelFlag           -- if true, current proc has labels
sequence    uses                -- holds subs/functions
constant    U_NAME  = 1,        -- index to name of sub/function
            U_LIST  = 2,        -- index to list of called subs & functions
            U_CODE  = 3         -- holds actual code of sub/function

sequence subName, subType
sequence typeName, typeDefault
integer  subDataType
integer elementIndex
sequence expr
sequence stack, eStack, typeStack
object tmp, tmp1, tmp2, tmp3
integer at
integer tmpCount
sequence passBackList
sequence loopIndex
sequence gotoLabels, procLabels
integer labelSubCount, labelGotoCount


-----------------------------------------------------------------------------
    


-- pretty printing the output
    integer indents
    integer commentCount

    indents = 0          -- depth of indent
    commentCount = 0    -- number of comments in a row

integer matchedActionToken
sequence act

-- data holder
sequence data
    data = {}


-----------------------------------------------------------------------------
-- File I/O    

    tmp = command_line()
    if length( tmp ) != 3 then        
        abort_error( "Error", "Expected a file name." )
    end if

    if not match( ".", tmp[3] ) then
        tmp[3] = tmp[3] & ".BAS"
    else
        if not match( ".bas", tmp[3] ) then        
            abort_error( "Error", "File name must end with '.BAS'." )
        end if
    end if

    constant inFile = upper( tmp[3] )
    constant outFile = open( tmp[3][1..length(tmp[3])-4] & ".ex" , "w" )


-----------------------------------------------------------------------------

-- blank the screen
    bk_color( BLUE )
    clear_screen()

-- menu bar         
    text_color( BLACK )
    bk_color( WHITE )
    puts( 1, "   File  Edit  View  Search  Run  Debug  Options                           Help " )
    
    -- top of window
    text_color( WHITE )
    bk_color( BLUE )    
    puts( 1, 218 & repeat( 196, 78 ) & 191 )

    -- center text
    text_color( BLUE )
    bk_color( WHITE )
    position( 2, 40 - floor( length( inFile ) / 2 ) )
    puts( 1, " " & inFile & " " )

    -- return colors
    text_color( WHITE )
    bk_color( BLUE )
    
    position( 3, 1 )
    
    -- window
    for i = 1 to 21 do
    
        -- screen
        puts( 1, 179 & repeat( ' ', 78 ) & 179 )
        
    end for          
    
    -- bottom of window
    puts( 1, 192 & repeat( 196, 78 ) & 217 )
    
    -- status line: don't wrap
    wrap(0)
    position( 25, 1 )
    text_color( BRIGHT_WHITE )
    bk_color( CYAN )    
    puts( 1, repeat( ' ', 80 ) )
    wrap(1)
    
    -- message
    position( 25, 2 )
    puts( 1, "eBasic " & VERSION )


-----------------------------------------------------------------------------
-- splash screen

-- clear controls
    controls = {}

-- create controls
    add_control( Label, "Welcome To eBasic!", 2, 12, 1, 0, 0 )
    add_control( Label, VERSION, 3, 7, 1, 0, 0 )        
    add_control( Label, "(c) 1997 David Cuny", 5, 12, 1, 0, 1 )
    add_control( Label, "Hardware Calls (c) Jacques Deschenes", 7, 3, 1, 0, 1 )
    add_control( Label, "Speaker Music (c) J. Craig Gilbert", 8, 4, 1, 0, 1 )

-- create buttons
    add_button( "OK", 6 )

-- run dialog
    tmp = run_dialog( 40, 12, "" )

-----------------------------------------------------------------------------
-- Prompt user for options

-- clear controls
    controls = {}

-- create controls
    add_control( Check, "Compatible Loops", 2, 8, 1, 0, 1 )
    add_control( Check, "Compatible Printing", 3, 8, 1, 0, 1 )
    add_control( Check, "File Reading", 4, 8, 1, 0, 1 )    
    add_control( Check, "Compatible Arrays", 5, 8, 1, 0, 1 )
    add_control( Check, "Pass by Reference", 6, 8, 1, 0, 1 )

-- create buttons
    add_button( "OK", 6 )
    add_button( "Cancel", 8 )

-- run dialog
    tmp = run_dialog( 40, 10, "eBasic Compatibility Options" )

-- cancelled?
    if compare( tmp[1], "Cancel" ) = 0 then
        abort_error( "Alert", "eBasic cancelled." )
    end if

-- set values
    ForLoopMode     = tmp[2][1]
    PrintMode       = tmp[2][2]
    FileReadMode    = tmp[2][3]
    ArrayMode       = tmp[2][4]
    PassByReference = tmp[2][5]





-----------------------------------------------------------------------------
procedure on_screen( sequence s1, sequence s2 )

    -- displays a line on the screen
    
    sequence s

    -- scroll the display
    s = save_text_image( {4,2}, {23,79} )
    display_text_image( {3,2}, s )

    -- set colors
    bk_color( BLUE )
    text_color( WHITE )


    -- clear bottom line      
    position( 23, 2 )
    puts( 1, repeat( ' ', 78 ) )

    -- add text
    position( 23, 2 )
    if length( s2 ) then
        -- use mask
        printf( 1, s1, s2 )
    else
        -- just print string
        puts( 1, s1 )
    end if

end procedure



-----------------------------------------------------------------------------
function val( sequence s )
    -- convert s to a number
    sequence result
    
    result = value( s )
    if result[1] = GET_SUCCESS then
        return result[2]
    else
        return 0
    end if

end function


---------------------------------------------------------------------------
procedure indent( integer n )

    -- alter the indentation
    indents = indents + n

end procedure


---------------------------------------------------------------------------
procedure to_buffer( sequence s )

    integer i

    -- point to last item    
    i = length( uses )

    -- append code   
    if deferFlag then    
        -- place in code section
        uses[i][U_CODE] = append( uses[i][U_CODE], s )
    else
        -- write to file
        puts( outFile, s )
    end if

end procedure

---------------------------------------------------------------------------
procedure break_line( sequence s, integer breakFlag )

    integer quoteFlag, breakAt
    sequence indent            

    -- calculate intentation
    if indents > 0 then
        indent = repeat( ' ', (indents+breakFlag)*2 )
    else
        indent = ""
    end if                    
    
    -- comments break if they are too long
    if length(s) > 2 then
        if compare( s[1..2], "--" ) = 0 then
            -- too long?
            if length( s ) > LINE_MAX then
            
                -- put in appropriate buffer
                to_buffer( indent & "-- REM" & s[3..LINE_MAX] & "\n" )

                -- break line
                break_line( "-- REM ... " & s[LINE_MAX+1..length(s)], breakFlag )

            else
                -- put in appropriate buffer
                to_buffer( indent & s )
            end if
            -- exit.
            return
        end if
    end if 
    
    -- no quote encountered
    quoteFlag = 0

    -- break the line, if needed
    if length( s ) < LINE_MAX then
        to_buffer( indent & s )
    else
        -- find a place to break the line
        breakAt = 0
        for i = 2 to LINE_MAX do
            if s[i] = '\"' then
                quoteFlag = not quoteFlag
                
            elsif quoteFlag then
                -- can't break in a string
                
            elsif find( s[i], BREAK_CHAR ) then
                breakAt = i
                
            end if
        end for

        if quoteFlag then
            -- max reached without an end quote     

            -- break with quote...  
            to_buffer( indent & s[1..LINE_MAX] & "\"\n" )
            
            -- ...recurse with remainder, adding append and quote
            break_line( "& \"" & s[LINE_MAX+1..length(s)], 1 )                
        
        elsif breakAt = 0 then
            abort_error( "Error", "Line too long - unable to find break point." )
        
        else
            -- break line
            to_buffer( indent & s[1..breakAt] )
            
            -- line feed?
            if s[breakAt] != '\n' then             
                to_buffer( "\n" )
            end if
            
            -- recurse with remainder
            break_line( s[breakAt+1..length(s)], 1 )                
                                
        end if

    end if
    
end procedure

---------------------------------------------------------------------------
procedure out( sequence mask, sequence vals )

    -- write to output file

    integer defer
    sequence s

    -- check indent level
    if indents < 0 then
        errText = "Nesting Error - too many END clauses."
        return
    end if

    -- resolve string
    if length( vals ) = 0 then
        -- just the mask
        s = mask
    else
        -- merge
        s = sprintf( mask, vals )
    end if

    -- output the line
    break_line( s, 0 )

    
end procedure


---------------------------------------------------------------------------
procedure add_qb_as_comment( sequence s )

    -- add qb code as comment, if not all white space

    -- trim leading white space
    -- note: the line feed is already on the sequence

    integer oldIndent
    sequence s2

    s2 = s   
    for i = 1 to length( s ) do        
        if  s[i] != ' ' and s[i] != '\t' then
            s2 = s[i..length(s)]
            exit
        end if
    end for

    if length( s2 ) then 
        -- preceed with blank line?
        if s2[1] = '\'' then  
        
            -- comment
            if commentCount = 0 then
                -- line feed
                out( "\n", {} )         
            end if
                
            -- comment is not indented
            oldIndent = indents            
            indents = 0
            out( "-- " & s2[2..length(s2)], {} )
            indents = oldIndent
            
            -- increment count
            commentCount = commentCount + 1

        else
            -- clear count
            commentCount = 0
                                                                                
            -- preceed by blank line
            out( "\n", {} )

            -- print if more than cr/lf
            if length( s2 ) > 2 then
                out( "-- " & s2, {} )
            end if
            
        end if
    end if
    return
    
    
end procedure


---------------------------------------------------------------------------
function storage_text( integer index )

    if    index = NUM           then return "Num"
    elsif index = STR           then return "Str"
    elsif index = NUMS          then return "Nums"
    elsif index = STRS          then return "Strs"
    elsif index = USER_TYPE     then return "User"
    end if

    abort_error( "Error", sprintf( "Unknown data type index %d.", {index} ) )

end function

---------------------------------------------------------------------------
procedure guess_data_type( sequence s )

    -- set data type based on the suffix
    integer lastChar
    
    lastChar = last( s )
    if    lastChar = '%' then   dataType = INTEGER    
    elsif lastChar = '!' then   dataType = SINGLE
    elsif lastChar = '&' then   dataType = LONG
    elsif lastChar = '#' then   dataType = DOUBLE
    elsif lastChar = '$' then   dataType = STRING
                                dataSize = 0
                                                                
    else                        dataType = SINGLE
    end if
    
end procedure
---------------------------------------------------------------------------
function get_scope()

    -- return the current scope, based on:
    --
    --  FLAG            MEANING
    --  ------------    -----------------------------
    --  sharedFlag      SHARED option is selected
    --  staticFlag      procedure had STATIC option selected
    --  useLocals       in subprogram, not main.


    if sharedFlag then
        if useLocals then
            -- shared in procedure
            return V_LOCAL
        else
            -- shared in global
            return V_SHARED
        end if
    elsif staticFlag then
        -- known to current procedure, but built in globals
        return V_STATIC
    else
        -- known only to current procedure
        return V_HIDDEN
    end if
         

end function

---------------------------------------------------------------------------
function var_proto( sequence name )

    -- return a prototype for a variable
    return { name, dataType, get_scope(), {} }
    
end function    

---------------------------------------------------------------------------
function v_ref( sequence name, sequence indexes, sequence elements )

    -- returns reference, sets data type
    -- will create a variable, if needed

    integer at, scope, storageIndex, arrayFlag, elementFlag
    sequence ref, search

    -- set flags
    arrayFlag = length( indexes )
    elementFlag = length( elements )

    -- limit search base on index/elements
    if elementFlag  then search = { USER_TYPE }
    elsif arrayFlag then search = { NUMS, STRS, USER_TYPE }
                    else search = { NUM, STR, USER_TYPE }
    end if

    -- clear match flag
    at = 0

    -- case 0: loop index
    if  ForLoopMode = FAST 
    and at = 0 then

        -- look in loop indexes
        at = a_find( name, loopIndex )
        if at then
            
                -- data type                    
                dataType = SINGLE

                -- reference "real" index
                return loopIndex[at][2]
        
        end if
    
    end if

    -- case 1: local?
    if  at = 0                  -- no prior match
    and useLocals               -- locals are in use
    and not staticFlag then     -- static is off
    
        for i = 1 to length( search ) do        -- for each valid index
            
            -- get index
            storageIndex = search[i]

            at = a_find( name, locals[storageIndex] )

            -- found?
            if at then

                -- data type                    
                dataType = locals[storageIndex][at][V_TYPE]

                -- reference
                return sprintf( "v%s[%d]", {storage_text(storageIndex), at})
                
                -- leave loop
                exit
            end if
        end for
    end if

    -- case 2: local static?
    if  at = 0              -- no prior match
    and useLocals           -- locals are in use
    and staticFlag then     -- locals are static
        for i = 1 to length( search ) do        -- for each valid index
            
            -- get index
            storageIndex = search[i]

            -- search for non-static name
            for j = 1 to length( globals[storageIndex] ) do
                if  compare( name, globals[storageIndex][j][V_NAME] ) = 0
                and globals[storageIndex][j][V_SCOPE] = V_STATIC then
                
                    -- set flag
                    at = j     
                    
                    -- leave loop
                    exit         
                        
                end if
            end for

            -- found?
            if at then

                -- data type                    
                dataType = globals[storageIndex][at][V_TYPE]

                -- reference
                return sprintf( "g%s[%d]", {storage_text(storageIndex), at})
            
                -- leave loop
                exit

            end if
        end for
    end if


    -- case 3: global mode is off, but is visible global
    if  at = 0 then      -- no prior match    
        for i = 1 to length( search ) do        -- for each valid index
            
            -- get index
            storageIndex = search[i]            

            -- search for non-static name
            for j = 1 to length( globals[storageIndex] ) do
                if  compare( name, globals[storageIndex][j][V_NAME] ) = 0
                and globals[storageIndex][j][V_SCOPE] > 0
                and globals[storageIndex][j][V_SCOPE] != V_STATIC then
                
                    -- set flag
                    at = j     
                    
                    -- leave loop
                    exit
                             
                end if
            end for
            
            -- found?
            if at then

                -- data type
                dataType = globals[storageIndex][at][V_TYPE]
                   
                -- reference
                return sprintf( "g%s[%d]", {storage_text(storageIndex), at})

                -- leave loop
                exit
                
            end if
        end for
    end if

    -- case 4: global mode or shared is on
    if  at = 0                                  -- no prior match
    and not useLocals then                      -- globals visible
    
        for i = 1 to length( search ) do        -- for each valid index
            
            -- get index
            storageIndex = search[i]
        
            at = a_find( name, globals[storageIndex] )  -- look for it
            if at then                                  -- matched?

                -- data type
                dataType = globals[storageIndex][at][V_TYPE] 
                       
                -- reference
                return sprintf( "g%s[%d]", {storage_text(storageIndex), at})
                                
                -- leave loop
                exit

            end if
        end for
    end if

    
    -- case 5: constant
    if at = 0 then
        if  arrayFlag = 0                       -- can't index a constant
        and elementFlag = 0 then                -- can't have element
        
            at = a_find( name, constants )          -- check constants    
            if at > 0 then                          -- found?
    
                -- data type
                dataType = constants[at][V_TYPE]    -- data type        
                
                -- reference
                return sprintf( "c[%d]", at )
         
            end if
        end if    
    end if    

    -- not found?
    if at = 0 then

        -- build a reference: { name, data type, scope, array offsets }

        -- DATA TYPE
        
        -- look at suffix to override
        if    last(name) = '%'  then dataType = INTEGER                                        
        elsif last(name) = '!'  then dataType = SINGLE
        elsif last(name) = '&'  then dataType = LONG
        elsif last(name) = '#'  then dataType = DOUBLE
        elsif last(name) = '$'  then dataType = STRING
                                     dataSize = 0
        elsif elementFlag = 0
        and   dataType = 0      then dataType = SINGLE
        end if

        -- storage type
        if dataType > 0                         then storageIndex = USER_TYPE
        elsif dataType = STRING and arrayFlag   then storageIndex = STRS             
        elsif dataType = STRING                 then storageIndex = STR                  
        elsif arrayFlag                         then storageIndex = NUMS
                                                else storageIndex = NUM                                                                                        
        end if                               

        scope = get_scope()

        if not useLocals        -- in main ...
        or staticFlag then      -- .. or static flag set

            -- create a global/static variable
    
            -- enough room?
            if length( globals ) > MAX_GLOBALS then
                abort_error( "Error", "Maximum number of globals exceeded." )
            end if            

            -- build reference { name, dataType, scope, array indexes }
            globals[storageIndex] = append( globals[storageIndex],
                                        var_proto( name ) )
        
            -- index
            at = length( globals[storageIndex] )
        
            -- reference            
            return sprintf( "g%s[%d]", {storage_text(storageIndex),at})
    
    else
    
            -- create a local variable
    
            -- enough room?
            if length( locals ) > MAX_LOCALS then
                abort_error( "Error", "Maximum number of locals exceeded." )
            end if            

            -- build reference { name, dataType, scope, array indexes }
            locals[storageIndex] = append( locals[storageIndex], 
                                        var_proto( name ) )
        
            -- index
            at = length( locals[storageIndex] )
        
            -- reference
            return sprintf( "v%s[%d]", {storage_text(storageIndex),at})

        end if        
    end if


end function



---------------------------------------------------------------------------
function make_unindexed_var( sequence varName )

    -- return the Euphoria handle for a variable

    -- get reference information    
    return v_ref( varName, {}, {} )

end function


---------------------------------------------------------------------------
function var_to_indexes( sequence s )

    -- convert a reference to indexes
    -- for example, gNum[sin(x)][1] --> { "gNum", "sin(x)", "1" }

    integer index, outside, startAt, brackets
    sequence indexes

    -- initialize index list
    indexes = {}

    -- quick exit
    if length( s ) < 7 then
        -- too short
        return "{}"

    elsif s[length(s)] != ']' then
        -- more than just indexes
        return "{}"

    elsif compare( s[1..5], "gNum[" )=0 then
        indexes = "'g',NUM"
        startAt = 5
        
    elsif compare( s[1..5], "gStr[" )=0 then
        indexes = "'g',STR"
        startAt = 5

    elsif compare( s[1..6], "gNums[" )=0 then
        indexes = "'g',NUMS"
        startAt = 6

    elsif compare( s[1..6], "gStrs[" )=0 then
        indexes = "'g',STRS"
        startAt = 6
        
    elsif compare( s[1..6], "gUser[" )=0 then
        indexes = "'g',USER_TYPE"
        startAt = 6

    elsif compare( s[1..5], "lNum[" )=0 then
        indexes = "'l',NUM"
        startAt = 5
        
    elsif compare( s[1..5], "lStr[" )=0 then
        indexes = "'l',STR"
        startAt = 5

    elsif compare( s[1..6], "lNums[" )=0 then
        indexes = "'l',NUMS"
        startAt = 6

    elsif compare( s[1..6], "lStrs[" )=0 then
        indexes = "'l',STRS"
        startAt = 6
        
    elsif compare( s[1..6], "lUser[" )=0 then
        indexes = "'l',USER_TYPE"
        startAt = 6
        
    else
        -- something else. exit
        return "{}"

    end if

    -- read the indexes
    outside = -1
    brackets = 0
    for i = startAt to length( s ) do

        -- beginning of an index?
        if outside then
            if s[i] = '[' then
                -- prepare to read an index's value
                startAt = i + 1
                brackets = 1
                outside = 0
            else
                -- not well formed, may be an expression i.e. --> [2]+3
                return "{}"
            
            end if

        elsif s[i] = '[' then
            -- add to bracket count
            brackets = brackets + 1

        elsif s[i] = ']' then
            -- decrement bracket count
            brackets = brackets - 1

            -- end of index?
            if brackets = 0 then
                -- add index to list
                indexes = indexes & "," & s[startAt..i-1]

                -- flag as outside
                outside = -1
            end if

        end if
    end for

    -- return accumulated indexes
    return "{" & indexes & "}"

end function


---------------------------------------------------------------------------
procedure var_assign( sequence varRef, sequence varVal )

    -- do variable assignment
    sequence varName, strLimit
    
    if dataType = INTEGER then            
        -- cast as integer
        out( varRef & " = qb_fix(" & varVal & ")\n", {} )
    
    elsif dataType = STRING
    and   dataSize != 0 then
        -- cast as limited size string
        strLimit = sprintf( "%d", {dataSize} )
        out( varRef & " = qb_pad( " & varVal & ", " & strLimit & " )\n", {} )
    
    else                      
        -- normal assignment
        out( varRef & " = " & varVal & "\n", {} )
    end if

end procedure

-----------------------------------------------------------------------------
function clean_name( sequence s )

    -- convert sequence to legal name

    integer at
    sequence cleansed
    
    cleansed = ""
    for i = 1 to length( s ) do
        if s[i] = '.' then  
            -- replace '.' with '_'
            cleansed = cleansed & "_"
            
        elsif find( s[i], SUFFIXES )
        or s[i] = ':' then
            -- don't include
            
        else
            cleansed = cleansed & s[i]
            
        end if
    end for
            
    return cleansed
    
end function

-----------------------------------------------------------------------------
function proc_name( sequence s )

    -- convert sequence to legal name

    return "proc_" & clean_name( s )

end function

-----------------------------------------------------------------------------
function label_name( sequence s )

    -- convert sequence to legal name
    
    return "label_" & clean_name( s )
    
end function



---------------------------------------------------------------------------
procedure push( object s )

    stack = append( stack, s )
    
end procedure


---------------------------------------------------------------------------
function pop()

    object s
    
    if length( stack ) then    
        s = last(stack)  
        stack = rest( stack )
        return s
    else
        abort_error( "Internal Error", "Stack underflow." )            
    end if
    
end function

---------------------------------------------------------------------------
procedure ePush( object s )

    eStack = append( eStack, s )
    
end procedure


---------------------------------------------------------------------------
function ePop()

    object s

    if length( eStack ) then    
        s = last(eStack)
        eStack = rest( eStack )
        return s
    else
        abort_error( "Internal Error", "Expression stack underflow." )        
    end if
    
end function


---------------------------------------------------------------------------
procedure eSwap()

    -- swap top two items on expression stack

    sequence s1, s2

    if length( eStack ) > 1 then        
        s1 = ePop()
        s2 = ePop()
        ePush( s1 )
        ePush( s2 )
    else
        abort_error( "Internal Error", "Expression stack underflow in swap." )
    end if
        
        
    
end procedure

---------------------------------------------------------------------------
procedure save_type()

    -- save current data type
    typeStack = append( typeStack, { dataType, dataSize } )
    
end procedure


---------------------------------------------------------------------------
procedure restore_type()

    sequence s

    if length( typeStack ) then    
        s = last( typeStack )
        dataType = s[1]
        dataSize = s[2]
        typeStack = rest( typeStack )
    else
        abort_error( "Internal Error", "Type Stack underflow." )            
    end if
    
end procedure

---------------------------------------------------------------------------
function list_to_text( sequence s )

    -- convert an arg list into a single string
    sequence s2

    -- trivial
    if length( s ) = 0 then
        return "{}"
    end if

    s2 = "{ "    
    for i = 1 to length( s )-1 do
        s2 = s2 & s[i] & ", "
    end for
    s2 = s2 & s[length(s)] & " }"
    
    return s2
    
end function

---------------------------------------------------------------------------
function builtin( sequence called, sequence args, sequence list )

    -- find the correct builtin, and return the code
    -- builtins are in the form { name, args, code }
    
    integer index
    
    -- find the function
    index = 0
    for i = 1 to length( list ) do
        if  compare( called, list[i][1] ) = 0   -- same name
        and length( args ) = list[i][2] then    -- same number of args
            -- match
            index = i
            exit
        end if
    end for   

    -- match?            
    if index = 0 then
        errText = "Wrong number of arguments in " & upper( called ) & "."
        return {}
    else    
        -- apply args to call
        return sprintf( list[index][3], args )
    end if

end function


---------------------------------------------------------------------------
procedure end_proc()

    -- close off prior sub/functions
    
    -- end the labels
    if labelFlag then
        indent(-1)
        out("end if\n", {} )
    end if

    -- handle passbacks
    -- don't add list if option not selected
    if PassByReference then
        out( "cleanup( %s, passBacks )\n", {passBackList} )                
    else
        out( "cleanup( {}, {} )\n", {} )
    end if    
            
    -- return result for functions
    if compare( subType, "function" ) = 0 then
        out( "\n", {} )
        out( "return returnValue\n", {} )
    end if

    -- end        
    indent( -1 )
    out( "\n", {} )
    out( "end %s --%s\n", {subType,subName} )
    out( "\n", {} )
    out( "\n", {} )
    indent(1)
        
end procedure
        




---------------------------------------------------------------------------
procedure build_header( sequence args )

    -- args is in the form { { name, dataType } ... }

    integer index

    -- set default return value based on function name
    guess_data_type( subName )
    if dataType = STRING then
        out( "returnValue = \"\"\n", {} )
        out( "\n", {} )
    else
        out( "returnValue = 0\n", {} )
        out( "\n", {} )
    end if        

    -- clear the passBackList
    passBackList = ""
    for i = 1 to length( args ) do
            -- build passBackList
            if i = 1 then
                passBackList = args[1][1]
            else
                passBackList = passBackList & ", " & args[i][1]
            end if
    end for
    passBackList = "{ " & passBackList & " }"

    -- find the labels
    index = a_find( subName, labels )

    -- clear the string limit
    dataSize = 0

    -- no lables?
    if length( labels[index] ) = 1 then

        -- clear flag
        labelFlag = 0
    
        -- assign the args       
        out( "-- assign the arg list\n", {} )
        for i = 1 to length( args ) do
            -- set the data type
            dataType = args[i][2]

            -- create the variable
            out( "%s = args[%d]\n", {args[i][1], i} )
        end for
        
        return  
    else
        -- set flag
        labelFlag = -1
        
    end if

    -- skip setup if called from gosub           
        out( "-- direct %s call?\n", {subType} )
        out( "if  not goto \n", {} )
        out( "and not gosub then\n", {} )            
        indent(1)                  
        
            -- parameters?
            if length( args ) then       
                out( "-- assign parameters\n", {} )
                for i = 1 to length( args ) do
                    -- set the data type
                    dataType = args[i][2]
        
                    -- create the variable
                    out( "%s = args[%d]\n", {args[i][1], i} )
                end for
                out( "\n", {} )
            end if                     

        indent(-1)     
        out("end if\n", {} )

        out("if goto = 0 or gosub then\n", {} )
            indent(1)              
            out("if gosub then\n", {} )
                indent(1)
                out( "goto = gosub\n", {} )
                out( "gosub = 0\n", {} )
            indent(-1)
            out("else\n", {} )            
                indent(1)
                out( "goto = %s\n", {label_name( subName & "_setup" )} )
            indent(-1)
            out("end if\n", {} )
            out("\n",{})
            out( "-- label handling loop\n", {} )
            out( "while goto do\n", {} )             
        
                -- setup label
                indent(1)        
                tmp = subName & "_setup"
                
                -- functions return values
                if compare( subType, "function" ) = 0 then
                    -- call function
                    out("if jump_to( %s ) then x = %s( {}, {} ) end if\n", 
                        {label_name(tmp),proc_name(subName)} )
                else   
                    -- call procedure
                    out("if jump_to( %s ) then %s( {}, {} ) end if\n", 
                        {label_name(tmp),proc_name(subName)} )
                end if

                -- remaining labels...        
                for i = 2 to length( labels[index] ) do 
                    tmp = labels[index][i]         
                    
                    if compare( subType, "function" ) = 0 then
                        -- call function
                        out("if jump_to( %s ) then x = %s( {}, {} ) end if\n", 
                            {label_name(tmp),proc_name(subName)} )
                    else   
                        -- call procedure
                        out("if jump_to( %s ) then %s( {}, {} ) end if\n", 
                            {label_name(tmp),proc_name(subName)} )
                    end if
                end for
                indent(-1)
        
            out( "end while\n", {} )
            out( "returnFlag = 0\n", {} )
        
        indent(-1)        
        out( "\n", {} )

        -- close the current call, and start the setup label
        out( "\n", {} )
                
        out( "elsif goto = %s_setup then\n", {label_name(subName)} )
            indent(1)
            out( "-- clear goto\n", {} )
            out( "goto = 0\n", {} )
            out( "\n", {} )

end procedure


--
-- THE ACTIONS
--

---------------------------------------------------------------------------
function action_is( sequence s )
    
    -- true if action matches s, set flag
    
    if compare( act, s ) = 0 then
        matchedActionToken = 1
        return 1
    else
        return 0
    end if

end function



---------------------------------------------------------------------------
function act_print( sequence arg )

    -- PRINTING

        if action_is( "#PrintScreen" ) then
            -- push { output device, cr, mask, arglist }
            push( {"0", 1, "", {}} )
        
        elsif action_is( "#PrintTo" ) then
            -- specify the output device
            if PrintMode = FAST then
                push( { "file[" & arg & "][F_HANDLE]", 1, "", {} } )
            else
                push( { arg, 1, "", {} } )
            end if
        
        elsif action_is( "#LPrint" ) then
            -- specify the output device
            if PrintMode = FAST then
                push( { "PRN", 1, "", {} } )
            else
                push( { "-1", 1, "", {} } )                
            end if
        
        elsif action_is( "#Print" ) then
            -- print

            tmp = pop()
            expr = ePop()

            if PrintMode = COMPATIBLE then
                if dataType = STRING then                
                    out( "print_str( %s, %s )\n", {tmp[1],expr} )
                else
                    out( "print_num( %s, %s )\n", {tmp[1],expr} )
                end if


            else      
                -- format
                if dataType = STRING then
                    tmp[3] = tmp[3] & "%s"
                else                
                    tmp[3] = tmp[3] & " %g "
                end if   
                
                -- data
                tmp[4] = append( tmp[4], expr )
            end if

            -- flag cr
            tmp[2] = 1

            push( tmp )
            
        elsif action_is( "#EndPrint" ) then
            -- end of print statement

            -- get args
            tmp = pop()

            -- cr?
            if tmp[2] then
                if PrintMode = COMPATIBLE then
                    out( "print_cr( %s )\n", {tmp[1]} )
                else
                    tmp[3] = tmp[3] & "\\n"
                end if
            end if

            if PrintMode = FAST then
                -- build arg list
                expr = ""
                for i = 1 to length( tmp[4] ) do
                
                    -- add to list
                    expr = expr & tmp[4][i]
                    
                    -- comma
                    if i < length( tmp[4] ) then
                        expr = expr & ","
                    end if                    
                end for

                -- output device
                if compare( tmp[1], "-1" ) = 0 then
                    -- printer
                    tmp[1] = "PRN"
                elsif compare( tmp[1], "0" ) = 0 then
                    -- screen
                    tmp[1] = "1"
                end if
                
                -- output
                out( "printf( " & tmp[1]                -- output device
                        & ", \"" & tmp[3] & "\", {"     -- mask
                        & expr & "} )\n", {} )          -- args
            end if

        elsif action_is( "#PrintComma" ) then    
            -- print tab
            tmp = pop()

            if PrintMode = COMPATIBLE then
                out( "print_tab( %s )\n", {tmp[1]} )
            else
                tmp[3] = tmp[3] & "\\t\\t"
            end if
            
            -- flag no cr
            tmp[2] = 0
            push( tmp )

        elsif action_is( "#NoCr" ) then    
            -- flag no cr     
            tmp = pop()
            tmp[2] = 0
            push( tmp )

        elsif action_is( "#Cr" ) then    
            -- flag cr
            tmp = pop()
            tmp[2] = 1
            push( tmp )

        elsif action_is( "#Tab" ) then    
            -- tab
            if PrintMode = COMPATIBLE then
                tmp = pop()
                out( "print_tab_to( %s, %s )\n", {tmp[1],ePop()} )
                push( tmp )
            else
                errText = "TAB is only supported in COMPATIBLE PRINT mode."
            end if


        elsif action_is( "#PrintUsing{" ) then
            -- add mask to list:
            -- push { output device, cr, mask, arglist, using }
            tmp = pop()
            tmp = append( tmp, ePop() )
            push( tmp )

        elsif action_is( "#PrintUsingArg" ) then
            -- build expression
            tmp = pop()
            
            -- add expression
            tmp[4] = append( tmp[4], ePop() )
            
            -- return to stack
            push( tmp )


        elsif action_is( "#PrintUsingSemi" ) then
            -- set cr flag to 0
            tmp = pop()
            tmp[2] = 0
            push( tmp )


        elsif action_is( "#}PrintUsing" ) then
            -- add to arg list
            tmp = pop()

            -- FAST mode print control
            -- attach tab/cr to end of mask?
            if  PrintMode = FAST
            and tmp[2] = 1 then
                    -- cr         
                    tmp[5] = tmp[5][1..length(tmp[5])-1] & "\\n\""
            end if
            
            -- output: device, mask, args
            out( "qb_print_using( %s, %s, %s )\n",
                {tmp[1],tmp[5],list_to_text(tmp[4])} )

            -- COMPATIBLE mode print control
            if PrintMode = COMPATIBLE
            and tmp[2] = 1 then
                -- cr    
                out( "print_cr( %s )\n", {tmp[1]} )
            end if

        end if

        -- return match result
        return matchedActionToken

end function

---------------------------------------------------------------------------
function act_input( sequence arg )

    -- INPUT
        if action_is( "#Input{" ) then
            -- push { source, prompt, "?" flag, crFlag, lineInFlag, vars }
            push( { "0", "\"\"", -1, 0, 0, {} } )
        
        elsif action_is( "#LineInput{" ) then
            -- push { source, prompt, "?" flag, crFlag, lineInFlag, vars }
            push( { "0", "\"\"", 0, 0, 1, {} } )
        
        elsif action_is( "#InputFile" ) then
            -- store file index
            tmp = pop()
            tmp[1] = arg
            push( tmp )
        
        elsif action_is( "#InputPrompt" ) then
            -- push prompt
            tmp = pop()
            tmp[2] = arg
            push( tmp )
        
        elsif action_is( "#InputNoQFlag" ) then
            -- flag question mark off
            tmp = pop()
            tmp[3] = 0
            push( tmp )
        
        elsif action_is( "#InputCrFlag" ) then
            -- flag cr
            tmp = pop()
            tmp[4] = 1
            push( tmp )
        
        elsif action_is( "#Input" ) then
            -- add var and type to stack                    
            tmp = pop()
            if dataType = STRING then
                -- string input
                tmp[6] = append( tmp[6], { ePop(), 's' } )
            else
                -- numeric input
                tmp[6] = append( tmp[6], { ePop(), 'n' } )
            end if
            push( tmp )
        
        elsif action_is( "#}Input" ) then
            -- numeric input
            tmp = pop()

            -- build the type list
            tmp1 = ""
            for i = 1 to length( tmp[6] ) do
                -- append types
                tmp1 = tmp1 & tmp[6][i][2]
            end for

            -- call input routine
            if compare( tmp[1], "0" ) = 0 then
                -- screen input
                
                -- call to input function    
                expr = sprintf( "input_screen( %s, %d, %d, %d, \"%s\" )",
                        {tmp[2],tmp[3],tmp[4],tmp[5],tmp1} )
                        
            else                          
                -- file input
                
                -- call to input function
                expr = sprintf( "input_file( %s, %d, \"%s\" )", 
                                    {tmp[1],tmp[5],tmp1} )
            end if
                                                
            -- single or multiple return values?
            if length( tmp1 ) = 1 then
                -- single assignment
                var_assign( tmp[6][1][1], expr )
            else
                -- multi assignment              
                
                -- call to input...                
                out( "inputBuffer = %s\n", {expr} )
                
                -- ... actual assignment
                indent( 1 )
                for i = 1 to length( tmp1 ) do
                    var_assign( tmp[6][i][1], 
                        sprintf( "inputBuffer[%d]", i ) )
                end for
                indent( -1 )
            end if            
                        
        end if

        -- return match result
        return matchedActionToken
            
end function


---------------------------------------------------------------------------
function act_expr( sequence arg )


    -- BUILDING NUMERIC EXPRESSIONS        
        
        if action_is( "#Cons" ) then    
            -- accumulate arg in expression
            ePush( ePop() & arg )
        
        elsif action_is( "#PushExpr" ) then
            -- place arg on expression stack       
            ePush( arg )
    
    -- ASSIGNMENT
            
        elsif action_is( "#Assign{" ) then
            -- get expression to assign   
            tmp = ePop()
        
            -- save the reference
            push( tmp )    
            
            -- push the data type
            save_type()
    
        elsif action_is( "#}Assign" ) then
            -- assign value 

            -- get the data type
            restore_type()

            -- get the reference            
            tmp = pop()
            expr = ePop()

            -- strip off the suffix for comparison

            -- assigned to function name?
            if compare( remove_numeric_suffix( tmp ), subName ) = 0 then  
                -- assign to function's return value
                out( "returnValue = %s\n", {expr} )
            else  
                -- normal assignment
                var_assign( tmp, expr )
            end if            


        elsif action_is( "#Variable{" ) then
            -- push { name, indexes, element }
            push( { ePop(), {}, {} } )
            
            
        elsif action_is( "#}Variable" ) then
            -- place the variable expressions on the expression stack
            -- reference is: { name, indexes, elements }

            -- get the reference
            tmp = pop()
            
            -- get the reference
            tmp1 = v_ref( tmp[1], tmp[2], tmp[3] )

            -- initial value
            expr = tmp1
            
            -- indexes?
            if  length( tmp[2] )                    -- indexed
            and compare( tmp[2], "[]" )!=0 then     -- not a()

                -- if compatible mode, need to point to data
                if ArrayMode = COMPATIBLE then
                    -- access the array
                    expr = expr & "[VALS]"
                end if
                
                for i = 1 to length( tmp[2] ) do    
                    if ArrayMode = FAST then
                        -- assume offset of 1
                        expr = sprintf( "%s[ %s+1 ]", {expr,tmp[2][i]} )
                    else
                        -- use stored offset
                        expr = sprintf( "%s[ %s+%s[OFS][%d] ]", 
                                    {expr,tmp[2][i],tmp1,i} )
                    end if

                end for

            end if


            -- element?
            if length( tmp[3] ) then

                for i = 1 to length( tmp[3] ) do

                    -- get an element
                    tmp2 = tmp[3][i]                       
                       
                    -- remove the '.' from the element
                    tmp2 = tmp2[2..length(tmp2)]
                    
                    -- match flag                    
                    at = 0                   
                    -- search                    
                    for j = 1 to length( elements ) do       
                        -- matching element, same data type
                        if  compare( tmp2, elements[j][E_NAME] ) = 0
                        -- and elements[j][E_OWNER] = dataType
                        then

                            -- set flag                                                    
                            at = j
                            
                            -- leave loop
                            exit         
                            
                        end if
                        
                    end for
                    
                    if at = 0 then
                        -- wrong element
                        abort_error("Error",
                            tmp[1] &  " does not have element '"
                            & tmp2 & "'." )
                    end if
                                        
                    -- get the index
                    tmp3 = elements[at][E_INDEX]

                    -- add the index
                    expr = sprintf( "%s[%d]", {expr,tmp3} )
    
                    -- set the type and size
                    dataType = elements[at][E_TYPE]
                    dataSize = elements[at][E_SIZE]
                    
                end for
            end if

            -- push on expression stack
            ePush( expr )

        elsif action_is( "#MakeVar" ) then
            -- place the arg, an unindexed variable, on the stack
            ePush( make_unindexed_var( arg ) )

        elsif action_is( "#Parens" ) then
            -- place parens around expression
            ePush( "(" & ePop() & ")" )

        elsif action_is( "#UnaryMinus" ) then
            -- place parens around expression
            ePush( "-(" & ePop() & ")" )

    -- BUILDING EXPRESSIONS

        elsif action_is( "#Add" ) then                    
            eSwap()
            if dataType = STRING then
                ePush( sprintf( "%s & %s", {ePop(),ePop()} ) )                
            else
                ePush( sprintf( "%s + %s", {ePop(),ePop()} ) )
            end if

        elsif action_is( "#Concat" ) then                    
            eSwap()
            ePush( sprintf( "%s & %s", {ePop(),ePop()} ) )                

        elsif action_is( "#Sub" ) then                    
            eSwap()
            ePush( sprintf( "%s - %s", {ePop(),ePop()} ) )            

        elsif action_is( "#Mul" ) then                    
            eSwap()
            ePush( sprintf( "%s * %s", {ePop(),ePop()} ) )            

        elsif action_is( "#Div" ) then                    
            eSwap()
            ePush( sprintf( "%s / %s", {ePop(),ePop()} ) )

        elsif action_is( "#IntDiv" ) then                    
            eSwap()
            ePush( sprintf( "qb_fix(%s / floor(%s+.5))", {ePop(),ePop()} ) )            

        elsif action_is( "#Power" ) then
            eSwap()
            ePush( sprintf( "power(%s, %s)", {ePop(),ePop()} ) )            

        elsif action_is( "#Mod" ) then
            eSwap()
            ePush( sprintf( "remainder(%s, %s)", {ePop(),ePop()} ) )            

        elsif action_is( "#Eq" ) then                    
            eSwap()
            if dataType = STRING then
                ePush( sprintf( "compare(%s, %s)=0", {ePop(),ePop()} ) )
            else
                ePush( sprintf( "(%s = %s)", {ePop(),ePop()} ) )
            end if

        elsif action_is( "#Ne" ) then                    
            eSwap()
            if dataType = STRING then
                ePush( sprintf( "compare(%s, %s)!=0", {ePop(),ePop()} ) )
            else
                ePush( sprintf( "(%s != %s)", {ePop(),ePop()} ) )
            end if

        elsif action_is( "#Gt" ) then                    
            eSwap()
            if dataType = STRING then
                ePush( sprintf( "compare(%s, %s)>0", {ePop(),ePop()} ) )
            else
                ePush( sprintf( "(%s > %s)", {ePop(),ePop()} ) )
            end if

        elsif action_is( "#Ge" ) then                    
            eSwap()
            if dataType = STRING then
                ePush( sprintf( "compare(%s, %s)>=0", {ePop(),ePop()} ) )
            else
                ePush( sprintf( "(%s >= %s)", {ePop(),ePop()} ) )
            end if
                                            
        elsif action_is( "#Lt" ) then                    
            eSwap()
            if dataType = STRING then
                ePush( sprintf( "compare(%s, %s)<0", {ePop(),ePop()} ) )
            else
                ePush( sprintf( "(%s < %s)", {ePop(),ePop()} ) )
            end if

                                        
        elsif action_is( "#Le" ) then                    
            eSwap()
            if dataType = STRING then
                ePush( sprintf( "compare(%s, %s)<=0", {ePop(),ePop()} ) )
            else
                ePush( sprintf( "(%s <= %s)", {ePop(),ePop()} ) )
            end if

            
        elsif action_is( "#And" ) then                    
            eSwap()
            ePush( sprintf( "and_bits(%s, %s)", {ePop(),ePop()} ) )


        elsif action_is( "#Or" ) then
            eSwap()
            ePush( sprintf( "or_bits(%s, %s)", {ePop(),ePop()} ) )
            
            
        elsif action_is( "#Xor" ) then
            eSwap()
            ePush( sprintf( "xor_bits(%s, %s)", {ePop(),ePop()} ) )


        elsif action_is( "#Not" ) then
            ePush( sprintf( "not_bits(%s)", {ePop()} ) )


    -- DATA TYPE FLAGS

        elsif action_is( "#NumType" ) then    
            -- data type is numeric

            dataType = SINGLE
            
        elsif action_is( "#StrType" ) then    
            -- pop data type
            dataType = STRING
            dataSize = 0

        elsif action_is( "#ClearType" ) then    
            -- push data type
            dataType = 0
            dataSize = 0

        elsif action_is( "#PushType" ) then    
            -- push data type
            save_type()
            
        elsif action_is( "#PopType" ) then    
            -- pop data type
            restore_type()

        end if

        -- return match result
        return matchedActionToken

end function

---------------------------------------------------------------------------
procedure dim_array( sequence dim )

    -- info is in the form
    -- { name, type, scope, {indexPair1 ... indexPair n} }

    integer at
    sequence offsets, data, size, ref

    -- set the data type      
    dataType = dim[V_TYPE]

    -- default values for array    
    if dataType > 0 then
        -- user defined
        data = types[dim[V_TYPE]][T_DEFAULT]
        ref = v_ref( dim[V_NAME], "", "UserDefined" )
        
    elsif dim[V_TYPE] = STRING then
        -- string
        data = "\"\""
        ref = v_ref( dim[V_NAME], "Array", "" )
        
    else
        -- numeric array
        data = "0"
        ref = v_ref( dim[V_NAME], "Array", "" )
        
    end if
    

    -- if index is "[]", it's just a reference. all we needed to do
    -- was allocate an array, so leave now.
    if compare( dim[V_INDEXES], "[]" ) = 0 then
        -- leave
        return
    end if

    -- build assignment
    for i = length( dim[V_INDEXES] ) to 1 by -1 do     

        -- calculate range to allocate
        if compare( dim[V_INDEXES][i][V_LOW], "0" ) = 0 then
            -- default starts at 0
            size = dim[V_INDEXES][i][V_HIGH]
            
        elsif ArrayMode = COMPATIBLE then
            -- allocate only as much space as needed
            size = dim[V_INDEXES][i][V_HIGH] & "- " 
                        & dim[V_INDEXES][i][V_LOW]  
                        
        else
            -- no offsets, assume all arrays start at zero
            -- ignore low, allocate up to high index
            size = dim[V_INDEXES][i][V_HIGH]
                                    
        end if
        
        -- allocation
        data = sprintf( "repeat( %s, 1+%s )", {data, size} )
        
    end for            

    -- run-time creation of the array
    if ArrayMode = COMPATIBLE then
        -- index to value portion
        out( "%s[VALS] = %s\n", {ref,data} )
    else
        -- only data
        out( "%s = %s\n", {ref,data} )
    end if

    -- store run-time ranges
    if ArrayMode = COMPATIBLE then

        -- build list
        offsets = ""

        -- need a seperate offset for each item
        for i = 1 to length( dim[V_INDEXES] ) do     
    
            -- calculate range to allocate
            offsets = offsets & "-(" & dim[V_INDEXES][i][V_LOW] & "-1)"

            -- add comma?
            if i < length( dim[V_INDEXES] ) then
                offsets = offsets & ","
            end if
            
        end for

        -- assign offsets                       
        out( "%s[OFS] = { %s }\n", {ref,offsets} )

    end if


end procedure

---------------------------------------------------------------------------
procedure build_indexes( sequence varName, sequence indexes )

    -- indexes takes the form:
    -- { index1, index2 ... }

    integer dType, at
    sequence info, offsets
    object result

    -- blank indexes are legal: a() can be passed as an arg
    if length( indexes ) = 0 then
        return
    end if

    -- clear info
    info = {}

    -- look for it first in global string arrays
    at = a_find( varName, globals[dType] )
    if at > 0 then
        -- visible within scope?
        if globals[dType][at][V_SCOPE] > 0
        or not useLocals then
            info = globals[dType][at]
        end if
    end if

    -- not found in globals?
    if length( info ) = 0 then
        -- look for it in local arrays
        at = a_find( varName, locals[dType] )
        if at > 0 then
            info = locals[dType][at]
        end if
    end if
            
    -- not found in either globals or locals?
    if length( info ) = 0 then
        -- assume dim( 1, 10 )
        offsets = { {"1", "10"} }                
    else
        -- in list. get values
        offsets = info[V_INDEXES]
    end if

    -- does assigned index match size of array?
    if length( offsets ) != length( indexes ) then
        errText = sprintf( "Expected %d dimensions for '%s', not %d.",
                                {length(indexes),varName,length(offsets)} )
    end if

    expr = ""

    -- build indexes            
    for i = 1 to length( offsets ) do
        if compare( offsets[i][V_LOW], "0" ) = 0 then
            expr = sprintf("%s[1+%s]",{expr,indexes[i]} )
        else
            -- calculate, including offset
            expr = sprintf("%s[1+%s- %s]",{expr,indexes[i],offsets[i][V_LOW]} )
        end if
    end for
            
    -- append result to expression on stack
    ePush( ePop() & expr )

end procedure

---------------------------------------------------------------------------
function act_array( sequence arg )


    -- SUBSCRIPTS
        if action_is( "#Dim{" ) then
            -- push { name, type, scope, indexes }
            guess_data_type( arg )
            push( {arg, dataType, 0, {} } )            

        elsif action_is( "#DimHigh" ) then
            -- add range {0..number} to list
            tmp = pop()
            tmp[V_INDEXES] = append( tmp[V_INDEXES], {"0", ePop() } )
            push( tmp )

        elsif action_is( "#DimRange" ) then
            -- add index to list
            eSwap()    
            tmp = pop()
            tmp[V_INDEXES] = append( tmp[V_INDEXES], { ePop(), ePop() } )
            push( tmp )

        elsif action_is( "#DimAs" ) then
            -- add index to list

            tmp = pop()
            tmp[V_TYPE] = dataType
            push( tmp )

        elsif action_is( "#}Dim" ) then
            -- build the array assignment
            -- list is { name, type, scope, {low,high}, {low,high} ... }

            -- get the list
            tmp = pop()

            -- restore the type
            dataType = tmp[V_TYPE]

            -- indexed?   
            if length( tmp[V_INDEXES] ) > 0 then

                -- build indexed
                dim_array( tmp )
            
            else

                -- allocate space
                tmp1 = make_unindexed_var( tmp[V_NAME] )
                

                -- user type?
                if dataType > 0 then
                    -- need to assign
                    out( "%s = %s\n", {tmp1,types[dataType][T_DEFAULT]} )

                end if
            
            end if

        elsif action_is( "#Index" ) then
            -- add an index to the variable structure
            -- structure should be on the stack
            tmp1 = pop()

            -- only dummy index?
            if compare( tmp1[2], "[]" )= 0 then
                -- replace with real index
                tmp1[2] = { ePop() }
            else
                -- append item to list
                tmp1[2] = append( tmp1[2], ePop() )
            end if

            push( tmp1 )

        elsif action_is( "#DummyIndex" ) then
            -- add a "dummy" index to note that the
            -- variable is indexed

            tmp1 = pop()

            -- append item to list
            tmp1[2] = "[]"

            push( tmp1 )


    -- DATA TYPES

        elsif action_is( "#Integer" ) then
            -- set data type
            dataType = INTEGER

        elsif action_is( "#Long" ) then
            -- set data type
            dataType = LONG

        elsif action_is( "#Single" ) then
            -- set data type
            dataType = SINGLE
            
        elsif action_is( "#Double" ) then
            -- set data type
            dataType = DOUBLE
            
        elsif action_is( "#String" ) then
            -- set data type
            dataType = STRING

        elsif action_is( "#StringSize" ) then
            -- set data size
            dataSize = val( arg )
            
        elsif action_is( "#StringSizeZero" ) then
            -- set data size
            dataSize = 0

        elsif action_is( "#UserType" ) then
            -- set as user-defined data type
            dataType = a_find( arg, types )

        elsif action_is( "#GuessType" ) then
            -- set data type based on name
            guess_data_type( arg )

    -- USER DEFINED TYPES
        elsif action_is( "#Type{" ) then
            -- ordinarily, elements are hidden so they don't conflict
            -- with vars. Making them visible lets the parser read
            --      %element AS <type>
            -- correctly

            hideElements = 0

        elsif action_is( "#}Type" ) then
            -- return to hiding elements

            hideElements = -1

        elsif action_is( "#Element" ) then
            -- add an element to the variable structure
            -- structure should be on top of stack

            tmp1 = pop()

            -- append item to list
            tmp1[3] = append( tmp1[3], arg )

            push( tmp1 )

        end if

        -- return match result
        return matchedActionToken


end function


---------------------------------------------------------------------------
function act_shared( sequence arg )

    -- SHARED/COMMON/CONST

        if action_is( "#SharedOn" ) then
            if useLocals then
                -- only be set in main
                errText = "Illegal in Procedure or DEF FN."
            end if
            
            -- set shared flag on
            sharedFlag = -1
            
        elsif action_is( "#SharedOff" ) then
            -- set shared flag off
            sharedFlag = 0            

        elsif action_is( "#Shared{" ) then
            if not useLocals then
                -- only be set in main
                errText = "Illegal outside SUB/FUNCTION."
            end if
        
            -- push name of variable on stack
                push( arg )

        elsif action_is( "#SharedVar" ) then
            -- flag as visible (#1)

            -- get name
            expr = pop()

            -- push name and type
            if last( expr ) = '$' then
                push( {expr, STR} )
            else
                push( {expr, NUM} )
            end if

        elsif action_is( "#SharedArray" ) then
            -- flag as visible (#1)

            -- get name
            expr = pop()

            -- push name and type
            if last( expr ) = '$' then
                push( {expr, STRS} )
            else
                push( {expr, NUMS} )
            end if

        elsif action_is( "#}Shared" ) then
            -- stack contains { name, storage index }

            -- get name, datatype
            tmp = pop()

            -- look for it
            tmp1 = a_find( tmp[1], globals[tmp[2]] )

            -- doesn't exist yet?
            if tmp1 = 0 then
                -- allot in table
                globals[tmp[2]] = append( globals, var_proto( tmp[1] ) )
                tmp1 = length( globals[tmp[2]] )
            end if

            -- find variable
            tmp1 = a_find( tmp[1], globals[tmp[2]] )

            -- if not globally visible...
            if globals[tmp[2]][tmp1][V_SCOPE] != V_SHARED then
                --- set as locally visible
                globals[tmp[2]][tmp1][V_SCOPE] = V_LOCAL
            end if

        elsif action_is( "#Common{" ) then
            -- beginning of common list
            
            -- only in main
            if useLocals then
                -- only be set in main
                errText = "Illegal in Procedure or DEF FN."
            end if
        
            -- push name of variable on stack
                push( arg )

        elsif action_is( "#CommonVar" ) then
            -- flag as visible (#1)

            -- get name
            expr = pop()

            -- push name and type
            if last( expr ) = '$' then
                push( {expr, STR} )
            else
                push( {expr, NUM} )
            end if

        elsif action_is( "#CommonArray" ) then
            -- flag as visible (#1)

            -- get name
            expr = pop()

            -- push name and type
            if last( expr ) = '$' then
                push( {expr, STRS} )
            else
                push( {expr, NUMS} )
            end if

        elsif action_is( "#}Common" ) then
            -- stack contains { name, datatype }

            -- get name, datatype
            tmp = pop()

            -- look for it
            tmp1 = a_find( tmp[1], globals[tmp[2]] )

            -- doesn't exist yet?
            if tmp1 = 0 then
                -- allot in table
                globals[tmp[2]] = append( globals, var_proto( tmp[1] ) )
                tmp1 = length( globals[tmp[2]] )
            end if

            -- set to globally visible...
            globals[tmp[2]][tmp1][V_SCOPE] = V_SHARED

        elsif action_is( "#Const{" ) then
            -- push text onto stack
            push( arg )


        elsif action_is( "#}Const" ) then
            -- pop the name off the stack
            tmp = pop()

            -- enough room?
            if length( constants ) > MAX_CONSTS then
                abort_error( "Internal Error",
                    sprintf( "Number of constants exceeds maximum of %d.", 
                                {MAX_CONSTS} ) )
            end if

            -- make it a constant
            constants = append( constants, var_proto( tmp ) )
            
            -- index
            at = length( constants )

            -- assign it
            out( "c[%d] = %s\n", {at, ePop()} )


    -- STATIC

        elsif action_is( "#StaticOn" ) then
            if not useLocals then
                -- only be set outside main
                errText = "Illegal outside of Procedure or DEF FN."
            end if
            
            -- set static flag on
            staticFlag = -1
            
        elsif action_is( "#StaticOff" ) then
            -- set shared flag off
            staticFlag = 0            
                                
        end if

        -- return match result
        return matchedActionToken


end function

---------------------------------------------------------------------------
procedure add_dependency( sequence s )

    -- add s as current dependancy of current sub/function

    integer this
    
    this = length( uses )

    -- skip if already in list        
    if not find( s, uses[this][U_LIST] ) then
            -- place on the dependancies list
            uses[this][U_LIST] = append( uses[this][U_LIST], s )
    end if

end procedure    


---------------------------------------------------------------------------
function act_sub( sequence arg )


    -- DEFINE SUB/FUNCTION

        if action_is( "#UserName{" ) then        

            -- closePrior
            end_proc()
            
            -- save current sub/function name
            subName = arg
            if find( subName, userSubs ) then
                subType = "procedure"
            else
                subType = "function"
            end if

            -- build new "uses" item
            uses = append( uses, { subName, {}, {}, {} } )

            -- build the proc
            indent(-1)
            out( "%s %s( sequence args, sequence passBacks )\n",
                    {subType, proc_name(subName)} )
            indent( 1 )
            
            -- guess data type
            guess_data_type( subName )
            subDataType = dataType
                
            -- remove data type suffix
            subName = remove_numeric_suffix( subName )

            -- push the list of arguments            
            push( {} )

            -- set local flag on
            useLocals = -1

            -- clear the static flag
            staticFlag = 0

            -- clear globals flagged as static by hiding the names            
            -- clear globals flagged as local
            
            -- for each type of global
            for i = 1 to length( globals ) do
                -- for each variable within the type
                for j = 1 to length( globals[i] ) do
            
                    -- if static, hide name
                    if globals[i][j][V_SCOPE] = V_STATIC then
                        -- zap name
                        globals[i][j][V_NAME] = ""
                        -- make invisible
                        globals[i][j][V_SCOPE] = V_HIDDEN
                        
                    -- if not shared, hide scope
                    elsif globals[i][j][V_SCOPE] != V_SHARED then
                        -- make invisible
                        globals[i][j][V_SCOPE] = V_HIDDEN
                        
                        
                    end if
                end for
            end for

            -- clear the local variables
            locals = repeat( {}, NUM_TYPES )

        elsif action_is( "#UserParm" ) then

            -- get the data
            tmp = pop()

            -- get the name
            tmp1 = ePop()

            -- save name and type
            push( append( tmp, {tmp1, dataType} ) )

        elsif action_is( "#UserParmType" ) then
            -- set the type of the last parm

            -- get the data
            tmp = pop()

            -- change the type
            tmp[length(tmp)][2] = dataType

            -- save it
            push( tmp )

        elsif action_is( "#}UserName" ) then
            -- user defined function/sub

            -- set up locals
            out( "setup_frame()\n", {} )

            -- get the args                        
            tmp = pop()       

            -- label handling code
            build_header( tmp )

            
        elsif action_is( "#UserEnd" ) then         
            -- no action needed
            
    -- CALL BUILT IN/USER FUNCTIONS

        elsif action_is( "#Func{" ) then
            -- push name, arglist
            push( { arg, {}, "" } )
                           
        elsif action_is( "#Arg" ) then
            -- add new arg to arglist

            tmp = pop()

            -- get new arg
            tmp1 = ePop()

            -- add arg to parameter list
            tmp[2] = append( tmp[2], tmp1 )

            if PassByReference then
                -- add indexes to address list
                if length( tmp[3] ) then
                    tmp[3] = tmp[3] & ","
                end if
                tmp[3] = tmp[3] & var_to_indexes( tmp1 )
            end if

            -- save results
            push( tmp )

        elsif action_is( "#MissingArg" ) then
            -- add -1 to arglist

            -- get arg list            
            tmp = pop()

            -- add arg to parameter list
            tmp[2] = append( tmp[2], -1 )
            
            -- push back
            push( tmp )
            
        elsif action_is( "#StaticArg" ) then
            -- add new arg to arglist
            -- null to address list
            tmp = pop()

            -- get new arg
            tmp1 = ePop()

            -- add arg to parameter list
            tmp[2] = append( tmp[2], tmp1 )

            if PassByReference then
                -- add indexes to address list
                if length( tmp[3] ) then
                    tmp[3] = tmp[3] & ","
                end if
                tmp[3] = tmp[3] & var_to_indexes( tmp1 )
            end if

            -- save results
            push( tmp )
                           
        elsif action_is( "#}Func" ) then
            -- pop name, args
            tmp = pop()
            ePush( builtin( tmp[1], tmp[2], builtin_functions ) )
            
        elsif action_is( "#}UserFunc" ) then
            -- pop name, args
            tmp = pop()

            -- call function            
            tmp1 = proc_name( tmp[1] )
            ePush( tmp1 & "( " 
                        & list_to_text( tmp[2] )
                        & ", {" & tmp[3] & "} )" )
                        
            -- add as dependancy
            add_dependency( tmp[1] )                        
    
    -- CALL BUILT IN/USER SUBS

        elsif action_is( "#Sub{" ) then
            -- push name, arglist, callback list
            push( { arg, {}, "" } )
            
        elsif action_is( "#}Sub" ) then
            -- pop name, args, callbacks 
            tmp = pop()

            -- call built in sub
            out( builtin( tmp[1], tmp[2], builtin_subs ), {} )

        elsif action_is( "#}UserSub" ) then
            -- pop name, args
            tmp = pop()     
            
            -- call sub
            tmp1 = proc_name( tmp[1] )
            out( tmp1 & "( "  
                        & list_to_text( tmp[2] )
                        & ", {" & tmp[3] & "}"
                        & " )\n", {} )

            -- add to list of dependancies
            add_dependency( tmp[1] )

        end if

        -- return match result
        return matchedActionToken

end function

---------------------------------------------------------------------------
function act_label_name( sequence arg )

    -- LABELS
    
        if action_is( "#Label" ) then

            -- close the current call, and start the label

            -- ensure it ends with a colon
            if last( arg ) != ':' then
                add_dependency( arg & ":" )
            else
                add_dependency( arg )
            end if            

            indent(-1)
            out( "elsif goto = %s then \n", {label_name(arg)} )
            indent(1)
            out( "-- clear goto\n", {} )
            out( "goto = 0\n", {} )
            out( "\n", {} )

        end if

        -- return match result
        return matchedActionToken

end function

---------------------------------------------------------------------------
function act_end( sequence arg )


    -- END
    
        if action_is( "#End" ) then
            out( "abort(0)\n", {} )


    -- IF...ELSE...END IF
    
        elsif action_is( "#If" ) then    
            -- if clause

            -- numeric expression?

            if dataType = STRING then
                errText = "Expected a numeric expression in IF clause."
            else            
                out( "if %s then\n", {ePop()} )
                indent( 1 )
            end if
            
        elsif action_is( "#ElseIf" ) then
            -- else if clause
            
            -- numeric expression?
            if dataType = STRING then
                errText = "Expected a numeric expression in ELSE clause."
            else
                indent( -1 )
                out( "elsif %s then\n", {ePop()} )
                indent( 1 )
            end if
    
        elsif action_is( "#Else" ) then
            -- else clause
            indent( -1 )
            out( "else\n", "" )
            indent( 1 )
            
        elsif action_is( "#EndIf" ) then
            -- end if
            indent( -1 )
            out( "end if\n", "" )            
    
        end if

        -- return match result
        return matchedActionToken

end function

---------------------------------------------------------------------------
function act_select( sequence arg )


    -- SELECT...END SELECT

        if action_is( "#CaseStart" ) then    
            -- need to place in local var
            
            -- create a temp variable
            tmp1 = sprintf("_Case%d",{tmpCount})    -- step
            tmpCount = tmpCount + 1
            
            -- make it a string, if needed
            if dataType = STRING then
                tmp1 = tmp1 & '$'
            end if
            
            -- make it a "real" variable
            tmp1 = make_unindexed_var( tmp1 )

            out( "-- assign expression for multiple evaluations\n", {} )            
            out( "%s = %s\n", {tmp1, ePop()} )
            
            -- push expression, type, clauses, count of if statements
            push( {tmp1, dataType, {}, 0} )

            -- mark start of case statment
            out( "-- case statement\n", {} )
            
            -- indent
            indent( 1 )
            
        elsif action_is( "#CaseElse" ) then    
            -- push expression, type, and count
            tmp = pop()
            
            -- outdent
            indent( -1 )
            
            -- only clause?    
            if tmp[4] = 0 then
                out( "if 1 then\n", {} )
            else
                out( "\n", {} )
                out( "else\n", {} )            
            end if
            
            -- indent
            indent( 1 )
            
            -- return to stack
            push( tmp )
            
        elsif action_is( "#CaseEq" )
        or    action_is( "#CaseNe" ) 
        or    action_is( "#CaseGt" ) 
        or    action_is( "#CaseGe" ) 
        or    action_is( "#CaseLt" ) 
        or    action_is( "#CaseLe" ) then

            -- select operator        
            if      action_is( "#CaseEq" )  then tmp2 = "="
            elsif   action_is( "#CaseNe" )  then tmp2 = "!="
            elsif   action_is( "#CaseGt" )  then tmp2 = ">"
            elsif   action_is( "#CaseGe" )  then tmp2 = ">="
            elsif   action_is( "#CaseLt" )  then tmp2 = "<"
            elsif   action_is( "#CaseLe" )  then tmp2 = "<="
            end if

            -- compare to expression
            tmp = pop()

            -- use compare for strings            
            if dataType = STRING then     
                expr = sprintf( "compare( %s, %s ) %s 0", {tmp[1],ePop(),tmp2} )
            else
                expr = sprintf( "%s %s %s ", {tmp[1], tmp2, ePop()} )
            end if

            -- add to tests
            tmp[3] = append( tmp[3], expr )

            -- place on stack
            push( tmp )

                       
        elsif action_is( "#CaseTo" ) then    
            -- within range  
            tmp = pop()
            
            -- check data type
            -- if (dataType = STRING) and (tmp[2] != STRING) then
            --    errText = "Mis-matched data types."
            -- end if

            eSwap()            
            if dataType = STRING then
                expr = "(compare( " & tmp[1] & ", " & ePop() & " ) >= 0" &
                        " and compare( " & tmp[1] & ", " & ePop() & " ) <= 0)"
                tmp[3] = append( tmp[3], expr )
            else
                expr = "( " & tmp[1] & ">=" & ePop() 
                            & " and " & tmp[1] & "<=" & ePop() & " )"
                tmp[3] = append( tmp[3], expr )
            end if
            
            push( tmp )
            
            
        elsif action_is( "#Case" ) then    
            -- build the expression
            tmp = pop()

            -- outdent
            indent( -1 )

            -- if
            if tmp[4] = 0 then
                expr = "if    "
            else
                out( "\n", {} )
                expr = "elsif "            
            end if
            
            -- increment clause count
            tmp[4] = tmp[4] + 1
            
            -- test
            for i = 1 to length( tmp[3] ) do
                
                -- expression
                out( expr & tmp[3][i] & "\n", {} )
                
                -- rest of clause is ors
                expr = "or    "                
                
            end for
            out( "then\n", {} )

            -- clear the expressions
            tmp[3] = {}
            
            -- indent
            indent( 1 )
            
            push( tmp )
            
            
        elsif action_is( "#EndSelect" ) then    
            -- pop
            tmp = pop()          

            -- outdent
            indent( -1 )


            -- only generate end if not empty statement
            if tmp[4] != 0 then
            
                -- close if
                out( "end if\n", {} )
    
            end if
            
        end if

        -- return match result
        return matchedActionToken

end function

---------------------------------------------------------------------------
function act_wend( sequence arg )
    
        
    -- WHILE...WEND
    
        if action_is( "#While" ) then    
            -- if clause
            out( "while %s do\n", {ePop()} )
            indent( 1 )
            
        elsif action_is( "#Wend" ) then
            -- else if clause
            indent( -1 )
            out( "end while\n", "" )            

        end if
        
        -- return match result
        return matchedActionToken

end function

---------------------------------------------------------------------------
function act_exit( sequence arg )

    -- EXIT

        if action_is( "#Exit" ) then    
            -- exit control structure
            out( "exit\n", {} )

        elsif action_is( "#ExitProc" ) then    
            -- exit Sub or Function
            -- cleanup frame

            out( "goto = 0\n", {} )
            out( "returnFlag = -1\n", {} )

            -- don't add list if option not selected
            if PassByReference then
                out( "cleanup( %s, passBacks )\n", {passBackList} )                
            else
                out( "cleanup( {}, {} )\n", {} )
            end if

            -- exit
            if compare( subType, "function" ) = 0 then
                -- return value
                out( "return returnValue\n", {} )
            else
                -- exit
                out( "return\n", {} )
            end if


        elsif action_is( "#Return" ) then    
            -- exit from a GOSUB
            -- no frame cleanup

            out( "goto = 0\n", {} )
            out( "returnFlag = -1\n", {} )            
            if compare( subType, "function" ) = 0 then
                out( "return returnValue\n", {} )
            else
                out( "return\n", {} )
            end if

        elsif action_is( "#ReturnLabel" ) then    
            -- exit from a GOSUB
            -- no frame cleanup

            out( "goto = %s\n", {label_name(arg)} )
            out( "returnFlag = -1\n", {} )
            if compare( subType, "function" ) = 0 then
                out( "return returnValue\n", {} )
            else
                out( "return\n", {} )
            end if
    
        end if

        -- return match result
        return matchedActionToken

end function

---------------------------------------------------------------------------
function act_subs( sequence arg )


    -- FOR LOOPS

        if action_is( "#For{" ) then
            -- initialize loop values
            -- { variable, start, end, step }
            push( { arg, "", "", "1" } )
            
        elsif action_is( "#ForStart" ) then

            -- start value of loop      
            tmp = pop()

            -- numeric expression?
            if dataType = STRING then                    
                errText = "Expected a numeric expression in FOR value."
            end if

            -- store expression            
            tmp[2] = ePop()
            push( tmp )
            
        elsif action_is( "#ForTo" ) then
            -- end value of loop
            tmp = pop()
                      
            -- numeric expression?
            if dataType = STRING then                    
                errText = "Expected a numeric expression in TO value."
            end if
                      
            tmp[3] = ePop()
            push( tmp )
            
        elsif action_is( "#ForStep" ) then
            -- { variable, start, end, step }

            -- step value of loop
            tmp = pop()

            -- numeric expression?
            if dataType = STRING then                    
                errText = "Expected a numeric expression in STEP value."
            end if
            
            tmp[4] = ePop()
            push( tmp )
            
        elsif action_is( "#}For" ) then

            -- generate loop
            tmp = pop()

            -- increment for loop count
            tmpCount = tmpCount + 1

            if ForLoopMode = FAST then            
                -- use built-in loop for speed

                -- build a "real" Euphoria variable
                tmp1 = sprintf( "i%d", {tmpCount} )
                
                -- the For loop
                out( "for %s = %s to %s by %s do\n",
                        {tmp1,tmp[2],tmp[3],tmp[4]} )
                                            
                
                -- indent
                indent( 1 )
                
                -- set local to index
                tmp2 = make_unindexed_var( tmp[1] )
                var_assign( tmp2, tmp1 ) 
                
                -- push the variable and the "real" on the loop stack
                loopIndex = append( loopIndex, {tmp[1],tmp1} )
                
            else
                -- simulate For with While; more compatible with QBasic
                -- this takes up 2 locals

                -- index
                expr = make_unindexed_var( tmp[1] )
                
                -- step size
                tmp1 = make_unindexed_var( sprintf("_Step%d",{tmpCount} ) )
                
                -- size limit
                tmp2 = make_unindexed_var( sprintf("_Max%d",{tmpCount}) )

                out("-- set loop range\n", {} )   
                out( "%s = %s\n", {tmp1,tmp[4]} )    -- step
                out( "%s = %s\n", {tmp2,tmp[3]} )    -- limit

                -- initial value, less step
                out("-- initialize loop index, start less step\n", {} )
                out( "%s = %s- %s\n", {expr,tmp[2],tmp1} )

                -- the loop
                out("-- for loop\n", {} )
                out( "while 1 do\n", {} )

                -- indent
                indent( 1 )

                -- increment

                out( "%s = %s + %s\n", {expr,expr,tmp1} )

                -- past range?
                out( "if %s > 0 then\n", {tmp1} )
                out( "\tif %s > %s then exit  end if\n", {expr,tmp2} )
                out( "else\n", {} )
                out( "\t if %s < %s then exit  end if\n", {expr,tmp2} )
                out( "end if\n", {} )

            end if
    
        elsif action_is( "#Next" ) then       

            -- outdent
            indent( -1 )

            -- close loop
            if ForLoopMode = FAST then     
                -- close the loop
                out( "end for\n", "" )
                
                -- pop the loop index from the stack
                loopIndex = loopIndex[1..length(loopIndex)-1]
                
            else                        
                -- close the loop
                out( "end while\n", "" )
            end if
        
        end if

        -- return match result
        return matchedActionToken

end function

---------------------------------------------------------------------------
function act_while( sequence arg )

    -- DO WHILE/UNTIL

        if action_is( "#Do" ) then
            -- do ... while|until test
            out( "while 1 do\n", {} )
            indent( 1 )

        elsif action_is( "#LoopWhile" ) then
            out( "\tif not " & ePop() & " then  exit  end if\n", {} )
            indent( -1 )
            out( "end while\n", {} )
            
        elsif action_is( "#LoopUntil" ) then
            out( "\tif " & ePop() & " then  exit  end if\n", {} )
            indent( -1 )
            out( "end while\n", {} )            

        elsif action_is( "#DoWhile" ) then
            out( "while " & ePop() & " do\n", {} )
            indent( 1 )

        elsif action_is( "#DoUntil" ) then
            out( "while not " & ePop() & " do\n", {} )
            indent( 1 )

        elsif action_is( "#EndDo" ) then
            indent( -1 )
            out( "end while\n", {} )            

        end if
        
        -- return match result
        return matchedActionToken

end function

---------------------------------------------------------------------------
function act_comment( sequence arg )


    -- COMMENTS

        if action_is( "#Comment" ) then    
            -- write comment
            -- out( "--%s\n", {arg[2..length(arg)]} )
            
        end if

        -- return match result
        return matchedActionToken

end function

---------------------------------------------------------------------------
function act_go( sequence arg )

        
    -- GOTO/GOSUB
            
        if action_is( "#Goto" ) then
            -- exit function, set goto
            
            -- make sure a colon is on the end
            if last(arg) != ':' then
                arg = arg & ':'
            end if
            
            -- set goto and exit
            out( "goto = %s  return{}\n", {label_name(arg)} )
            
        elsif action_is( "#Gosub" ) then
            -- call sub, check for return values
            
            -- call sub            
            out( "gosub = %s\n", {label_name(arg)} )
            out( "%s( {}, {} )\n", {proc_name(subName)} )
            
            -- check for result of RETURN [label]
            if compare( subType, "function" ) = 0 then
                out( "if goto then  return {}  end if\n", {} )
            else            
                out( "if goto then  return  end if\n", {} )
            end if

        elsif action_is( "#ReturnLabel" ) then
            -- exit function, return to caller
            if compare( subType, "function" ) = 0 then
                out( "goto = %s  return {}\n", {label_name(arg)} ) 
            else
                out( "goto = %s  return \n", {label_name(arg)} ) 
            end if
            

        elsif action_is( "#Return" ) then
            -- exit function, return to caller            
            if compare( subType, "function" ) = 0 then
                out( "return {}\n", {label_name(arg)} ) 
            else
                out( "return \n", {label_name(arg)} ) 
            end if
            
        end if
        
        -- return match result
        return matchedActionToken

end function

---------------------------------------------------------------------------
function act_data( sequence arg )

    -- READ DATA
    
        if action_is( "#Read" ) then        
            -- read data
            
            if  dataType = STRING then
                -- string read
                var_assign( ePop(), "qb_read()" )
            else            
                -- numeric read
                var_assign( ePop(), "qb_val(qb_read())" )
            end if

        end if
                                    
        -- return match result
        return matchedActionToken

end function


---------------------------------------------------------------------------
function act_file( sequence arg )

    -- FILE ACCESS
        if action_is( "#FileOpen{" ) then        
            -- { name, mode, access, number, len }
            push( { ePop(), "b", "u", "1", "128" } )
        

        elsif action_is( "#AppendMode" ) then
            -- file access mode
            tmp = pop()
            tmp[2] = "a"
            push( tmp )

        elsif action_is( "#BinaryMode" ) then
            -- file access mode
            tmp = pop()
            tmp[2] = "b"
            push( tmp )

        elsif action_is( "#InputMode" ) then
            -- file access mode
            tmp = pop()

            tmp[2] = ""

            -- set mode based on compatibility
            if FileReadMode = FAST then
                -- read as text file with gets()
                tmp[3] = "r"
            else
                -- read as binary 1 char at a time with getc()
                tmp[3] = "rb"
            end if

            -- save results
            push( tmp )            

        elsif action_is( "#OutputMode" ) then
            -- file access mode
            tmp = pop()
            tmp[2] = ""
            tmp[3] = "w"
            push( tmp )            

        elsif action_is( "#ReadAccess" ) then
            -- file access mode
            tmp = pop()
            tmp[3] = "r"
            push( tmp )

        elsif action_is( "#WriteAccess" ) then
            -- file access mode
            tmp = pop()
            if compare( tmp[3], "r" ) = 0 then
                tmp[3] = "u"
            else
                tmp[3] = "w"
            end if
            push( tmp )

        elsif action_is( "#FileNumber" ) then
            -- file number
            tmp = pop()
            tmp[4] = arg
            push( tmp )


        elsif action_is( "#FileLen" ) then
            -- buffer size
            tmp = pop()
            tmp[5] = arg
            push( tmp )

        elsif action_is( "#}FileOpen" ) then
            -- validate file number
            tmp = pop()
            
            -- code it
            out( "file_open( %s, \"%s%s\", %s )\n", tmp[1..4] )

        elsif action_is( "#CloseOne" ) then        
            -- close the given file
        
            -- instruction            
            out( "file_close( %s )\n", {arg} )      
            
        elsif action_is( "#CloseAll" ) then                    
        
            -- instruction            
            out( "file_close_all()\n", "" )
                        
        end if
        
        -- return match result
        return matchedActionToken

end function


---------------------------------------------------------------------------
function act_etc( sequence arg )

    -- NAME <STRING> AS <STRING>
    
        if action_is( "#Name" ) then        
            -- rename
            eSwap()
            out( "system( \"rename \" & %s & \" \" & %s, 2 )\n",
                    {ePop(), ePop()} )

    -- DEF SEG [= <EXPR>]

        elsif action_is( "#DefSeg" ) then        
            -- define segment
            out( "qb_def_seg( %s )\n", ePop() )

        elsif action_is( "#DefSegZero" ) then        
            -- define segment
            out( "qb_def_seg( 0 )\n", {} )

    -- LOCATE [<EXPR>], [<EXPR>]

        elsif action_is( "#Locate{" ) then
            -- push 2 nulls
            push( {"-1", "-1", "1" } )

        elsif action_is( "#LocateRow" ) then
            -- save row
            tmp = pop()
            tmp[1] = ePop()
            push( tmp )

        elsif action_is( "#LocateCol" ) then
            -- save column
            tmp = pop()
            tmp[2] = ePop()
            push( tmp )

        elsif action_is( "#LocateVisible" ) then
            -- save visible flag
            tmp = pop()
            tmp[3] = ePop()
            push( tmp )


        elsif action_is( "#}Locate" ) then
            -- set up call
            tmp = pop()
            out( "qb_locate( %s, %s, %s )\n", tmp )

    -- MID$( <STR_VAR>, <START> [, <LONG>] ) = <STR_EXPR>

        elsif action_is( "#Mid{" ) then
            -- push { var, start, long }
            push( { "", "", "" } )

        elsif action_is( "#MidVar" ) then
            -- set varriable
            tmp = pop()
            tmp[1] = ePop()
            push( tmp )


        elsif action_is( "#MidStart" ) then
            -- set start
            tmp = pop()
            tmp[2] = ePop()
            push( tmp )

        elsif action_is( "#MidLong" ) then
            -- set long
            tmp = pop()
            tmp[3] = ePop()
            push( tmp )

        elsif action_is( "#}Mid" ) then
            -- create expression
            tmp = pop()     -- mid$ args

            -- 3 args?
            if length( tmp[3] ) then
                -- 3 args
                out( "%s = qb_assign_mid3( %s, %s, %s, %s)\n",
                        {tmp[1],tmp[1],tmp[2],tmp[3],ePop()} )

            else
                -- 2 args
                out( "%s = qb_assign_mid2( %s, %s, %s )\n",
                        {tmp[1],tmp[1],tmp[2],ePop()} )

            end if

    -- SWAP <VALUE>, <VALUE>
    
        elsif action_is( "#Swap" ) then        
            tmp[1] = ePop()
            tmp[2] = ePop()
            out( "tmp = %s\n", {tmp[1]} )
            out( "%s = %s\n", {tmp[1],tmp[2]} )
            out( "%s = tmp\n", {tmp[2]} )


        end if

        -- return match result
        return matchedActionToken

end function




---------------------------------------------------------------------------
global procedure action( sequence token, sequence arg )

    -- handle actions from parser
    
    -- set up global
    act = token
    
    -- flag no match
    matchedActionToken = 0

-- printf( traceFile, "Action %s  - Arg %s  - Type is %d\n", {token,arg,dataType} )
-- on_screen( "%s[ %s ] eStack:%d stack:%d\n", {act,arg,length(eStack),length(stack)} ) )
-- show_seq( eStack )

    if    act_print( arg )      then
    elsif act_input( arg )      then
    elsif act_expr( arg )       then
    elsif act_array( arg )      then
    elsif act_shared( arg )     then
    elsif act_sub( arg )        then
    elsif act_label_name( arg )      then
    elsif act_end( arg )        then
    elsif act_select( arg )     then
    elsif act_while( arg )      then
    elsif act_exit( arg )       then
    elsif act_subs( arg )       then
    elsif act_wend( arg )       then
    elsif act_comment( arg )    then
    elsif act_go( arg )         then
    elsif act_file( arg )       then
    elsif act_data( arg )       then
    elsif act_etc( arg )       then
    else
        errText = "Unknown internal command: " & act
    end if

end procedure

-----------------------------------------------------------------------------
function read_file( sequence fName )

    -- read a file, return contents in sequence
    integer handle
    sequence buffer
    object line

    -- open the file
    handle = open( fName, "r" )
    if handle = -1 then
        errText = "Unable to open file \"" & fName & "\"."
        return {}
    end if

    -- read a text file into a sequence
    buffer = {}
    while 1 do
        line = gets(handle)
        if atom(line) then
            exit   -- end of file
        end if
        buffer = append(buffer, line)
    end while

    -- close the file
    close( handle )

    -- return the file contents    
    return buffer

end function

-----------------------------------------------------------------------------
function data_to_seq( sequence s )           

    -- convert a data statement to a sequence
    integer start, n, quote
    sequence word, words
    

    -- move past "DATA"
    start = match( "data", lower( s ) ) + 5
    
    -- init
    word = ""
    words = {}
    quote = 0 
    
    -- parse    
    for i = start to length( s )-1 do

        if s[i] = '"' then 
            quote = not quote            
        elsif quote then 
            word = word & s[i]
        elsif s[i] = ' ' or s[i] = 8 then
            -- ignore            
        elsif s[i] = ',' then 
            words = append( words, word )
            word = ""         
        elsif s[i] = '\\' then
            word = word & "\\"
        else 
            word = word & s[i]            
        end if            
    end for                      
    
    -- last word
    words = append( words, word )

    -- reconstruct
    s = ""
    for i = 1 to length( words ) do
        s = s & "\"" & words[i] & "\", "
    end for 

    return s
    
end function


-----------------------------------------------------------------------------
function text_to_type( sequence s )

    -- return data type for text name
    -- only works with pre-defined data types

    s = lower( s )

    -- find type
       if compare( s, "integer" ) = 0 then      return INTEGER
    elsif compare( s, "long" ) = 0 then         return LONG
    elsif compare( s, "single" ) = 0 then       return SINGLE
    elsif compare( s, "double" ) = 0 then       return DOUBLE
    elsif compare( s, "string" ) = 0 then       return STRING
    end if

    -- no matches
    errText = "Expected a data type, not " & upper( s ) & "."
    return 0

end function

-----------------------------------------------------------------------------
sequence text, tokenList, firstToken

-- initialize
    -- clear error flag
        errText = ""
    
    -- init stacks
        stack = {}  
        eStack = {}
        typeStack = {}
    
    -- temp variable counter
        tmpCount = 0    
        
    -- loop index stack
        loopIndex = {}
   
    -- initialize the deferring
        deferFlag = 0
        
    -- no lables encountered yet
        labelFlag = 0
        
    
    -- initialize the data structures
    userSubs = {}                   -- user defined subs
    userFuncs = {}                  -- user defined functions
    labels = { { "main" } }         -- labels
    types = {}                      -- user defined data types
    elements = {}                   -- elements of types    
    typeName = "" 
    uses = { {"main", {}, {}, {}} }     -- stuff called
    
    -- clear sub name
    subName = "main"
    subType = "procedure"

    -- clear label numbering          
    labelSubCount = 100
    labelGotoCount = labelSubCount + 1
    
    procLabels = { {"main", labelSubCount} }
    gotoLabels = { {"main_setup:", labelGotoCount} }
    
    -- trace file title
    printf( traceFile, "File Name:%s\n\n", {inFile} )

-- read the file
    text = read_file( inFile )
    -- error?
    if length( errText ) > 0 then
        abort_error( "Read Error", errText )        
    end if


--
--
-- PASS 1:  Scan for labels, subs and functions
--
--

    text_color( BRIGHT_WHITE )
    on_screen( "PASS 1: Scanning the file.\n", {} )
    text_color( WHITE )

    for i = 1 to length( text ) do
    
        -- current line, for error message
        abortLine = i
        abortCode = text[i]
    
        -- tokenize
        tokenList = tokenize( text[i] )

        -- sub?
        if length( tokenList ) > 1 then              
        
            -- get first token
            firstToken = lower( tokenList[1] )

            -- new subroutine?
            if compare( firstToken, "sub" ) = 0 then

                -- get name
                subName = tokenList[2]

                userSubs = append( userSubs, subName )
                labels = append( labels, { subName } )        
                
                -- update label counts
                labelSubCount = labelSubCount + 100
                labelGotoCount = labelSubCount + 1

                -- add labels                
                procLabels = append( procLabels, { subName, labelSubCount } )
                gotoLabels = append( gotoLabels, { subName & "_setup:", labelGotoCount } )
                
            -- new function?
            elsif compare( firstToken, "function" ) = 0 then

                -- get name, less suffix
                subName = remove_numeric_suffix( tokenList[2] )

                userFuncs = append( userFuncs, subName )
                labels = append( labels, { subName } )
                
                -- update label counts
                labelSubCount = labelSubCount + 100
                labelGotoCount = labelSubCount + 1

                -- add labels                
                procLabels = append( procLabels, { subName, labelSubCount } )
                gotoLabels = append( gotoLabels, { subName & "_setup:", labelGotoCount } )
                
            end if

        end if

        -- label?
        if length( tokenList ) > 0 then

            -- get first token
            firstToken = tokenList[1]
            
            -- label? ends in colon & not a comment
            if  firstToken[length(firstToken)] = ':'    -- ends in colon
            and firstToken[1] != '\'' then              -- not a comment
            
                -- add to the list
                labels[length(labels)] =
                    append( labels[length(labels)], firstToken )
                    
                -- update label counts & label
                labelGotoCount = labelGotoCount + 1
                gotoLabels = append( gotoLabels, { firstToken, labelGotoCount } )
                
                    
                        
            end if
        end if

        -- start of new type definition?
        if length( tokenList ) > 0 then
            if compare( lower( tokenList[1] ), "type" )=0 then

                -- save name
                typeName = tokenList[2]
                typeDefault = {}
                elementIndex = 0

            end if
        end if

        -- new type?
        if length( tokenList ) > 2 then
            if compare( lower( tokenList[2] ), "as" )=0 then

                -- build new structure
                tmp = repeat( {}, 5 )

                -- increment the index
                elementIndex = elementIndex + 1

                -- fill in
                tmp[E_NAME]     = tokenList[1]
                tmp[E_INDEX]    = elementIndex
                tmp[E_TYPE]     = text_to_type( tokenList[3] )                
                tmp[E_OWNER]    = length( types ) + 1         
                
                -- data size?
                if tmp[E_TYPE] = STRING then
                    -- check grammar
                    if compare( tokenList[4], "*" ) != 0 then
                        errText = "Expected \"*\"."
                    else
                        -- convert the data size
                        tmp1 = value( tokenList[5] )
                        if tmp1[1] != GET_SUCCESS then
                            -- failed
                            errText = "Expected a number, not " &
                                tokenList[5] & "."
                        else
                            -- got number
                            dataSize = tmp1[2]
                            tmp[E_SIZE] = dataSize                            
                        end if
                    end if
                    
                else
                    -- zero
                    tmp[E_SIZE] = 0
                end if

                -- default type

                -- comma?
                if length( typeDefault ) then
                    typeDefault = typeDefault & ", "
                end if

                -- string or numeric
                if tmp[E_TYPE] = STRING then

                    -- string
                    typeDefault = typeDefault 
                        & "\"" & repeat( ' ', dataSize ) & "\""
                else
                    -- number
                    typeDefault = typeDefault & "0"
                end if

                -- append to elements
                elements = append( elements, tmp )

            end if
        end if

        -- end of type definition?
        if length( tokenList ) > 2 then
            if  compare( lower( tokenList[1] ), "end" )=0
            and compare( lower( tokenList[2] ), "type" )=0 then

                -- add braces to typeDefault
                typeDefault = "{ " & typeDefault & " }"

                -- append type header to elements
                types = append( types, { typeName, typeDefault } )

                -- clear type name
                typeName = ""

            end if
        end if

        -- error?
        if length( errText ) > 0 then
            -- line number and text
            text_color( BRIGHT_WHITE )
            on_screen( "LINE %d: %s", {i,text[i]} )
            -- leave loop
            exit
        end if

    end for

    -- closed all types?
    if length( typeName ) 
    and length( errText ) = 0 then
        errText = "END TYPE not found for " & typeName & "."
    end if

    -- error?
    if length( errText ) > 0 then
        abort_error( "Tokenize Error", errText )
    end if


    -- results of pass 1

        puts( traceFile, "Results of Pass 1:\n" )
        puts( traceFile, "------------------\n\n" )

        -- subs
        printf( traceFile, "There are %d User Subs\n", length( userSubs ) )
        for i = 1 to length( userSubs ) do
            printf( traceFile, "\t%s\n", {userSubs[i]} )
        end for
        puts( traceFile, "\n" )

        -- functions
        printf( traceFile, "There are %d User Functions\n", length( userFuncs ) )
        for i = 1 to length( userFuncs ) do
            printf( traceFile, "\t%s\n", {userFuncs[i]} )
        end for
        puts( traceFile, "\n" )

        -- types
        printf( traceFile, "There are %d Types.\n", {length(types)} )
        for i = 1 to length( types ) do
            printf( traceFile, "\t%s\n", {types[i][T_NAME]} )
            for j = 1 to length( elements ) do
                -- get owner's name
                if compare( elements[j][E_OWNER], i ) = 0 then
                    printf( traceFile, "\t\t%s\n", {elements[j][E_NAME]} )
                end if
            end for
        end for

        -- labels
        tmp = 0
        for i = 1 to length( labels ) do
            if length( labels[i] ) > 1 then
                tmp = tmp + 1
            end if
        end for

        printf( traceFile, "There are %d Labels\n", tmp )
        for i = 1 to length( labels ) do
            if length( labels[i] ) > 1 then
                printf( traceFile, "\t%s\n", {labels[i][1]} )
                for j = 2 to length(labels[i]) do
                    printf( traceFile, "\t\t%s\n", {labels[i][j]} )
                end for
            end if
        end for
        puts( traceFile, "\n" )

--
--
--  PASS 2: Generate Code 
--
--

    text_color( BRIGHT_WHITE )    
    on_screen( "PASS 2: Generating code...\n\n", {} )
    text_color( WHITE )

    -- clear sub name
    subName = "main"
   
    -- file header     
    indents = 0
    out( "-- QBasic program %s\n", {inFile} )
    out( "-- Generated with eBasic %s\n", {VERSION} )
    out( "--   (c) 1997 David Cuny\n", "" )
    out( "--   Hardware Calls (c) Jacques Deschenes\n", {} )
    out( "--   PC Music (c) J. Craig Gilbert\n", {} )
    out( "\n", "" )

    out( "without warning\n", {} )
    out( "without trace\n", {} )
    out( "without type_check\n", {} )

    -- array mode: fast or compatible
    if ArrayMode = FAST then
        -- assumes arrays start at zero
        out( "global constant FAST_ARRAYS = -1\n", {} )
    else
        -- arrays can start at any position
        out( "global constant FAST_ARRAYS = 0\n", {} )
    end if
    out( "\n", {} )

    out( "without warning\n", "" )
    out( "include io.e\n", "" )    
    out( "include ebasic.e\n", "" )
    out( "\n", "" )
    out( "object returnValue\n", "" )
    out( "\n", "" )
    out( "-- initialize return value\n", {} )
    out( "returnValue = {}\n", "" )
    out( "\n", "" )
    
    -- label names     
    tmp = 0         -- label count
    out( "-- labels\n", {} )    
    indent(1)
    for i = 1 to length( labels ) do
        -- does this procedure contain labels?
        if length( labels[i] ) > 1 then
            -- setup label  
            tmp = tmp + 1  
            out( "constant %s_setup = %d\n", {label_name(labels[i][1]),tmp} )

            -- 'real' labels            
            for j = 2 to length(labels[i]) do
                tmp = tmp + 1
                out( "constant %s = %d\n", {label_name(labels[i][j]),tmp} )                
            end for
        end if        
    end for
    indent(-1)
    out( "\n", {} )

    -- start of code
    deferFlag = 1    
    out( "procedure proc_main( sequence args, sequence passBacks )\n", "" )
    out( "\n", "" )
    out( "\n", "" )
    indent(1)
                
    
    -- label handling code
    build_header( {} )

    -- determine token types
    for i = 1 to length( text ) do

        -- for error message
        abortLine = i
        abortCode = text[i]

        -- add QBasic as comment        
        add_qb_as_comment( text[i] )

        -- tokenize again
        tokenList = tokenize( text[i] )

        --
        -- SPECIAL: Pre-process before the parser gets to it
        --

            -- data?
            if compare( lower(tokenList[1]), "data" ) = 0 then
                -- move
                data = append( data, data_to_seq( text[i] ) )
                -- clear it
                tokenList = ""
            end if


        -- determine types
        tokenList = add_token_types( tokenList )

        -- display current line
        if length( text[i] ) > 79 then
            -- clip
            on_screen( text[i][1..79] & "\n", {} )
        else
            on_screen( text[i], {} )
        end if
        
        -- clear actions
            actions = {}

        -- start parsing at rule 1
        
        thisToken = 1
        tokens = tokenList
        
        while thisToken < length( tokenList )
        and length( errText ) = 0 do
            parse( {"$instruction"}, 0 )
        end while
        
        -- parse error?
        if length( errText ) > 0  then

            -- line number and text
            text_color( BRIGHT_WHITE )
            on_screen( "LINE %d: %s", {i,text[i]} )

            -- error message
            abort_error( "Parse Error", errText )
        end if

        -- clear the data type
        dataType = 0

        -- generate code
        for j = 1 to length( actions ) do

            -- action
            action( actions[j][1], actions[j][2] )
            
            -- compile error?
            if length( errText ) > 0  then
                -- line number and text
                text_color( BRIGHT_WHITE )
                on_screen( "LINE %d: %s", {i,text[i]} )

                -- message
                abort_error( "Compiler Error", errText )
            end if
            
        end for

    end for
            

-- end of main loop    
    end_proc()     
    
    indent(-1)    
    indents = 0    
    deferFlag = 0
    
-- check indent/nesting
    if indents > 0 then
        abort_error( "Nesting Error", "Too few END clauses." )
    end if

-- sort based on dependancies
    for i = 1 to length( uses ) do 
        for j = 1 to length( uses )-1 do
            -- does prior routine NOT use this?
            if not find( uses[j][U_NAME], uses[j+1][U_LIST] ) then
                -- swap then
                tmp1 = uses[j]
                uses[j] = uses[j+1]
                uses[j+1] = tmp1
            end if        
        end for
    end for

-- output the code    
    for i = 1 to length( uses ) do
        -- sub/function
        for j = 1 to length( uses[i][U_CODE] ) do
            -- out has already split the lines 
            puts( outFile, uses[i][U_CODE][j] )
        end for
    end for

-- data statement         
    if length( data ) then
        out( "\n", {} )
        indent(1)
        out( "-- data\n", {} )
        out( "data = {\n", {} )
        indent( 1 )
        for i = 1 to length( data ) do
            -- data
            if i = length( data ) then
                out( data[i] & "\"END OF DATA\" }\n", {} )
            else
                out( data[i] & "\n", {} )
            end if
        end for

        indent(-2)
        out( "\n\n", {} )
    end if    

-- call main
    out( "-- call main\n", {} )
    out( "proc_main( {}, {} )\n", {} )
    
    out( "-- clean up\n", {} )
    out( "set_text_mode()\n", {} )

-- were all dependancies successfully resolved?
    for i = length( uses ) to 2 by - 1 do 
        for j = i-1 to 1 by -1 do
            -- does prior routine use this?
            if find( uses[i][U_NAME], uses[j][U_LIST] ) then
                -- forward reference
                abort_error( "Compile Error",
                                "Illegal forward reference to " &
                                    uses[j][U_NAME] & 
                                    " in " & uses[i][U_NAME] & "." )
            end if        
        end for
    end for


-- report success    
    abortLine = 0    
    abort_error( "Complete!", "eBasic finished successfully." )



