-- GRAMMAR.E
-- Parsing code for eBasic Translator
-- (c) 1997 David Cuny

-----------------------------------------------------------------------------
global constant grammar = {

    -- one line
    {"$instruction",        "<eol>" },
    {"$instruction",        ":", "#ClearType", "$instruction" },
    {"$instruction",        "$command", "$instruction" },
    
    -- comment
    {"$command",            "%comment", "#Comment" },

    -- defint, deflng: ignore
    {"$command",            "defint", "$defRange" },
    {"$command",            "deflng", "$defRange" },
    {"$command",            "defsng", "$defRange" },
    {"$command",            "defdbl", "$defRange" },
    {"$command",            "defstr", "$defRange" },
    {"$defRange",           "%variable", "$defHighEnd" },
    {"$defHighEnd",         "-", "%variable" },
    {"$defHighEnd",         "NULL" },

    -- type
    {"$command",            "type", "%userType", "#Type{" },
    {"$command",            "%element", "as", "$dataType" },
    {"$element",            "%element", "#Element" },
    {"$element",            "NULL" },

    -- declare: ignore
    {"$command",            "declare", "$ignore" },
    {"$ignore",             "?END" },
    {"$ignore",             "ANY", "$ignore" },

    -- randomize
    {"$command",            "randomize", "$expr" },


    -- PRINT [#filenumber%,] [expressionlist] [{; | ,}]
    -- LPRINT [expressionlist] [{; | ,}]
    {"$command",            "print", "$printTo", "$printUsing" },
    {"$command",            "lprint", "#LPrint", "$printUsing" },    
    {"$printTo",            "#", "%number", "#PrintTo", "," },
    {"$printTo",            "NULL", "#PrintScreen" },
    {"$printUsing",         "using", "$expr", "#PrintUsing{", ";", "$printUsingItem" },
    {"$printUsing",         "$printTail" },
    {"$printTail",          ",", "#PrintComma", "$printTail" },
    {"$printTail",          ";", "#NoCr", "$printTail" },
    {"$printTail",          "tab", "(", "$expr", "#Tab", ")", "$printTail" },
    {"$printTail",          "?EXPR", "#ClearType", "$expr", "#Print", "$printTail" },
    {"$printTail",          "NULL", "#EndPrint" },

    {"$printUsingItem",     "?EXPR", "$expr", "#PrintUsingArg", "$printUsingTail" },
    {"$printUsingItem",     "NULL", "#PrintUsingSemi", "#}PrintUsing" },      
    {"$printUsingTail",     ";", "$printUsingItem" },    
    {"$printUsingTail",     "NULL", "#}PrintUsing" },

    -- input command
    -- INPUT [;] ["prompt"{; | ,}] variablelist
    -- INPUT #filenumber%, variablelist
    -- LINE INPUT #filenumber%, variable$
    -- LINE INPUT [;] ["prompt";] variable$

    {"$command",            "input", "#Input{", "$input" },
    {"$command",            "line", "input", "#LineInput{", "$input" },
    {"$input",              "#", "%number", "#InputFile", ",", "$inputVar" },
    {"$input",              "$inputSemi", "$inputPrompt", "$inputVar" },
    {"$inputSemi",          ";" },
    {"$inputSemi",          "NULL", "#InputCrFlag" },
    {"$inputPrompt",        "%string", "#InputPrompt", "$inputPromptDelim" },
    {"$inputPrompt",        "NULL" },
    {"$inputPromptDelim",   ",", "#InputNoQFlag" },
    {"$inputPromptDelim",   ";" },
    {"$inputVar",           "$variable", "#Input", "$inputVarTail" },
    {"$inputVarTail",       ",", "$inputVar" },
    {"$inputVarTail",       "NULL", "#}Input" },

    -- for loop

    {"$command",            "for", "%variable", "#MakeVar", "#GuessType", "#For{", "=", "$expr", "#ForStart", "to", "$expr", "#ForTo", "$step" },
    {"$step",               "step", "$expr", "#ForStep", "$stepTail" },
    {"$step",               "by", "$expr", "#ForStep", "$stepTail" },
    {"$step",               "$stepTail" },
    {"$stepTail",            "do", "#}For" },
    {"$stepTail",            "NULL", "#}For" },

    -- next
    {"$command",            "next", "#Next", "$nextTail" },
    {"$nextTail",           "%variable" },
    {"$nextTail",           "NULL" },    

    -- do while|until test ... loop
    -- do ... loop while|until test
    {"$command",            "do", "$do" },
    {"$command",            "loop", "$loop" },
    {"$do",                 "while", "$expr", "#DoWhile" },
    {"$do",                 "until", "$expr", "#DoUntil" },
    {"$do",                 "NULL", "#Do" },    
    {"$loop",               "while", "$expr", "#LoopWhile" },
    {"$loop",               "until", "$expr", "#LoopUntil" },
    {"$loop",               "NULL", "#EndDo" },

    -- if command
    {"$command",            "if", "$expr", "then", "#If", "$ifThen" },
    -- this is a patch, because ?END sees END as an end
    {"$ifThen",             "end", "BACKUP", "$ifThenTail" },
    {"$ifThen",             "?END" },
    {"$ifThen",             "NULL", "$ifThenTail" },
    {"$ifThenTail",         "else", "#Else", "$elseTail" },
    -- these patches come before ?END, so the ":" and "END" get processed
    -- as part of the code, and not an end of the clause
    {"$ifThenTail",         ":", "$ifThenTail" },
    {"$ifThenTail",         "end", "BACKUP", "$command", "$ifThenTail" },    
    {"$ifThenTail",         "?END", "#EndIf" }, 
    {"$ifThenTail",         "$command", "$ifThenTail" },
    {"$elseTail",           "?END", "#EndIf" },
    {"$elseTail",           "$command", "$elseTail" },
    {"$command",            "elseif", "$expr", "then", "#ElseIf"  },
    {"$command",            "elsif", "$expr", "then", "#ElseIf" },
    {"$command",            "else", "#Else" },    
    {"$elseTail",           "if", "$expr", "then", "#ElseIf" },
    {"$elseTail",           "NULL", "#Else" },

    -- while command
    {"$command",            "while", "$expr", "#While" },
    {"$command",            "wend", "#Wend" },

    -- case command
    {"$command",            "select", "case", "$expr", "#CaseStart" },    
    {"$command",            "case", "$case" },
    {"$case",               "else", "#CaseElse" },
    {"$case",               "NULL", "$caseExpr" },
    {"$caseExpr",           "is", "$caseIs" },
    {"$caseExpr",           "$expr", "$caseOptionalTo" },
    {"$caseOptionalTo",     "to", "$expr", "#CaseTo", "$caseTail" },
    {"$caseOptionalTo",     "NULL", "#CaseEq", "$caseTail" },
    {"$caseTail",           ",", "$caseExpr" },
    {"$caseTail",           "NULL", "#Case" },

    --  expression[,expression]...
    --  expression TO expression
    --  IS relational-operator expression
    --      <, <=, >, >=, <>, or =    
    {"$caseIs",             "<", "$expr", "#CaseLt", "$caseTail" },
    {"$caseIs",             ">", "$expr", "#CaseGt", "$caseTail" },
    {"$caseIs",             "=", "$expr", "#CaseEq", "$caseTail" },
    {"$caseIs",             ">=", "$expr", "#CaseGe", "$caseTail" },
    {"$caseIs",             "<=", "$expr", "#CaseLe", "$caseTail" },
    {"$caseIs",             "!=", "$expr", "#CaseNe", "$caseTail" },

    -- read [variable]      
    {"$command",            "read", "$readVar" },
    {"$readVar",            "$variable", "#Read", "$readTail" },
    {"$readTail",           ",", "$readVar" },
    {"$readTail",           "NULL" },

    -- exit command
    {"$command",            "exit", "$exit" },
    {"$exit",               "do", "#Exit" },
    {"$exit",               "for", "#Exit" },
    {"$exit",               "function", "#ExitProc" },
    {"$end",                "select", "#EndSelect" },    
    {"$exit",               "sub", "#ExitProc" },


    -- SUB name[(parameterlist)] [STATIC]
    {"$command",            "sub", "%userSub", "#UserName{", "$parmList", "$parmStatic", "#}UserName" },

    {"$parmList",           "(", "$parmListItems", ")" },
    {"$parmList",           "NULL" },
    {"$parmListItems",      "$parmVar", "#UserParm", "$parmMore" },

    {"$parmMore",           ",", "$parmListItems" },
    {"$parmMore",           "NULL" },
    
    {"$parmStatic",         "static", "#StaticOn" },
    {"$parmStatic",         "NULL", "#StaticOff" },
    

    {"$dataType",           "integer", "#Integer" },
    {"$dataType",           "long", "#Long" },
    {"$dataType",           "single", "#Single" },
    {"$dataType",           "double", "#Double" },
    {"$dataType",           "string", "$dataSize", "#String" },
    {"$dataSize",           "*", "%number", "#StringSize" },
    {"$dataSize",           "NULL", "#StringSizeZero" },    
    {"$dataType",           "%userType", "#UserType" },



    -- FUNCTION name[(parameterlist)] [STATIC]
    {"$command",            "function", "$userFuncDef" },
    {"$userFuncDef",        "%userFunc", "#UserName{", "$parmList", "$parmStatic", "#}UserName" },
    {"$userFuncDef",        "%userStrFunc", "#UserName{", "$parmList", "$parmStatic", "#}UserName" },


    -- dim statement: 
    -- DIM %variable( %number [ to %number ]{%number [ to %number ]} ), ...
    {"$command",            "dim", "$dimShared" },    
    {"$dimShared",          "shared", "#SharedOn", "$dimList" },
    {"$dimShared",          "NULL", "$dimList" },
    {"$dimList",            "%variable", "#GuessType", "#Dim{", "$dimSize" },
    {"$dimSize",            "(", "$dimRange", "$dimListTail" },
    {"$dimSize",            "NULL", "$dimAs", "#}Dim", "$dimTail" },
    {"$dimListTail",        ",", "$dimRange", "$dimListTail" },
    {"$dimListTail",        ")", "$dimAs", "#}Dim", "$dimTail" },
    {"$dimTail",            ",", "$dimList" },
    {"$dimTail",            "NULL", "#SharedOff" },
    {"$dimRange",           "$expr", "$dimTo" },
    {"$dimTo",              "to", "$expr", "#DimRange" },
    {"$dimTo",              "NULL", "#DimHigh" },
    {"$dimAs",              "as", "$dataType", "#DimAs" },
    {"$dimAs",              "NULL" },    

    -- common declaration
    -- only COMMON SHARED is supported
    {"$command",            "common", "shared", "$commonList" },
    {"$commonList",         "%variable", "#GuessType", "#Common{", "$commonParens" },
    {"$commonParens",       "(", ")", "#CommonArray", "#}Common", "$commonTail" },
    {"$commonParens",       "NULL", "#CommonVar", "#}Common", "$commonTail" },
    {"$commonTail",         ",", "$commonList" },
    {"$commonTail",         "NULL" },

    -- shared declaration
    {"$command",            "shared", "$sharedList" },
    {"$sharedList",         "%variable", "#GuessType", "#Shared{", "$sharedParens" },
    {"$sharedParens",       "(", ")", "#SharedArray", "#}Shared", "$sharedTail" },
    {"$sharedParens",       "NULL", "#SharedVar", "#}Shared", "$sharedTail" },
    {"$sharedTail",         ",", "$sharedList" },
    {"$sharedTail",         "NULL" },

    -- constant declaration
    {"$command",            "const", "$constList" },
    {"$constList",          "%variable", "#Const{", "=", "$expr", "#}Const", "$constTail" },
    {"$constTail",          ",", "$constList" },
    {"$constTail",          "NULL" },

    
    -- end clauses    
    {"$command",            "end", "$end" },
    {"$end",                "sub", "#UserEnd" },
    {"$end",                "function", "#UserEnd" },
    {"$end",                "if", "#EndIf" },
    {"$end",                "for", "#Next" },
    {"$end",                "type", "#}Type" },
    {"$end",                "NULL", "#End" },

    -- CALL user defined sub
    {"$command",            "call", "%userSub", "#Sub{", "$callArgList" },
    {"$command",            "%userSub", "#Sub{", "$subArgList", "#}UserSub"},

    -- call requires parens, or nothing
    {"$callArgList",        "(", "$subArgList", ")", "#}UserSub" },
    {"$callArgList",        "NULL", "#}UserSub" },    
    
    {"$subArgList",         "?END" },
    {"$subArgList",         "(", "$expr", "#StaticArg", ")", "$subArgMore" },
    {"$subArgList",         ",", "#MissingArg", "$subArgList" },
    {"$subArgList",         "$expr", "#Arg", "$subArgMore" },
    {"$subArgMore",         ",", "$subArgList" },
    {"$subArgMore",         "NULL" },


    -- call builtin subs
    {"$command",            "%builtinSub", "#Sub{", "$subArgList", "#}Sub" },
    

    -- goto and gosub..return
    {"$command",            "%label", "#Label" },
    {"$command",            "goto", "%variable", "#Goto" },
    {"$command",            "gosub", "%variable", "#Gosub" },
    {"$command",            "return", "$return" },
    {"$return",             "%variable", "#ReturnLabel" },
    {"$return",             "NULL", "#Return" },
    
    -- OPEN file$ [FOR mode] [access] [lock] AS [#]filenumber% [LEN=reclen%]
    -- [lock] is parsed, but not supported
    {"$command",            "open", "$expr", "#FileOpen{", "$fileFor", "$fileAccess", "$fileLock", "$fileAs", "$fileLen", "#}FileOpen" },
    {"$fileFor",            "for", "$fileMode" },
    {"$fileFor",            "NULL" },
    {"$fileMode",           "append", "#AppendMode" },
    {"$fileMode",           "binary", "#BinaryMode" },
    {"$fileMode",           "input", "#InputMode" },
    {"$fileMode",           "output", "#OutputMode" },
    {"$fileMode",           "random", "#RandomMode" },
    {"$fileMode",           "NULL" },
    {"$fileAccess",         "$fileReadAccess", "$fileWriteAccess" },
    {"$fileReadAccess",     "read", "#ReadAccess" },
    {"$fileReadAccess",     "NULL" },
    {"$fileWriteAccess",    "write", "#WriteAccess" },
    {"$fileWriteAccess",    "NULL" },
    {"$fileLock",           "shared" },
    {"$fileLock",           "lock", "$fileLockRead", "$fileLockWrite" },
    {"$fileLock",           "NULL" },
    {"$fileLockRead",       "read" },
    {"$fileLockRead",       "NULL" },
    {"$fileLockWrite",      "WRITE" },
    {"$fileLockRead",       "NULL" },

    {"$fileAs",              "as", "$fileNumber" },
    {"$fileNumber",         "#", "%number", "#FileNumber" },
    {"$fileNumber",         "%number", "#FileNumber" },
    {"$fileLen",            "len", "=", "%number", "#FileLen" },
    {"$fileLen",             "NULL" },

    -- CLOSE [[#]filenumber%[,[#]filenumber%]...]
    {"$command",            "close", "$closeItem" },
    {"$closeItem",          "#", "%number", "#CloseOne", "$closeTail" },
    {"$closeItem",          "%number", "#CloseOne", "$closeTail" },
    {"$closeItem",          "?END", "#CloseAll" },            
    {"$closeTail",          ",", "$closeItem" },
    {"$closeTail",          "NULL" },

    -- optional variable indexes
    {"$index",              "(", "#PushType", "#DummyIndex", "$indexItem" },
    {"$index",              "NULL" },
    {"$indexItem",          ",", "$indexItem" },
    {"$indexItem",          ")", "#PopType" },
    {"$indexItem",          "$expr", "#Index", "$indexItem" },


    -- swap command
    {"$command",            "swap", "$expr", ",", "$expr", "#Swap" },

    -- on error goto {label|0} -- DUMMY
    {"$command",            "on", "error", "goto", "ANY" },

    -- resume {label} -- DUMMY
    {"$command",            "resume", "$resume" },    
    {"$resume",             "%variable" },
    {"$resume",             "next" },
    {"$resume",             "NULL" },

    -- option base <number> -- DUMMY
    {"$command",            "option", "base", "%number" },

    -- NAME str$ as str$
    -- note that basicExpr is used instead of expr because 'as' confuses
    -- the parser
    {"$command",            "name", "$basicExpr", "as", "$basicExpr", "#Name" },

    -- DEF SEG expr
    {"$command",            "def", "seg", "$defSeg" },
    {"$defSeg",             "=", "$expr", "#DefSeg" },
    {"$defSeg",             "NULL", "#DefSegZero" },

    -- LOCATE  [row%] [,[column%] [,[cursor%] [,start% [,stop%]]]]
    {"$command",            "locate", "#Locate{", "$locateRow" },
    {"$locateRow",          "?EXPR", "$expr", "#LocateRow", "$locateCol" },
    {"$locateRow",          "$locateCol" },
    {"$locateCol",          ",", "$locateOptCol" },
    {"$locateCol",          "NULL", "#}Locate" },
    {"$locateOptCol",       "?EXPR", "$expr", "#LocateCol", "$locateVisible" },
    {"$locateOptCol",       "NULL", "$locateVisible" },
    {"$locateVisible",      ",", "$expr", "#LocateVisible", "#}Locate" },
    {"$locateVisible",      "NULL", "#}Locate" },

    -- MID$( var, start [, long] ) = expr
    {"$command",            "mid$", "#Mid{", "(", "$midVar" },
    {"$midVar",             "$expr", "#MidVar", ",", "$midStart" },
    {"$midStart",           "$expr", "#MidStart", "$midLong" },
    {"$midLong",            ")", "$midSubStr" },
    {"$midLong",            ",", "$expr", "#MidLong", ")", "$midSubStr" },
    {"$midSubStr",          "=", "$expr", "#}Mid" },

    -- assignment
    {"$command",            "%variable", "BACKUP", "$assign" },
    {"$command",            "let", "%variable", "BACKUP", "$assign" },
    {"$assign",             "$variable", "=", "#Assign{", "$expr", "#}Assign" },
    {"$command",            "%variable", "BACKUP", "$variable", "=", "#Assign{", "$expr", "#}Assign" },

    -- variable handler. takes care of arrays, and indexes
    {"$variable",           "%variable", "#PushExpr", "#Variable{", "$index", "$element", "#}Variable" },


    -- like variable, but includes "as"
    {"$parmVar",            "%variable", "#PushExpr", "#Variable{", "$index", "$element", "$parmType", "#}Variable" },
    {"$parmType",           "as", "$dataType" },
    {"$parmType",           "NULL" },


    -- this should only happen in the same function...
    {"$command",            "%userFunc", "#PushExpr", "#Assign{", "=", "$expr", "#}Assign" },
    {"$command",            "%userStrFunc", "#PushExpr", "#Assign{", "=", "$expr", "#}Assign" },    

    -- function arg list
    {"$argList",            "(", "$listItem" },
    {"$argList",            "NULL" },
    {"$listItem",           ")" },    
    {"$listItem",           "$expr", "#Arg", "$listTail" },
    {"$listTail",           ",", "$listItem" },
    {"$listTail",           ")" },
                
                
    -- basic expressions
    -- this is the initial call, so the expression is pushed onto the stack

    -- string expressions
    {"$basicExpr",          "%builtinStr", "#Func{", "$argList", "#}Func", "#StrType", "$optOperator" },
    {"$basicExpr",          "%userStrFunc", "#Func{", "$argList", "#}UserFunc", "#StrType", "$optOperator" },
    {"$basicExpr",          "%string", "#PushExpr", "#StrType", "$optOperator" },

    -- numeric expressions
    {"$basicExpr",          "not", "$basicExpr", "#Not", "#NumType", "$optEquality" },
    {"$basicExpr",          "%builtinNum", "#Func{", "$argList", "#}Func", "#NumType", "$optOperator", "$optEquality" },
    {"$basicExpr",          "%userFunc", "#Func{", "$argList", "#}UserFunc", "#NumType", "$optOperator", "$optEquality" },
    {"$basicExpr",          "%number", "#PushExpr", "#NumType", "$optOperator", "$optEquality" },
    {"$basicExpr",          "%variable", "BACKUP", "$variable", "$optOperator", "$optEquality" },
    {"$basicExpr",          "(", "$expr", ")", "#Parens", "#NumType", "$optOperator", "$optEquality"  },
    {"$basicExpr",          "-", "$basicExpr", "#UnaryMinus", "#NumType", "$optOperator", "$optEquality" },
    
    -- stringing the expression together with operators
    {"$optOperator",        "+", "#PushType", "$exprVal", "#PopType", "#Add", "$optOperator" },
    {"$optOperator",        "&", "$exprVal", "#Concat", "$optOperator" },
    {"$optOperator",        "-", "$exprVal", "#Sub", "$optOperator" },
    {"$optOperator",        "*", "$exprVal", "#Mul", "$optOperator" },
    {"$optOperator",        "/", "$exprVal", "#Div", "$optOperator" },
    {"$optOperator",        "\\", "$exprVal", "#IntDiv", "$optOperator" },
    {"$optOperator",        "^", "$exprVal", "#Power", "$optOperator" },
    {"$optOperator",        "mod", "$exprVal", "#Mod", "$optOperator" },
    {"$optOperator",        "NULL" },

    -- these follow optional operators

    -- string values
    {"$exprVal",            "%builtinStr", "#Func{", "$argList", "#}Func", "#StrType" },
    {"$exprVal",            "%userStrFunc", "#Func{", "$argList", "#}UserFunc", "#StrType" },
    {"$exprVal",            "%string", "#PushExpr", "#StrType" },

    -- numeric values
    {"$exprVal",            "not", "$basicExpr", "#Not", "#NumType" },
    {"$exprVal",            "%builtinNum", "#Func{", "$argList", "#}Func", "#NumType" },
    {"$exprVal",            "%userFunc", "#Func{", "$argList", "#}UserFunc", "#NumType" },
    {"$exprVal",            "%number", "#PushExpr", "#NumType" },
    {"$exprVal",            "%variable", "BACKUP", "$variable" },
    {"$exprVal",            "(", "$expr", ")", "#Parens", "#NumType" },
    {"$exprVal",            "-", "$exprVal", "#UnaryMinus", "#NumType" },


    {"$optEquality",       "=", "$basicExpr", "#Eq", "#NumType" },
    {"$optEquality",       ">", "$basicExpr", "#Gt", "#NumType" },
    {"$optEquality",       "<", "$basicExpr", "#Lt", "#NumType" },
    {"$optEquality",       "!=", "$basicExpr", "#Ne", "#NumType" },
    {"$optEquality",       ">=", "$basicExpr", "#Ge", "#NumType" },
    {"$optEquality",       "<=", "$basicExpr", "#Le", "#NumType" },
    {"$optEquality",       "NULL" },
    
    {"$exprEquality",      "$basicExpr", "$optEquality" },
    
    {"$logicTail",          "and", "$exprEquality", "#And", "#NumType", "$logicTail" },
    {"$logicTail",          "or", "$exprEquality", "#Or", "#NumType", "$logicTail" },
    {"$logicTail",          "xor", "$exprEquality", "#Xor", "#NumType", "$logicTail" },    
    {"$logicTail",          "NULL" },

    {"$expr",               "$exprEquality", "$logicTail" }}


-----------------------------------------------------------------------------
global atom thisToken
global sequence userSubs, userFuncs, labels
global sequence tokens, tokenValue, actions


global constant builtin_functions = {
    { "abs", 1, "qb_abs( %s )" },
    { "asc", 1, "qb_asc( %s )" },
    { "atn", 1, "arctan( %s )" },
    { "chr$", 1, "{%s}" },
    { "cint", 1, "%s" },
    { "clng", 1, "%s" },
    { "command$", 0, "qb_command()" },
    { "cos", 1, "cos(%s)" },
    { "csng", 1, "%s" },
    { "csrlin", 1, "qb_csrlin(%s)" },
    { "cvd", 1, "%s" },
    { "cvdmbf", 1, "%s" },
    { "cvi", 1, "%s" },
    { "cvl", 1, "%s" },
    { "cvsmbf", 1, "%s" },
    { "date$", 0, "qb_date( %s )" },
    { "eof", 1, "file_eof( %s )" },
    { "environ$", 0, "qb_environ(%s)" },
    { "exp", 1, "exp( %s )" },
    { "fix", 1, "qb_fix( %s )" },
    { "hex$", 1, "sprintf( \"%%x\", {%s} )" },
    { "inkey$", 0, "qb_inkey()" },
    { "instr", 2, "qb_instr2( %s, %s )" },
    { "instr", 3, "qb_instr3( %s, %s, %s )" },
    { "inp", 1, "qb_inp( %s )" },
    { "int", 1, "floor(%s)" },
    { "left$", 2, "qb_left(%s, %s)" },
    { "ltrim$", 1, "qb_ltrim( %s )" },
    { "len", 1, "length(%s)" },
    { "lcase$", 1, "lower(%s)" },
    { "log", 1, "log( %s )" },
    { "mid$", 2, "qb_mid(%s, %s, MISSING)" },
    { "mid$", 3, "qb_mid(%s, %s, %s)" },
    { "oct$", 0, "sprintf(\"%%o\",{%s})" },
    { "peek", 1, "qb_peek( %s )" },
    { "pos", 1, "qb_pos( %s )" },
    { "right$", 2, "qb_right(%s, %s)" },
    { "rtrim$", 1, "qb_rtrim( %s )" },
    { "rnd", 0, "qb_rnd()" },
    { "rnd", 1, "qb_rnd()" },
    { "screen", 3, "qb_fn_screen( %s, %s, %s )" },
    { "screen", 2, "qb_fn_screen( %s, %s, 0 )" },
    { "sgn", 1, "compare( %s, 0 )" },
    { "sin", 1, "sin(%s)" },
    { "spc", 1, "repeat( 32, %s )" },
    { "space$", 1, "repeat( 32, %s )" },
    { "sqr", 1, "sqrt( %s )" },
    { "str$", 1, "sprintf( \" %%d \", {%s} )" },
    { "string$", 2, "qb_string(%s, %s)" },
    { "tan", 1, "tan(%s)" },
    { "timer", 0, "time()" },
    { "time$", 0, "qb_time()" },
    { "ucase$", 1, "upper(%s)" },
    { "val", 1, "qb_val(%s)" },
    
    -- included for compatibility with Euphoria
    { "repeat$", 2, "qb_string(%s, %s)" }}

global sequence builtin_subs 
    builtin_subs = {
    
    { "beep", 0, "qb_beep()\n" },
    { "cls", 0, "clear_screen()\n" },
    { "color", 1, "text_color( %s )\n" },
    { "color", 2, "text_color( %s )  bk_color( %s )\n" },    
    { "chdir", 1, "system( \"cd \" & %s, 2 )\n" },
    { "kill", 1, "system( \"del \" & %s, 2 )\n" },
    { "mkdir", 1, "system( \"mkdir \" & %s, 2 )\n" },
    { "out", 2, "qb_out( %s, %s )\n" },
    { "poke", 2, "qb_poke( %s, %s )\n" },
    { "play", 1, "qb_play( %s )\n" },
    { "restore", 0, "dataPos = 0\n" },    
    { "rmdir", 1, "system( \"rmdir \" & %s, 2 )\n" },
    { "screen", 4, "qb_sub_screen( %s, %s, %s, %s )\n" },
    { "screen", 3, "qb_sub_screen( %s, %s, %s, MISSING )\n" },
    { "screen", 2, "qb_sub_screen( %s, %s, MISSING, MISSING )\n" },
    { "screen", 1, "qb_sub_screen( %s, MISSING, MISSING, MISSING )\n" },
    { "shell", 1, "system( %s, 2 )\n" },
    { "sleep", 1, "qb_sleep( %s )\n" },
    { "sound", 2, "qb_sound( %s, %s)\n" },
    { "stop", 0, "set_text_mode()  abort(0)\n" },
    { "system", 0, "abort(0)\n" },
    { "wait", 1, "qb_sleep( %s )\n" },
    { "width", 1, "qb_width( %s, MISSING )\n" },
    { "width", 2, "qb_width( %s, %s )\n" }}

-----------------------------------------------------------------------------
global constant keywords = { 
    "print", "lprint", "tab", "using",
    "input", "line",
    "open", "output", "as", "close",
    "if", "then", "else", "elsif", "elseif",
    "do", "while", "until", "loop", "exit",
    "for", "to", "step", "by", "next", "append", "random", "output",
    "shared", "lock",
    "while", "wend",
    "call", "sub", "function", "end", "static",
    "goto", "gosub", "return",
    "type", "as", "single", "double", "string", "integer", "long",
    "and", "or", "xor", "not", "mod", 
    "select", "case", "is",
    "data", "read",
    "const",
    "defint", "deflng", "defsng", "defdbl", "defstr",
    "randomize", "declare",
    "dim", "redim", "common", "global", "shared",
    "name",
    "def", "seg",
    "locate",
    "swap",
    "on", "error", "resume",
    "let",
    "option", "base" }

                                                                        
-----------------------------------------------------------------------------
-- first tokens in an expression
global constant ExprHeads = {
    "%builtinStr",
    "%userStrFunc",
    "%string",
    "not",
    "%builtinNum",
    "%userFunc", 
    "%number",
    "%variable",
    "(", 
    "-" }

-----------------------------------------------------------------------------
-- delimit end of commands                 
global constant EndMarkers = {
    "end",
    ":",
    "<eol>",
    "else",
    "%comment" }




-- token assignment
global constant
    TokenValue  = 1,
    TokenType   = 2
    
--
-- SUPPORT TOOLS
--              


-----------------------------------------------------------------------------
global function last( sequence s )

    -- return last item in a sequence
    if length( s ) then
        return s[length(s)]
    else
        return {}
    end if
    
end function 


-----------------------------------------------------------------------------
global function rest( sequence s )

    -- return everything but first item in sequence
    if length( s ) then
        return s[1..length(s)-1]
    else
        return {}
    end if
    
end function 


-----------------------------------------------------------------------------
global function a_find( sequence s, sequence list )

    -- return index in associated list
    -- list items are in the form { key, ... }

    for i = 1 to length( list ) do
        if compare( s, list[i][1] ) = 0 then
            return i
        end if
    end for
    return 0

end function


-----------------------------------------------------------------------------
global function reserved_word( sequence s )

    -- return true s is a key word
    -- also checks for label name

    -- empty?
    if length( s ) then

        -- set to lower case
        s = lower( s )
    
        -- remove colon if on end    
        if last(s) = ':' then
            s = s[1..length(s)-1]
        end if
    
        if find( s, keywords )                  then return -1
        elsif a_find( s, builtin_functions )    then return -1        
        elsif a_find( s, builtin_subs )         then return -1
                                                else return 0
        end if
        
    else
    
        return 0
        
    end if
        
    
end function



-----------------------------------------------------------------------------
global function remove_numeric_suffix( sequence s )

    -- removes suffix, if not '$'
    if find( last(s), SUFFIXES ) then
        if last(s) != '$' then
            return rest(s)
        end if
    end if
    
    return s
    
end function
    

-----------------------------------------------------------------------------
global function add_token_types( sequence tokens )

    -- add token types to tokens

    integer at, firstToken
    sequence lToken, pToken

    for i = 1 to length( tokens ) do

        -- lower case token
        lToken = lower( tokens[i] )
        
        -- plain (no suffix) token
        pToken = remove_numeric_suffix( tokens[i] )
        
        if compare( lToken, "spc" ) = 0 then
            -- looks like numeric, but returns a string
                tokens[i] = { lToken, "%builtinStr" }

        elsif compare( lToken, "mid$" ) = 0
        or    compare( lToken, "screen" ) = 0 then

            -- this is a kludge to handle something that
            -- can be either a function and a command

            -- try to determine if this is the first token
            if i = 1 then 
                -- first token on line                              
                firstToken = 1
            elsif compare( tokens[i-1][1], ":" ) = 0 then  
                -- first token following colon
                firstToken = 1
            else
                -- failure
                firstToken = 0
            end if


            -- determine how to handle            
            if firstToken then
            
                -- built-in, or handled by parser?
                if a_find( lToken, builtin_subs ) then
                    
                    -- builtin sub
                    tokens[i] = { lToken, "%builtinSub" }            
                else                    

                    -- give to parser
                    tokens[i] = { lToken, lToken }
                    
                end if

            elsif find( tokens[i], EndMarkers ) then
            
                -- key word, or built-in sub?
                if a_find( lToken, builtin_subs ) then
                
                    -- builtin sub
                    tokens[i] = { lToken, "%builtinSub" }            
                
                else                
                    
                    -- give to parser
                    tokens[i] = { lToken, lToken }
                    
                end if
                
            elsif last( tokens[i] ) = '$' then
                -- string function
                tokens[i] = { lToken, "%builtinStr" }
            else
                -- numeric function
                tokens[i] = { lToken, "%builtinNum" }
            end if

        elsif tokens[i][1] = '"' then
            -- string literal
            tokens[i] = { tokens[i], "%string" }

        elsif tokens[i][1] = '\'' then
            -- comment
            tokens[i] = { tokens[i], "%comment" }

        elsif a_find( lToken, builtin_functions ) then
            -- builtin function

            if last(tokens[i]) = '$' then
                -- string function
                tokens[i] = { lToken, "%builtinStr" }
            else
                -- numeric function
                tokens[i] = { lToken, "%builtinNum" }
            end if

        elsif a_find( lToken, builtin_subs ) then
            -- builtin function

            tokens[i] = { lToken, "%builtinSub" }            


        elsif find( lToken, keywords ) then
            -- key word
            tokens[i] = { lToken, lToken }

        elsif find( tokens[i], userSubs ) then
            -- user defined subroutine
            tokens[i] = { tokens[i], "%userSub" }
            
        elsif find( pToken, userFuncs ) then
            -- user defined subroutine         
            if last(tokens[i]) = '$' then
                tokens[i] = { tokens[i], "%userStrFunc" }
            else
                tokens[i] = { tokens[i], "%userFunc" }
            end if

        elsif tokens[i][1] = '.' then

            -- number, or element
            if tokens[i][2] < '0'
            or tokens[i][2] > '9' then

                -- element
                tokens[i] = { tokens[i], "%element" }

            else
                -- number
                tokens[i] = { tokens[i], "%number" }

            end if 

        elsif find( tokens[i][1], "0123456789." ) then
        
            -- save as number
            tokens[i] = { tokens[i], "%number" }
       
        elsif tokens[i][1] = '#'
        and   length( tokens[i] ) > 1 then

            -- hex number
            tokens[i] = { tokens[i], "%number" }

        elsif length(tokens[i]) > 1         -- not just colon
        and   last(tokens[i]) = ':'         -- ends in colon
        and   tokens[i][1] != '\'' then     -- not a comment
        
            tokens[i] = { tokens[i][1..length(tokens[i])-1], "%label" }

        elsif   (tokens[i][1] >= 'A' and tokens[i][1] <= 'Z')
        or      (tokens[i][1] >= 'a' and tokens[i][1] <= 'z') then

            -- an alphanumeric symbol, but what kind?

            -- "found" flag
            at = 0

            -- check types
            if at = 0 then
                at = a_find( tokens[i], types )
                if at != 0 then
                    tokens[i] = { tokens[i], "%userType" }
                end if
            end if

            -- check elements
            if at = 0
            and not hideElements then
                at = a_find( tokens[i], elements )
                if at != 0 then
                    tokens[i] = { tokens[i], "%element" }
                end if
            end if

            -- check constants
            if at = 0 then
                at = a_find( tokens[i], constants )
                if at != 0 then
                    -- type of constant?
                    tokens[i] = { tokens[i], "%variable" }                    
                end if
            end if

            -- check local strings
            if at = 0 then  -- skip if prior find
                at = a_find( tokens[i], locals[STR] )
                if at != 0 then                    
                    tokens[i] = { tokens[i], "%variable" }
                end if
            end if

            -- check local string arrays
            if at = 0 then  -- skip if prior find
                at = a_find( tokens[i], locals[STRS] )
                if at != 0 then
                    tokens[i] = { tokens[i], "%variable" }                
                end if
            end if

            -- check global strings
            if at = 0 then  -- skip if prior find
                at = a_find( tokens[i], globals[STR] )
                if at != 0 then
                    -- hidden?
                    if globals[STR][at][V_SCOPE] = V_HIDDEN
                    or useLocals then
                        -- hidden
                        at = 0
                    else
                        -- match
                        tokens[i] = { tokens[i], "%variable" }
                    end if
                else
                    -- clear flag
                    at = 0
                end if
            end if

            -- check global string arrays
            if at = 0 then  -- skip if prior find
                at = a_find( tokens[i], globals[STRS] )
                if at != 0 then
                    -- visible?
                    if globals[STRS][at][V_SCOPE] = V_HIDDEN
                    or useLocals then
                        -- hidden
                        at = 0
                    else
                        -- visible
                        tokens[i] = { tokens[i], "%variable" }
                    end if
                else
                    -- clear flag
                    at = 0
                end if
            end if

            -- default to numeric
            if at = 0 then  -- no match
                tokens[i] = { tokens[i], "%variable" }
            end if
            

        else
        
            -- use literal value
            tokens[i] = { tokens[i], tokens[i] }
            
        end if

        -- clean up suffixes from numbers
        if compare( tokens[i][2], "%number" ) = 0 then
            -- trailing type indicator?    
            tokens[i][2] = remove_numeric_suffix( tokens[i][2] )
        end if

        

    end for

    -- print in trace file
    for i = 1 to length( tokens ) do
        puts( traceFile, tokens[i][1] & " " )
    end for
    puts( traceFile, "\n" )
    
    for i = 1 to length( tokens ) do
        puts( traceFile, tokens[i][2] & " " )
    end for
    puts( traceFile, "\n" )

    return tokens

end function


-----------------------------------------------------------------------------
global function tokenize( sequence text )

    -- converts a string of text into tokens
    -- in the form:
    -- { { "token", "value" } ... }

    atom char, nextChar 
    integer i, flag, at
    sequence words, word
    object tmp
    

    word = ""
    words = {}
    text = text & "  "

    -- convert text into tokens
    i = 1
    while i < length( text ) do
        
        -- current char
        char = text[i] 
        
        -- following char
        if i < length( text ) then
            nextChar = text[i+1]
        else
            nextChar = ' '
        end if


        -- normal word        
        if ( char >= 'A' and char <= 'Z' )
        or ( char >= 'a' and char <= 'z' )
        or ( char >= '0' and char <= '9' )
        or char = '_' then

            --part of a word
            word = word & char
        
        elsif char = '.' then

            -- check to see if a valid element name follows
            -- otherwise, it's part of the same word

            at = 0
            for j = 1 to length( elements ) do
                -- don't slice off the end
                if i + length( elements[j][E_NAME] ) < length( text ) then
                    if compare( text[i+1..i+length(elements[j][E_NAME])],
                                    elements[j][E_NAME] ) = 0 then

                        -- match
                        at = j

                        -- leave loop
                        exit

                    end if
                end if
            end for

            -- match?
            if at then

                -- clear match flag
                at = 0

                if text[i-1] = ')' then
                    -- foo(3).element
                    at = 1

                elsif word[1] = '.' then
                    -- foo.bar.element
                    at = 1
    
                else
                    -- see if accumulated word is a variable

                    -- search in locals
                    at = a_find( word, locals[USER_TYPE] )
                    
                    -- search in globals                    
                    if at = 0 then  -- no prior match
                        at = a_find( word, globals[USER_TYPE] )
                        if at then  -- found?
                            -- hidden ?
                            if globals[USER_TYPE][at][V_SCOPE] = V_HIDDEN 
                            and useLocals then
                                -- hidden, no match.
                                at = 0
                            end if
                        end if
                    end if
                end if
            end if
                    
            if at then
                -- element name

                -- add pending word
                if length( word ) then
                    words = append( words, word )
                    word = ""
                end if

            end if

            -- add to word
            word = word & "."


        elsif char = ':' then

            -- part of word?
            if length( word ) > 0               -- word accumulated?
            and length( words ) = 0             -- first word?
            and not reserved_word( word ) then  -- not a reserved word
            
                -- part of label
                word = word & char              
                
            else
                
                -- append word
                if length( word ) > 0 then
                    words = append( words, word )
                    word = ""
                end if
                
                -- divider; treat as end of line
                words = append( words, ":" )
                
            end if

        -- comment
        elsif char = '\''
        or    ( compare( lower(word), "rem" ) = 0 and char = ' ' ) then

            -- strip comment; make it a ' comment
            word = "'" & text[i+1..length( text )-3]
            words = append( words,  word )
            word = ""

            -- append eol
            words = append( words, "<eol>" )
            i = length( text )

        -- two character token
        elsif ( char = '!' and nextChar = '=' )     
        or    ( char = '<' and nextChar = '>' )
        or    ( char = '>' and nextChar = '=' )
        or    ( char = '<' and nextChar = '=' ) then
                
            --delimiter
            if length( word ) > 0 then
                --add accumulated word, if any
                words = append( words, word )
            end if                              
            
            -- convert <> to !=
            word = text[i..i+1]
            if compare( word, "<>" ) = 0 then
                word = "!="
            end if
            
            -- add word
            words = append( words, word )    
            word = ""        
            
            -- move ahead 1 byte
            i = i + 1

        elsif char = '"' then
            -- string
            
            -- save current word
            if length( word ) > 0 then
                --add accumulated word, if any
                words = append( words, word )
                word = ""
            end if

            -- read to next
            word = {char}
            i = i + 1
            while   text[i] != '"'
            and     i < length(text)-1 do
                -- accumulate
                if text[i] = '\\' then
                    word = word & "\\\\"
                else
                    word = word & text[i]
                end if

                -- move ahead
                i = i + 1
            end while

            -- end of string found?            
            if text[i] != '"' then
                -- error
                errText = "Expected '\"', not <eol>." 
                return {}
            else                     
                -- add last quote
                word = word & text[i]                
            end if

        elsif char = '&'
        and   nextChar = 'H'
        and   length( word ) = 0 then

            -- hex number
            word = "#"
            
            -- move ahead 1 byte
            i = i + 1

        elsif find( char, SUFFIXES )
        and   length( word ) > 0 then
            -- data type indicator
            
            -- append to current word
            word = word & char
            words = append( words, word )
            word = ""


        -- one character token            
        else              
            --delimiter
            if length( word ) > 0 then
                --add accumulated word, if any
                words = append( words, word )
                word = ""
            end if
            
            --treat delimiter as word, if not white space
            if char = '\n' then
                -- end of line is special token
                words = append( words, "<eol>" )
                
            elsif char = ' '
            or    char = '\t' then
                -- white space. ignore
            
            else
                -- single byte token
                words = append( words, {char} )
            end if
        end if        

        -- increment
        i = i + 1    
    
    end while

    return words
        
end function


-----------------------------------------------------------------------------    
global procedure expected_err( sequence expected, sequence actual )

    -- convert expected parse to an error message
    -- errText is a global    

        -- high level token?
        if length( expected ) > 1
        and find( expected[1], "$%" ) then

            -- remove $ or %
            expected = expected[2..length(expected)]                
            
            -- place "a" or "an" in front
            if find( expected[1], "aeiouAEIOU" ) then
                expected = "an " & expected
            else
                expected = "a " & expected
            end if
            
            errText = "Expected " & expected
        else
            -- print expected token value                
            errText = "Expected \"" & expected & "\""
        end if

        -- print errant token        
        if actual[1] = '"' then
            -- string
            errText = errText & ", not the string " & actual & "."        
        else
            -- token        
            errText = errText & ", not \"" & actual & "\"."
        end if

end procedure


-----------------------------------------------------------------------------    
global procedure parse( sequence rule, integer level )

    integer isExpr, isEnd
    atom matchRule
    sequence token, tokenVal

    tokenValue = ""
    
    while length( rule ) > 0            -- more steps in rule
    and   length( errText ) = 0  do     -- no error

        -- get next token    
        if thisToken > length( tokens ) then
            token = "<eol>"
            tokenVal = ""
        else
            token = tokens[thisToken][TokenType]
            tokenVal = tokens[thisToken][TokenValue]
        end if

        -- check if it could be part of an expression
        isExpr = find( token, ExprHeads )
        isEnd = find( token, EndMarkers )

        if compare( rule[1], token ) = 0 
        or compare( rule[1], "ANY" ) = 0 then

-- trace file
printf( traceFile, "\t%s  %s\n", {rule[1],tokenVal} )

            -- store value of last matched token
            if thisToken > length( tokens ) then
                tokenValue = ""    
            else
                tokenValue = tokens[thisToken][TokenValue]
            end if
            
            --move ahead to next token
            thisToken = thisToken + 1

        elsif rule[1][1] = '#' then
            --perform action

printf( traceFile, "\t\t%s\n", {rule[1]} )

            actions = append( actions, { rule[1], tokenValue } )

        elsif compare( rule[1], "NULL" ) = 0 then
            -- automatic match

        elsif compare( rule[1], "?EXPR" ) = 0
        and   isExpr then
            -- match

        elsif compare( rule[1], "?END" ) = 0
        and   isEnd then
            -- match

        elsif compare( rule[1], "BACKUP" ) = 0 then
            -- backup 1 token

            if thisToken > 0 then
                thisToken = thisToken - 1
            end if
            
        elsif rule[1][1] = '$' then
        
            -- puts( 1, "Find rule \"" & rule[1] & "\"\n" )
        
            -- is current token an expression?
            isExpr = find( token, ExprHeads )
            isEnd = find( token, EndMarkers )

            -- find matching rule
            matchRule = 0
            for i = 1 to length( grammar ) do
                -- same head?
                if compare( grammar[i][1], rule[1] ) = 0 then

                    -- match first word (or null, or definition)                
                    if compare( grammar[i][2], token ) = 0
                    or compare( grammar[i][2], "NULL" ) = 0
                    or compare( grammar[i][2], "ANY" ) = 0
                    or compare( grammar[i][2], "BACKUP" ) = 0                    
                    or grammar[i][2][1] = '$'
                    or ( compare( grammar[i][2], "?EXPR" ) = 0 and isExpr )
                    or ( compare( grammar[i][2], "?END" ) = 0 and isEnd )
                    then
                                                                         
            
                        -- match

                        -- save rule
                        matchRule = i
                        -- exit loop
                        exit
                    end if
                end if
            end for

            if matchRule = 0 then            

                -- no matching rule found    
                if thisToken > length( tokens ) then
                    expected_err( rule[1], "<eol>" )                
                else            
                    expected_err( rule[1], tokens[thisToken][TokenValue] )
                end if
                                
            else
                -- recurse
printf( traceFile, "RULE: %s\n", {grammar[matchRule][1]} )
                parse( grammar[matchRule][2..length(grammar[matchRule])],
                        level+1 )
                        
                -- error?
                if length( errText ) > 0 then                 
                    return
                end if
            end if
            
                
        else
                -- failed match
                if thisToken > length( tokens ) then
                    expected_err( rule[1], "<eof>" )                
                else            
                    expected_err( rule[1], tokens[thisToken][TokenValue] )
                end if
                        
            -- exit
            return
                                    
        end if

        -- move ahead
        rule = rule[2..length( rule )]
        
    end while    

end procedure


